theory Rule_Labeling_Impl
imports
  Decreasing_Diagrams2
  Critical_Pairs_Impl
  Map_Choice
begin

abbreviation DD2_last where "DD2_last \<equiv> Decreasing_Diagrams2.last"
  
(* TODO: consolidate with critical_pairs(_impl) *)
definition critical_peaks_impl :: "('f,string)rules \<Rightarrow> ('f,string)rules \<Rightarrow> (bool \<times> ('f,string) lpeak) list" 
  where "critical_peaks_impl P R \<equiv> concat (map (\<lambda> (l,r). 
    concat (map (\<lambda> p. let C = ctxt_of_pos_term p l; l'' = l |_ p; b = (C = \<box>) in 
      if is_Var l'' then [] else 
      concat (map (\<lambda> (l',r'). case mgu_var_disjoint_string l'' l' of 
        Some (\<sigma>,\<tau>) \<Rightarrow> [(b, (l \<cdot> \<sigma>, (l, r), \<epsilon>, \<sigma>, r \<cdot> \<sigma>), (l \<cdot> \<sigma>, (l', r'), p, \<tau>, (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>))] 
      | None \<Rightarrow> []) R)) (poss_list l))) P)"

lemma critical_peaks_impl: 
  "set (critical_peaks_impl R R) = critical_peaks (set R)" (is "?l = ?r")
proof -
  note cpdefs = critical_peaks_impl_def critical_peaks_def set_concat
    set_map Let_def poss_list_sound
  {
    fix b s1 s2
    assume "(b,s1,s2) \<in> ?r"
    from this[unfolded cpdefs]
    obtain l r l' r' l'' C \<sigma> \<tau> where
      b: "b = (C = \<box>)" and s1: "s1 = (l \<cdot> \<sigma>, (l, r), \<epsilon>, \<sigma>, r \<cdot> \<sigma>)" 
      and s2: "s2 = ( l \<cdot> \<sigma>, (l', r'), hole_pos C, \<tau>, (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>)"
      and lr: "(l, r) \<in> set R" and lr': "(l', r') \<in> set R"
      and l: "l = C\<langle>l''\<rangle>" and l'': "is_Fun l''"
      and mgu: "mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>)"
      by auto
    let ?p = "hole_pos C"
    from l have p: "?p \<in> poss l" by auto
    from l p have C: "C = ctxt_of_pos_term ?p l" by auto
    from l p have l: "l |_ ?p = l''" by auto
    have "(b,s1,s2) \<in> ?l" unfolding cpdefs b s1 s2
      by (rule, rule, rule lr, unfold cpdefs, rule, rule imageI[OF p],
      insert C l lr' l'' mgu, force)
  } 
  hence "?r \<subseteq> ?l" by auto
  moreover
  {
    fix b s1 s2
    assume "(b,s1,s2) \<in> ?l"
    from this[unfolded cpdefs]
    obtain l r p l' r' \<sigma> \<tau> where
      lr: "(l,r) \<in> set R" and p: "p \<in> poss l" and lp: "is_Fun (l |_ p)" 
      and lr': "(l',r') \<in> set R" 
      and mgu: "Some (\<sigma>,\<tau>) = mgu_var_disjoint_string (l |_ p) l'" 
      and b: "b = (ctxt_of_pos_term p l = \<box>)" 
      and s1: "s1 = (l \<cdot> \<sigma>, (l, r), \<epsilon>, \<sigma>, r \<cdot> \<sigma>)" 
      and s2: "s2 = (l \<cdot> \<sigma>, (l', r'), p, \<tau>, (ctxt_of_pos_term p l \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>)"  
    by force  
    from p have pp: "hole_pos (ctxt_of_pos_term p l) = p" by simp
    from s1 have s: "get_target s1 = r \<cdot> \<sigma>" unfolding get_target_def by simp    
    from s2 have t: "get_target s2 = (ctxt_of_pos_term p l \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>" 
      unfolding get_target_def by simp    
    from critical_peaksI[OF lr lr' _ lp mgu[symmetric] s t b] 
    have "(b,s1,s2) \<in> ?r" using  ctxt_supt_id[OF p, symmetric] s1 s2 pp
    unfolding get_target_def by auto
  }
  hence "?l \<subseteq> ?r" by auto
  ultimately show ?thesis by auto
qed

type_synonym ('f, 'v) rule_lab_repr = "(('f, 'v) rule \<times> nat) list"

definition rule_lab_repr_to_lab :: "('f :: key, 'v :: key ) rule_lab_repr \<Rightarrow> ('f, 'v) rule \<Rightarrow> nat"
where
  "rule_lab_repr_to_lab ps = fun_of_map (ceta_map_of ps) 0"

fun rseq_to_step_list :: "('f, 'v) term \<Rightarrow> ('f, 'v) rseq \<Rightarrow> ('f, 'v) step list"
where
  "rseq_to_step_list s [] = []"
| "rseq_to_step_list s ((p,lr,t) # steps) = (
     let u = subt_at s p;
         v = subt_at t p;
         \<sigma> = (case match_list Var [(fst lr, u), (snd lr, v)] of
           Some \<tau> \<Rightarrow> \<tau>
         | None \<Rightarrow> undefined)
      in (s,lr,p,\<sigma>,t) # rseq_to_step_list t steps)"

primrec rseq_to_ddseq :: "('f, 'v) term \<times> ('f, 'v) rseq \<Rightarrow> ('f, 'v) seq"
  where "rseq_to_ddseq (s, rseq) = (s, rseq_to_step_list s rseq)"

lemma check_rsteps'_sound_dd_seq:
  assumes "isOK (check_rsteps' R rseq s w)"
  shows "rseq_to_ddseq (s, rseq) \<in> seq (set R) \<and> DD2_last (rseq_to_ddseq (s, rseq)) = w"
using assms
proof (induct rseq arbitrary: s)
  case Nil
  have "rseq_to_ddseq (s, []) \<in> seq (set R)" using seq.intros by auto
  moreover have "DD2_last (rseq_to_ddseq (s, [])) = s" by auto
  moreover have "s = w" using Nil check_rsteps_sound by fastforce
  ultimately show ?case by auto
next
  case (Cons r rs)
  show ?case 
  proof (cases r)
    case (fields p lr t)
    with Cons have IH: "rseq_to_ddseq (t, rs) \<in> seq (set R)" and 
      "DD2_last (rseq_to_ddseq (t, rs)) = w" by auto
    with fields have last:"DD2_last (rseq_to_ddseq (s, r # rs)) = w" 
      by (simp add: get_target_def)
    let ?\<sigma> = "case match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)] of Some \<tau> \<Rightarrow> \<tau>"
    from fields Cons(2) obtain \<tau> where 
      "lr \<in> set R" and "p \<in> poss s" and "p \<in> poss t" and 
      "ctxt_of_pos_term p s = ctxt_of_pos_term p t" and
      "match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)] = Some \<tau>" 
      apply (simp del: check_rstep')
      unfolding check_rstep'_def unfolding check_prop_rstep'_def 
      unfolding check_prop_rstep_rule_def unfolding Let_def 
      apply (cases "match_list Var [(fst lr, s |_ p), (snd lr, t |_ p)]") by auto
    moreover from this  match_list_matches have "(fst lr) \<cdot> ?\<sigma> = s |_ p \<and> (snd lr) \<cdot> ?\<sigma> = t |_ p" 
      by fastforce
    ultimately have "(s, t) \<in> rstep_r_p_s (set R) lr p ?\<sigma>" 
      unfolding rstep_r_p_s_def' by (cases lr, auto simp: replace_at_ident)
    with IH have "rseq_to_ddseq (s, r # rs) \<in> seq (set R)" using fields seq.intros by fastforce
    with last show ?thesis by blast
  qed
qed

definition check_ELD_1'_nat :: "nat \<Rightarrow> nat \<Rightarrow> nat list \<Rightarrow> nat list \<Rightarrow> nat list \<Rightarrow> shows check"
where
  "check_ELD_1'_nat \<beta> \<alpha> \<sigma>1 \<sigma>2 \<sigma>3 = do {
     check_allm (\<lambda>x. check (x < \<beta>) (shows '' labels are not decreasing'')) \<sigma>1;
     check_allm (\<lambda>x. check (x \<le> \<alpha>) (shows '' labels are not decreasing'')) \<sigma>2;
     check (length \<sigma>2 \<le> 1) (shows '' length of middle sequence is greater 1 '');
     check_allm (\<lambda>x. check (x < \<alpha> \<or> x < \<beta>) (shows '' labels are not decreasing'')) \<sigma>3
   }"

lemma check_ELD_1'_nat :
  assumes "isOK (check_ELD_1'_nat \<beta> \<alpha> \<sigma>1 \<sigma>2 \<sigma>3)"
  shows "ELD_1' ({(n, m). n < m},{(n, m). n \<le> m}) \<beta> \<alpha> \<sigma>1 \<sigma>2 \<sigma>3"
using assms
unfolding check_ELD_1'_nat_def ELD_1'_def ds_def
by auto

definition split_seq :: "nat \<Rightarrow> nat \<Rightarrow> nat list \<Rightarrow> nat list \<times> nat list \<times> nat list"
where
  "split_seq \<alpha> \<beta> ss = (
    let (ss', r) = span (\<lambda>n. n < \<alpha>) ss in
      case r of
        Nil \<Rightarrow> (ss',[],[])
      | Cons h t \<Rightarrow> if h \<le> \<beta> then (ss',[h],t) else (ss',[],r)
  )"

lemma split_seq_is_split:
  assumes "split_seq \<alpha> \<beta> ss = (ss1,ss2,ss3)"
  shows "ss1 @ ss2 @ ss3 = ss"
using assms 
apply (auto simp add: split_seq_def) 
apply (split list.splits, auto)
apply (split if_splits, auto) 
by (metis takeWhile_dropWhile_id)

definition 
  check_ELD_rule_labeling :: "(('f, 'v) rule \<Rightarrow> nat) \<Rightarrow> ('f,'v) rule \<Rightarrow> ('f, 'v) rule \<Rightarrow> ('f,'v) seq \<Rightarrow> ('f, 'v) seq \<Rightarrow> shows check"
where
  "check_ELD_rule_labeling lab r1 r2 j1 j2 = do {
    let \<alpha> = lab r1;
    let \<beta> = lab r2;
    let \<tau> = map (rule_labeling lab) (snd j1);
    let \<sigma> = map (rule_labeling lab) (snd j2);
    let (\<tau>1,\<tau>2,\<tau>3) = split_seq \<alpha> \<beta> \<tau>;
    let (\<sigma>1,\<sigma>2,\<sigma>3) = split_seq \<beta> \<alpha> \<sigma>;
    check_ELD_1'_nat \<alpha> \<beta> \<tau>1 \<tau>2 \<tau>3;
    check_ELD_1'_nat \<beta> \<alpha> \<sigma>1 \<sigma>2 \<sigma>3
  }"
  
lemma check_ELD_rule_labeling :
  assumes "isOK (check_ELD_rule_labeling lab r1 r2 j1 j2)"
  shows "eld_seq (rule_labeling lab) ({(n, m) . n < m}, {(n, m) . n \<le> m}) ((s1,r1,p1,\<rho>1,t1),(s2,r2,p2,\<rho>2,t2)) j1 j2" 
  (is "eld_seq ?l ?r ?p _ _")
proof -
  from assms obtain \<sigma>1 \<sigma>2 \<sigma>3 \<tau>1 \<tau>2 \<tau>3  where 
    \<sigma>: "(\<sigma>1,\<sigma>2,\<sigma>3) = split_seq (lab r2) (lab r1) (map ?l (snd j2))" and
    \<tau>: "(\<tau>1,\<tau>2,\<tau>3) = split_seq (lab r1) (lab r2) (map ?l (snd j1))" and
    \<sigma>ok: "isOK (check_ELD_1'_nat (lab r2) (lab r1) \<sigma>1 \<sigma>2 \<sigma>3)" and
    \<tau>ok: "isOK (check_ELD_1'_nat (lab r1) (lab r2) \<tau>1 \<tau>2 \<tau>3)" 
    unfolding check_ELD_rule_labeling_def
    unfolding Let_def by auto
  show ?thesis unfolding eld_seq_def
  proof (intro exI conjI)
    from \<sigma> show "map ?l (snd j2) = \<sigma>1 @ \<sigma>2 @ \<sigma>3"
      using split_seq_is_split by auto
    from \<tau> show "map ?l (snd j1) = \<tau>1 @ \<tau>2 @ \<tau>3"
      using split_seq_is_split by auto 
    from check_ELD_1'_nat[OF \<tau>ok] 
    show "ELD_1' ?r  (?l (fst ?p)) (?l (snd ?p)) \<tau>1 \<tau>2 \<tau>3"
    by auto
    from check_ELD_1'_nat[OF \<sigma>ok] 
    show "ELD_1' ?r  (?l (snd ?p)) (?l (fst ?p))  \<sigma>1 \<sigma>2 \<sigma>3"
    by auto
  qed
qed

definition check_cpeak_eld :: 
  "('f :: show, 'v :: show) rules \<Rightarrow> ('f, 'v) lpeak \<Rightarrow> ('f, 'v) rule \<Rightarrow>('f, 'v) rseq  \<Rightarrow> ('f, 'v) rseq  \<Rightarrow> 
  (('f, 'v) rule \<Rightarrow> nat) \<Rightarrow> shows check"
where
  "check_cpeak_eld R p cp j1 j2 lab = 
     (case p of ((s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2)) \<Rightarrow>
      do {
        check (instance_rule (t1, t2) cp) (shows '' critical peak and pair do not match '');
        let u = rseq_last (fst cp) j1;
        let v = rseq_last (snd cp) j2;
        check_rsteps' R j1 (fst cp) u;
        check_rsteps' R j2 (snd cp) v;
        check (u = v) (shows ''the rewrite sequences end in different terms '' +@+ shows u +@+ shows '' and '' +@+ shows v);
        check_ELD_rule_labeling lab r1 r2 (rseq_to_ddseq (t1, j1)) (rseq_to_ddseq (t2, j2))
      })"
  
lemma check_cpeak_eld:
  defines "r \<equiv> ({(n, m) . n < m}, {(n, m) . n \<le> m})"
  assumes "(b, (s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2)) \<in> critical_peaks (set R)" (is "(b, ?p) \<in> critical_peaks (set R)")
  and "isOK (check_cpeak_eld R ((s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2)) cp j1 j2 lab)"
  shows "eld (set R) (rule_labeling lab) r ?p" (is "eld _ ?l _ _")
proof -
  let ?u = "rseq_last (fst cp) j1"
  let ?v = "rseq_last (snd cp) j2"
  let ?j1cp = "rseq_to_ddseq ((fst cp), j1)"
  let ?j2cp = "rseq_to_ddseq ((snd cp), j2)"
  let ?j1t = "rseq_to_ddseq (t1, j1)"
  let ?j2t = "rseq_to_ddseq (t2, j2)"
  from assms have 
    steps1: "isOK (check_rsteps' R j1 (fst cp) ?u)" and 
    steps2: "isOK (check_rsteps' R j2 (snd cp) ?v)"and 
    eq: "?u = ?v" and
    inst: "instance_rule (t1, t2) cp"
    by (auto simp add: Let_def check_cpeak_eld_def)
  from inst obtain \<rho> where \<rho>: "t1 = fst cp \<cdot> \<rho> \<and> t2 = snd cp \<cdot> \<rho>" 
    unfolding instance_rule_def by auto
  let ?j1\<rho> = "seq_close ?j1cp \<box> \<rho>"
  let ?j2\<rho> = "seq_close ?j2cp \<box> \<rho>"
  show ?thesis unfolding eld_def
  proof (intro exI conjI)
    from assms have "isOK (check_ELD_rule_labeling lab r1 r2 ?j1t ?j2t)"
      by (auto simp add: Let_def check_cpeak_eld_def)
    moreover have "map ?l (snd ?j1t) = map ?l (snd ?j1\<rho>)" by (induct j1, auto)
    moreover have "map ?l (snd ?j2t) = map ?l (snd ?j2\<rho>)" by (induct j2, auto)
    ultimately have "isOK (check_ELD_rule_labeling lab r1 r2 ?j1\<rho> ?j2\<rho>)"
      unfolding check_ELD_rule_labeling_def Let_def by auto
    from check_ELD_rule_labeling[OF this]
    show "eld_seq (rule_labeling lab) r ?p ?j1\<rho> ?j2\<rho>" unfolding r_def by auto
    show "ld_trs (set R) ?p ?j1\<rho> ?j2\<rho>"
    proof -
      from assms(2) critical_peak_is_local_peak have lp:"?p \<in> local_peaks (set R)" by fast
      from steps1 check_rsteps'_sound_dd_seq have j1cpseq: "?j1cp \<in> seq (set R)" and 
        j1cplast:"DD2_last ?j1cp = ?u" by auto
      from steps2 check_rsteps'_sound_dd_seq have j2cpseq: "?j2cp \<in> seq (set R)" and 
        j2cplast:"DD2_last ?j2cp = ?v" by auto
      from seq_super_closed1[OF j1cpseq, of "\<box>" \<rho>] \<rho> j1cplast have "?j1\<rho> \<in> seq (set R)" 
        and " get_target (fst ?p) = first ?j1\<rho>" and "DD2_last ?j1\<rho> = ?u \<cdot> \<rho>"
        unfolding get_target_def by auto
      moreover from seq_super_closed1[OF j2cpseq, of "\<box>" \<rho>] \<rho> j2cplast have "?j2\<rho> \<in> seq (set R)" 
        and " get_target (snd ?p) = first ?j2\<rho>" and "DD2_last ?j2\<rho> = ?v \<cdot> \<rho>"
        unfolding get_target_def by auto
      ultimately show ?thesis unfolding ld_trs_def using lp eq by auto
    qed
  qed
qed

lemma linear_imp_SN_rel_d_nd:
  assumes "linear_trs R"
  shows "SN_rel (rstep (R_d R)) (rstep (R_nd R))"
proof -
  from assms have "R_d R = {}" unfolding R_d_def linear_trs_def by auto
  thus ?thesis using SN_rel_empty1 by auto
qed

definition check_rule_labeling_dec ::
  "('f :: {key, show}, string) rules \<Rightarrow>  ('f :: {key, show}, string) rule_lab_repr \<Rightarrow> 
     ('f :: {key, show},string) crit_pair_joins \<Rightarrow> shows check"
where
  "check_rule_labeling_dec R lab js = do {
  let cps = critical_peaks_impl R R;
  let l = rule_lab_repr_to_lab lab;
  let joins = js @ (map (\<lambda> (u, j1, v, j2). (v, j2, u ,j1)) js);
  check_allm (\<lambda>(b, (s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2)). 
    try check (t1 = t2) (shows '' pair non-trivial '') catch
    (\<lambda>e. check_exm (\<lambda>(u, j1, v, j2). 
          (check_cpeak_eld R ((s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2)) (u, v) j1 j2 l))
      joins (\<lambda>es. shows_rule (t1,t2) +@+ shows '' could not be joined decreasingly '')))
    cps
  }"

lemma check_rule_labeling_dec:
  assumes "isOK (check_rule_labeling_dec R lab js)"
  and "SN_rel (rstep (R_d (set R))) (rstep (R_nd (set R)))"
  and "left_linear_trs (set R)"
  shows "CR (rstep (set R))"
using assms (3,2)
proof (rule rule_labeling_is_sound_ll)
  note ok = assms(1)[unfolded check_rule_labeling_dec_def, simplified]
  let ?l = "rule_lab_repr_to_lab lab"
  let ?rl = "rule_labeling ?l"
  let ?joins = "js @ (map (\<lambda> (u, j1, v, j2). (v, j2, u ,j1)) js)"
  def r \<equiv> "({(n, m). n < (m :: nat)}, {(n, m). n \<le> (m :: nat)})"
  show "\<forall>(b, p) \<in> critical_peaks (set R). eld (set R) ?rl r p"
  proof
    fix b p
    assume p:"(b, p) \<in> critical_peaks (set R)"
    hence pi:"(b, p) \<in> set (critical_peaks_impl R R)" using critical_peaks_impl by auto
    obtain s1 r1 p1 \<sigma>1 t1 s2 r2 p2 \<sigma>2 t2 where
      fields: "p = ((s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2))" by (cases p, auto)
    from bspec[OF ok pi]
    have "(t1 = t2) \<or> (isOK (existsM (\<lambda>(u, j1, v, j2). check_cpeak_eld R p (u, v) j1 j2 ?l) ?joins))"
      unfolding fields by simp
    thus "eld (set R) ?rl r p"
    proof
      assume "t1 = t2"
      moreover have "p \<in> local_peaks (set R)" using critical_peak_is_local_peak fields p by fast
      ultimately show ?thesis using trivial_peak_eld unfolding fields by fast
    next
      assume "isOK (existsM (\<lambda>(u, j1, v, j2). check_cpeak_eld R p (u, v) j1 j2 ?l) ?joins)"
      then obtain u j1 v j2 where "(u, j1, v, j2) \<in> set ?joins \<and> isOK(check_cpeak_eld R p (u, v) j1 j2 ?l)"
        unfolding isOK_existsM by (split prod.splits, force)
      thus ?thesis using check_cpeak_eld p unfolding r_def fields by fast
    qed
  qed
qed

end
