theory Usable_Replacement_Map
imports 
  Term_Order
  Q_Relative_Rewriting
  Icap
  Complexity_Framework
begin

subsection {* usable replacement maps and compatibility, taken from Diss Avanzini *}
context 
  fixes \<mu> :: "'f af"
begin
(* deviation: Avanzini uses positions and subterms instead *)
definition af_nf_compatible_terms :: "('f,'v)trs \<Rightarrow> ('f,'v)terms" where
  "af_nf_compatible_terms rel \<equiv> { t . \<forall> C s. t = C\<langle>s\<rangle> \<longrightarrow> af_regarded_pos \<mu> t (hole_pos C) \<or> s \<in> NF rel}"

lemma af_nf_compatible_terms_full_af: assumes full: "\<mu> = full_af"
  shows "s \<in> af_nf_compatible_terms rel"
  unfolding af_nf_compatible_terms_def
  by (auto simp: af_regarded_full full)

lemma af_nf_compatible_terms_mono: assumes rel: "rel' \<subseteq> rel"
  shows "af_nf_compatible_terms rel \<subseteq> af_nf_compatible_terms rel'" 
  unfolding af_nf_compatible_terms_def using NF_anti_mono[OF rel] by blast

definition usable_replacement_map :: "('f,'v)terms \<Rightarrow> bool \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)terms \<Rightarrow> ('f,'v)trs \<Rightarrow> bool" where
  "usable_replacement_map T nfs P Q R \<equiv> 
  (qrstep nfs Q P) ^* `` T \<subseteq> af_nf_compatible_terms (qrstep nfs Q R)"

lemma usable_replacement_map_mono: assumes T: "T' \<subseteq> T" and P: "P' \<subseteq> P" and R: "R' \<subseteq> R"
  shows "usable_replacement_map T nfs P Q R \<Longrightarrow> usable_replacement_map T' nfs P' Q R'"
  unfolding usable_replacement_map_def
  using rtrancl_mono[OF qrstep_mono[OF P subset_refl, of nfs Q]] T
    af_nf_compatible_terms_mono[OF qrstep_mono[OF R subset_refl, of nfs Q]] by blast

lemma avanzini_14_9_1: assumes us: "usable_replacement_map T nfs (S \<union> W) Q R" and "R \<subseteq> S \<union> W"
  and R: "R \<subseteq> ord" and mono: "af_monotone \<pi> ord" and af: "af_subset \<mu> \<pi>" and subst: "subst_closed ord"
  and s: "s \<in> (qrstep nfs Q (S \<union> W))^* `` T" and step: "(s,t) \<in> qrstep nfs Q R" 
  shows "(s,t) \<in> ord"
proof -
  from us[unfolded usable_replacement_map_def] s have 
  s_af: "s \<in> af_nf_compatible_terms (qrstep nfs Q R)" by auto
  from qrstepE[OF step] obtain l r C \<sigma> where 
    nf: "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" and
    lr: "(l, r) \<in> R" and
    s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and
    t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" and
    nfs: "NF_subst nfs (l, r) \<sigma> Q" .
  hence "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> qrstep nfs Q R" (is "(?l,?r) \<in> _") by auto
  with s_af[unfolded af_nf_compatible_terms_def, simplified, rule_format, OF s]
  have rp: "af_regarded_pos \<mu> C\<langle>l \<cdot> \<sigma>\<rangle> (hole_pos C)" by auto
  from lr R have "(l,r) \<in> ord" by auto
  from subst_closedE[OF subst this] have lr: "(?l,?r) \<in> ord" by auto
  have "(C\<langle>?l\<rangle>, C\<langle>?r\<rangle>) \<in> ord"
    by (rule af_monotone_af_regarded_posD[OF af_subset_af_monotone[OF af mono] rp lr])
  with s t show ?thesis by auto
qed
end

lemma usable_replacement_map_full_af[simp]: "usable_replacement_map full_af T nfs R Q R'"
  unfolding usable_replacement_map_def using af_nf_compatible_terms_full_af[OF refl] by auto


(* deviation: currently we do not support separate af for weak order! 
  reason: the reduction pair requirement already demands full monotonicity *)
context cpx_ce_af_redtriple_order
begin
definition NS_\<mu> :: "'f af" where "NS_\<mu> = full_af"
lemma NS_\<mu>: "af_monotone NS_\<mu> NS" 
  by (rule ctxt_closed_imp_af_monotone[OF wm_NS])

lemma usable_replacement_map_NS: "usable_replacement_map NS_\<mu> T nfs P Q R"
  unfolding usable_replacement_map_def
  using af_nf_compatible_terms_full_af[OF NS_\<mu>_def] by blast


lemma avanzini_14_9_2: assumes compatR: "Rs \<subseteq> S" "Rw \<subseteq> NS" and
  usable: "usable_replacement_map \<mu>' T nfs (Rs \<union> Rw) Q Rs"
  and af: "af_subset \<mu>' \<mu>"
  and step: "(s,t) \<in> relto (qrstep nfs Q Rs) (qrstep nfs Q Rw)"
  and s: "s \<in> (qrstep nfs Q (Rs \<union> Rw))^* `` T" (is "s \<in> ?T")
  shows "(s,t) \<in> S"
proof -
  have subset: "Rs \<subseteq> Rs \<union> Rw" "Rw \<subseteq> Rs \<union> Rw" by auto
  note strict = avanzini_14_9_1[OF usable subset(1) compatR(1) \<mu> af subst_S]
  note weak = avanzini_14_9_1[OF usable_replacement_map_NS subset(2) compatR(2) NS_\<mu> af_subset_refl subst_NS, of _ nfs Q T]
  let ?Rs = "qrstep nfs Q Rs" let ?Rw = "qrstep nfs Q Rw" let ?P = "qrstep nfs Q (Rs \<union> Rw)"
  from step obtain u v where su: "(s,u) \<in> ?Rw^*" and uv: "(u,v) \<in> ?Rs" and vt: "(v,t) \<in> ?Rw^*" by auto
  have [simp]: "\<And> s. (s,s) \<in> NS" using refl_NS[unfolded refl_on_def] by auto
  {
    fix s t R
    assume s: "s \<in> ?T" and R: "(s,t) \<in> qrstep nfs Q R"
      "R \<subseteq> Rs \<union> Rw"
    from qrstep_mono[OF R(2)] R(1) have "(s,t) \<in> ?P" by blast
    with s have "t \<in> ?T"
      by (metis r_into_rtrancl rtrancl_Image_step) 
  } note T = this
  {
    fix s u
    assume s: "s \<in> ?T" and su: "(s,u) \<in> ?Rw^*"
    from su have "(s,u) \<in> NS \<and> u \<in> ?T"
    proof (induct)
      case (step t u)
      from weak[OF _ step(2)] step(3) have "(t,u) \<in> NS" by auto
      from trans_NS_point[OF _ this] step(3) have su: "(s,u) \<in> NS" by auto
      from T[OF _ step(2)] step(3) have "u \<in> ?T" by auto
      with su show ?case by auto
    qed (insert s, auto)
    hence "(s,u) \<in> NS" "u \<in> ?T" by auto
  } note steps = this
  from steps[OF s su] have su: "(s,u) \<in> NS" and u: "u \<in> ?T" .
  from T[OF u uv] have v: "v \<in> ?T" by auto
  from strict[OF u uv] have uv: "(u,v) \<in> S" .
  from steps[OF v vt] have vt: "(v,t) \<in> NS" by auto
  from compat_NS_S_point[OF su uv] have sv: "(s,v) \<in> S" .
  from compat_S_NS_point[OF sv vt] show "(s,t) \<in> S" .
qed

theorem avanzini_14_10: assumes compatR: "Rs \<subseteq> S" "Rw \<subseteq> NS" and
  usable: "usable_replacement_map \<mu>' (terms_of cm) nfs (Rs \<union> Rw) Q Rs" and
  af: "af_subset \<mu>' \<mu>" and
  cpx: "cpx_class cm = Inr cc"
shows "deriv_bound_measure_class (relto (qrstep nfs Q Rs) (qrstep nfs Q Rw)) cm cc"
  (is "deriv_bound_measure_class ?P cm cc")
proof -
  let ?T = "(qrstep nfs Q (Rs \<union> Rw))^* `` terms_of cm"
  from cpx_class[OF cpx] have bound: "deriv_bound_measure_class S cm cc" .
  note d = deriv_bound_measure_class_def deriv_bound_rel_class_def deriv_bound_rel_def
  from bound[unfolded d] obtain f where f: "f \<in> O_of cc" 
    and bound: "\<And> n t. t \<in> terms_of_nat cm n \<Longrightarrow> deriv_bound S t (f n)"
    by auto
  show ?thesis unfolding d
  proof (intro exI conjI allI impI, rule f)
    fix n t
    assume t: "t \<in> terms_of_nat cm n"
    from bound[OF t] have bound: "deriv_bound S t (f n)" by auto
    from t have t: "t \<in> ?T" by (auto simp: terms_of)
    note d = deriv_bound_def
    note main = avanzini_14_9_2[OF compatR usable af]
    show "deriv_bound ?P t (f n)" 
    proof (rule ccontr)
      assume "\<not> ?thesis"
      from this[unfolded d]
      obtain s where ts: "(t,s) \<in> ?P ^^ (Suc (f n))" by auto
      def m \<equiv> "Suc (f n)"
      from ts have "(t,s) \<in> S ^^ (Suc (f n)) \<and> s \<in> ?T" unfolding m_def[symmetric]
      proof (induct m arbitrary: s)
        case (Suc m s)
        from Suc(2)[simplified] obtain u where tus: "(t,u) \<in> ?P^^m" and uss: "(u,s) \<in> ?P" by auto
        from Suc(1)[OF tus] have tu: "(t,u) \<in> S^^m" and u: "u \<in> ?T" by auto
        from main[OF uss u] have us: "(u,s) \<in> S" .
        from tu us have ts: "(t,s) \<in> S^^(Suc m)" by auto
        from uss have "(u,s) \<in> (qrstep nfs Q (Rs \<union> Rw))^*" unfolding qrstep_union by regexp
        with u have "s \<in> ?T" by (metis rtrancl_Image_step) 
        with ts show ?case by auto
      qed (insert t, auto)
      with bound show False unfolding d by blast
    qed
  qed
qed

end

subsection {* computing innermost usable replacement maps from Moser/Hirokawa *}

context
  fixes ecap :: "('f,string)cap_fun"
  and R :: "('f,string)trs"
  and Q :: "('f,string)terms"
begin

definition innermost_repl_map :: "(('f \<times> nat) \<times> nat) set" where
  "innermost_repl_map \<equiv> { ((f,length rs),i) | l C f rs i. 
    (l,C\<langle>Fun f rs\<rangle>) \<in> R \<and>
    i < length rs \<and>
    Inl () \<in> vars_term (ecap R Q (mv_xvar ` (set (args l))) (mv_xvar (rs ! i)))}"

definition \<mu>_i :: "'f af" where
  "\<mu>_i f = { i. (f,i) \<in> innermost_repl_map}"

context 
  assumes inn: "NF_terms Q \<subseteq> NF_trs R"
  and ecap: "is_ecap ecap"
  and wf: "wf_trs R"
begin

lemma hirokawa_moser_4_7_1_main:   
  assumes redex: "(redex,r') \<in> rqrstep nfs Q R"
  and lr: "(l,r) \<in> R"
  and nfs: "NF_subst True (l,r) \<sigma> Q"
  and args: "set (args l) \<cdot>set \<sigma> \<subseteq> NF_terms Q"
  shows "r \<cdot> \<sigma> \<unrhd> C \<langle> redex \<rangle> \<Longrightarrow> af_regarded_pos \<mu>_i (C \<langle> redex \<rangle>) (hole_pos C)"
proof (induct C)
  case (More f bef C aft)
  from supteq_trans[OF More(2)] have "r \<cdot> \<sigma> \<unrhd> C\<langle>redex\<rangle>" by auto
  from More(1)[OF this] have IH: "af_regarded_pos \<mu>_i C\<langle>redex\<rangle> (hole_pos C)" .
  from More(2) obtain D where id: "r \<cdot> \<sigma> = D \<langle> (More f bef C aft) \<langle> redex \<rangle> \<rangle>" 
    by (metis supteq_ctxtE)
  let ?hp = "hole_pos D" 
  let ?i = "length bef"
  let ?n = "Suc (?i + length aft)"
  from redex have "(redex,r') \<in> rstep R" unfolding rqrstep_def by auto
  hence redex_NF: "redex \<notin> NF_trs R" by auto
  {
    fix x
    assume "\<sigma> x \<in> NF_terms Q"
    with inn have "\<sigma> x \<in> NF_trs R" by auto
    from NF_subterm[OF this, of redex] redex_NF have "\<not> \<sigma> x \<unrhd> redex" by auto
  } note vars = this
  let ?mv = "mv_xvar"
  from id have "?hp \<in> poss (r \<cdot> \<sigma>)" by auto
  from poss_subst_choice[OF this]
  have "?i \<in> \<mu>_i (f,?n)" 
  proof
    assume in_r: "?hp \<in> poss r \<and> is_Fun (r |_ ?hp)"
    hence [simp]: "r \<cdot> \<sigma> |_ ?hp = r |_ ?hp \<cdot> \<sigma>" by auto
    from arg_cong[OF id, of "\<lambda> t. t |_ ?hp"]
    have id: "r |_ ?hp \<cdot> \<sigma> = Fun f (bef @ C\<langle>redex\<rangle> # aft)" by simp
    with in_r obtain rs where rhp: "r |_ ?hp = Fun f rs" by (cases "r |_ ?hp", auto)
    def E \<equiv> "ctxt_of_pos_term ?hp r"
    from in_r have "r = E \<langle> r |_ ?hp \<rangle>" unfolding E_def
      by (metis ctxt_supt_id)
    with rhp have r: "r = E \<langle> Fun f rs \<rangle>" by auto
    note lr = lr[unfolded r]
    show ?thesis unfolding \<mu>_i_def
    proof
      let ?m = "length rs"
      note id = id[unfolded rhp]
      from arg_cong[OF id, of num_args] 
      have sym: "((f,?n),?i) = ((f,?m),?i)" and im: "?i < ?m" by auto
      show "((f,?n),?i) \<in> innermost_repl_map" unfolding innermost_repl_map_def
      proof (rule, intro exI conjI, rule sym, rule lr, rule im)
        def ri \<equiv> "rs ! ?i"
        from arg_cong[OF id, of "\<lambda> t. args t ! ?i"] sym
        have ri_red: "ri \<cdot> \<sigma> = C\<langle>redex\<rangle>" unfolding ri_def by auto
        from r have "r \<unrhd> Fun f rs" by auto
        moreover have "Fun f rs \<unrhd> ri" using sym ri_def by auto
        ultimately have "r \<unrhd> ri" by (rule supteq_trans)
        hence "vars_term ri \<subseteq> vars_term r" by (rule supteq_imp_vars_subset)
        with nfs[unfolded NF_subst_def vars_rule_def] 
        have "\<sigma> ` vars_term ri \<subseteq> NF_terms Q" by auto
        with ri_red
        show "Inl () \<in> vars_term (ecap R Q (?mv ` set (args l)) (?mv (rs ! ?i)))"
          unfolding ri_def[symmetric]
        proof (induct ri arbitrary: C)
          case (Var x C)
          hence "\<sigma> x \<unrhd> redex" and "\<sigma> x \<in> NF_terms Q" by auto
          with vars[of x] show ?case by auto
        next
          case (Fun f rs C)
          show ?case
          proof (cases C)
            case Hole
            have id: "set (args l) \<cdot>set \<sigma> = ?mv ` set (args l) \<cdot>set mv_subst \<sigma>"
              using mv_xvar[of _ \<sigma>] by auto
            from Hole Fun have "Fun f rs \<cdot> \<sigma> = redex" by simp            
            with redex have "(Fun f rs \<cdot> \<sigma>, r') \<in> (qrstep nfs Q R)\<^sup>* O rqrstep nfs Q R" by auto
            also have "Fun f rs \<cdot> \<sigma> = Fun f (map ?mv rs) \<cdot> mv_subst \<sigma>" 
              unfolding mv_xvar[of _ \<sigma>] by simp
            finally have "cap R Q (?mv ` set (args l)) ((Fun f (map ?mv rs))) = Var (Inl ())" 
              using args[unfolded id]
              by force
            from cap_Fun_fresh[OF ecap this]
            show ?thesis by auto
          next
            case (More g bef D aft)
            let ?i = "length bef"
            note id = Fun(2)[unfolded More]
            from arg_cong[OF id, of num_args] have len: "length rs = Suc (?i + length aft)" by simp
            with arg_cong[OF id, of "\<lambda> t. args t ! ?i"] have id: "rs ! ?i \<cdot> \<sigma> = D\<langle>redex\<rangle>" by auto
            from len have mem: "rs ! ?i \<in> set rs" by auto
            from Fun(1)[OF mem id] Fun(3) mem 
            have IH: "Inl () \<in> vars_term (ecap R Q (?mv ` set (args l)) (?mv (rs ! ?i)))" by auto
            with ecap_Fun[OF ecap, of R Q "?mv ` set (args l)" f "map ?mv rs"]
            show ?thesis using mem by auto
          qed
        qed
      qed
    qed
  next
    assume "\<exists> x q1 q2. q1 \<in> poss r \<and> q2 \<in> poss (\<sigma> x) \<and>
     r |_ q1 = Var x \<and> x \<in> vars_term r \<and> ?hp = q1 <#> q2 \<and> r \<cdot> \<sigma> |_ ?hp = \<sigma> x |_ q2"
    then obtain x q2 where x: "x \<in> vars_term r" and idd: "r \<cdot> \<sigma> |_ ?hp = \<sigma> x |_ q2"
      and q2: "q2 \<in> poss (\<sigma> x)" by auto
    from idd[unfolded id] have "(More f bef C aft)\<langle>redex\<rangle> = \<sigma> x |_ q2" by simp
    with q2 have "\<sigma> x \<unrhd> redex" 
      by (metis ctxt_imp_supteq subt_at_imp_supteq subterm.le_imp_less_or_eq subterm.order.strict_trans2 supt_imp_supteq)
    with vars[of x] have "\<sigma> x \<notin> NF_terms Q" by auto
    with nfs x have False unfolding NF_subst_def vars_rule_def by auto
    thus ?thesis by auto
  qed
  thus ?case using IH by simp
qed simp

lemma hirokawa_moser_4_7_1: 
  assumes s_af: "s \<in> af_nf_compatible_terms \<mu>_i (qrstep nfs Q R)"
  and step: "(s,t) \<in> qrstep nfs Q R"
  shows "t \<in> af_nf_compatible_terms \<mu>_i (qrstep nfs Q R)"
proof -
  note d = af_nf_compatible_terms_def
  let ?NF = "NF (qrstep nfs Q R)"
  note wwf = wf_trs_imp_wwf_qtrs[OF wf, of Q]
  show ?thesis unfolding d
  proof (rule, intro allI impI)
    fix D u
    assume tD: "t = D\<langle>u\<rangle>"
    from qrstepE[OF step[unfolded wwf_qtrs_imp_nfs_switch[OF wwf, of nfs True]]]
      obtain C \<sigma> l r where 
      NF: "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" and
      lr: "(l, r) \<in> R" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t: "t = C\<langle>r \<cdot> \<sigma>\<rangle>" 
      and nfs: "NF_subst True (l, r) \<sigma> Q" .
    from nfs have nfs_F: "NF_subst nfs (l,r) \<sigma> Q" unfolding NF_subst_def by auto
    from NF lr nfs_F have rstep: "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> qrstep nfs Q R" by auto
    from NF[folded NF_terms_args_conv]
    have "set (args l) \<cdot>set \<sigma> \<subseteq> NF_terms Q" by (cases l, auto)
    note main = hirokawa_moser_4_7_1_main[OF _ lr nfs this]
    let ?l = "l \<cdot> \<sigma>" let ?r = "r \<cdot> \<sigma>"
    show "af_regarded_pos \<mu>_i t (hole_pos D) \<or> u \<in> ?NF"
    proof (cases "u \<in> ?NF")
      case False note u = this
      from s_af[unfolded d] have s_af: "\<And> D u. s = D\<langle>u\<rangle> \<Longrightarrow> af_regarded_pos \<mu>_i s (hole_pos D) \<or> u \<in> ?NF" by auto
      from s_af[OF s] rstep have reg1: "af_regarded_pos \<mu>_i s (hole_pos C)" by auto
      from s_af[of _ u] False have reg2: "\<And> D. s = D\<langle>u\<rangle> \<Longrightarrow> af_regarded_pos \<mu>_i s (hole_pos D)" by auto
      from reg1 reg2 have "af_regarded_pos \<mu>_i t (hole_pos D)" using tD unfolding t s
      proof (induct C arbitrary: D)
        case (More f bef C aft D) note C = this
        let ?i = "length bef"
        let ?n = "Suc (length bef + length aft)"
        from C(2) have i: "?i \<in> \<mu>_i (f,?n)" and reg: "af_regarded_pos \<mu>_i (C \<langle>?l\<rangle>) (hole_pos C)" by auto
        show ?case
        proof (cases D)
          case (More g bbef DD aaft) note D = this
          let ?i' = "length bbef"
          show ?thesis
          proof (cases "?i' = ?i")
            case True
            with C(4)[unfolded D] have id: "C\<langle>r \<cdot> \<sigma>\<rangle> = DD\<langle>u\<rangle>" by auto
            have "af_regarded_pos \<mu>_i C\<langle>r \<cdot> \<sigma>\<rangle> (hole_pos DD)"
            proof (rule C(1)[OF reg _ id])
              fix DD
              assume "C\<langle>l \<cdot> \<sigma>\<rangle> = DD\<langle>u\<rangle>"
              with C(3)[of "More f bef DD aft"] reg
              show "af_regarded_pos local.\<mu>_i C\<langle>l \<cdot> \<sigma>\<rangle> (hole_pos DD)" by auto
            qed
            thus ?thesis using True i unfolding D by simp
          next
            case False note neq = this
            from C(4)[unfolded D]
            have id: "bef @ C\<langle>r \<cdot> \<sigma>\<rangle> # aft = bbef @ DD\<langle>u\<rangle> # aaft" by simp
            from arg_cong[OF id, of length] have i': "?i' < ?n" by simp
            from arg_cong[OF id, of "\<lambda> ts. ts ! ?i'"]
            have idd: "DD\<langle>u\<rangle> = (bef @ C\<langle>?r\<rangle> # aft) ! ?i'" by simp
            show ?thesis
            proof (cases "?i < ?i'")
              case False
              with neq have i': "?i' < ?i" by auto
              hence bef: "bef ! ?i' = DD\<langle>u\<rangle>" unfolding idd by (auto simp: nth_append)
              from id_take_nth_drop[OF i', unfolded bef] i' obtain b1 b2 where
                bef: "bef = b1 @ DD\<langle>u\<rangle> # b2" and len: "length b1 = ?i'" by auto
              from C(3)[unfolded bef, of "More f b1 DD (b2 @ C\<langle>l \<cdot> \<sigma>\<rangle> # aft)"]
              have "length b1 \<in> \<mu>_i (f, Suc (Suc (length b1 + (length b2 + length aft))))"
                and "af_regarded_pos \<mu>_i DD\<langle>u\<rangle> (hole_pos DD)" by auto
              thus ?thesis unfolding bef D using len
                by (simp add: nth_append)
            next
              case True
              def i'' \<equiv> "?i' - Suc ?i"
              from True i''_def i' have i': "?i' = Suc ?i + i''" and i'': "i'' < length aft" by auto
              have aft: "aft ! i'' = DD\<langle>u\<rangle>" unfolding idd i' by (auto simp: nth_append)
              from id_take_nth_drop[OF i'', unfolded aft] i'' obtain a1 a2 where
                aft: "aft = a1 @ DD\<langle>u\<rangle> # a2" and len: "length a1 = i''" by auto
              from C(3)[unfolded aft, of "More f (bef @ C\<langle>l \<cdot> \<sigma>\<rangle> # a1) DD a2"]
              have " Suc (?i + length a1) \<in> \<mu>_i (f, Suc (Suc (?i + (length a1 + length a2))))"
              and "af_regarded_pos \<mu>_i ((bef @ C\<langle>l \<cdot> \<sigma>\<rangle> # a1 @ DD\<langle>u\<rangle> # a2) ! Suc (?i + length a1)) (hole_pos DD)" 
                by auto
              thus ?thesis unfolding aft D using len
                by (simp add: nth_append i')
            qed
          qed
        qed simp
      next
        case (Hole D)
        let ?hp = "hole_pos D"
        from Hole have idd: "r \<cdot> \<sigma> = D\<langle>u\<rangle>" by simp
        from False obtain v where "(u,v) \<in> qrstep nfs Q R" unfolding NF_def by auto
        {
          from qrstepE[OF this]
            obtain E \<sigma>' l' r' where 
            NF: "\<forall>u\<lhd>l' \<cdot> \<sigma>'. u \<in> NF_terms Q" and
            lr: "(l', r') \<in> R" and u: "u = E\<langle>l' \<cdot> \<sigma>'\<rangle>" and v: "v = E\<langle>r' \<cdot> \<sigma>'\<rangle>" 
            and nfs: "NF_subst nfs (l', r') \<sigma>' Q" .
          hence "(l' \<cdot> \<sigma>', r' \<cdot> \<sigma>') \<in> rqrstep nfs Q R" by auto
          with u have "\<exists> E red r'. u = E\<langle>red\<rangle> \<and> (red,r') \<in> rqrstep nfs Q R" by auto          
        }
        then obtain E red r' where u: "u = E\<langle>red\<rangle>" and red: "(red,r') \<in> rqrstep nfs Q R" by auto
        def rr \<equiv> "r \<cdot> \<sigma>"
        from main[OF red, of "D \<circ>c E"] idd[unfolded u]
        have "af_regarded_pos \<mu>_i (r \<cdot> \<sigma>) (hole_pos (D \<circ>c E))" by auto
        thus ?case unfolding ctxt_apply.simps rr_def[symmetric]
          by (simp add: af_regarded_pos_append)
      qed
      thus ?thesis by simp
    qed simp
  qed
qed

theorem hirokawa_moser_4_8_1:
  fixes T :: "('f,string)terms" and C D :: "('f \<times> nat)list"
  defines "T \<equiv> terms_of (Runtime_Complexity C D)"
  assumes C: "\<And> c. c \<in> set C \<Longrightarrow> \<not> defined R c"
  shows "usable_replacement_map \<mu>_i T nfs R Q R"
proof (unfold usable_replacement_map_def, rule)
  fix t
  assume "t \<in> (qrstep nfs Q R)\<^sup>* `` T"
  then obtain s where steps: "(s,t) \<in> (qrstep nfs Q R)\<^sup>*" and s: "s \<in> T" by blast
  from steps 
  show "t \<in> af_nf_compatible_terms \<mu>_i (qrstep nfs Q R)"
  proof (induct)
    case (step t u)
    from hirokawa_moser_4_7_1[OF step(3) step(2)] show ?case .
  next
    case base
    from s[unfolded T_def] obtain f cs where
    s: "s = Fun f cs" and f: "(f,length cs) \<in> set D" and 
    cs: "\<And> c. c \<in> set cs \<Longrightarrow> funas_term c \<subseteq> set C" by (force simp: funas_args_term_def)
    show ?case unfolding s af_nf_compatible_terms_def
    proof (rule, intro allI impI)
      fix D u
      assume id: "Fun f cs = D\<langle>u\<rangle>"
      show "af_regarded_pos \<mu>_i (Fun f cs) (hole_pos D) \<or> u \<in> NF (qrstep nfs Q R)"
      proof (cases D)
        case (More g bef E aft)
        with id have "E\<langle>u\<rangle> \<in> set cs" by auto
        from cs[OF this] have u: "funas_term u \<subseteq> set C" by auto
        {
          fix v
          assume "(u,v) \<in> qrstep nfs Q R"
          from qrstepE[OF this]
            obtain F l r \<sigma> where uu: "u = F\<langle>l\<cdot>\<sigma>\<rangle>" and lr: "(l,r) \<in> R" .
          with wf[unfolded wf_trs_def] obtain f ls where l: "l = Fun f ls" by blast
          let ?n = "length ls"
          from u[unfolded uu l] have "(f,?n) \<in> set C" by auto
          from C[OF this] lr[unfolded l] have False unfolding defined_def by auto
        }
        hence "u \<in> NF (qrstep nfs Q R)" by auto
        thus ?thesis by auto
      qed simp
    qed
  qed
qed
end
end

lemma avanzini_14_34: assumes us: "usable_replacement_map \<mu> (terms_of cm) nfs (RS \<union> R) Q RS'"
  and DP: "is_DP_complexity Cp FS F RS R cm"
  and RS': "RS' \<subseteq> RS"
  and \<mu>_com: "\<And> f. f \<in> Cp \<Longrightarrow> \<mu>_com f \<supseteq> \<mu> f"
  shows "usable_replacement_map \<mu>_com (terms_of cm) nfs (RS \<union> R) Q RS'"
  unfolding usable_replacement_map_def
proof
  fix t
  assume tt: "t \<in> (qrstep nfs Q (RS \<union> R))\<^sup>* `` terms_of cm"
  with avanzini_14_20(2)[OF DP] have t: "t \<in> Tsharp_terms Cp FS F" by auto
  from tt us[unfolded usable_replacement_map_def] 
  have tt: "t \<in> af_nf_compatible_terms \<mu> (qrstep nfs Q RS')" by auto
  note d = af_nf_compatible_terms_def
  show "t \<in> af_nf_compatible_terms \<mu>_com (qrstep nfs Q RS')" unfolding d
  proof (rule, intro allI impI)
    fix C s
    assume ts: "t = C\<langle>s\<rangle>"
    with tt[unfolded d] have af: "af_regarded_pos \<mu> t (hole_pos C) \<or> s \<in> NF (qrstep nfs Q RS')" by auto
    show "af_regarded_pos \<mu>_com t (hole_pos C) \<or> s \<in> NF (qrstep nfs Q RS')"
    proof (cases "s \<in> NF (qrstep nfs Q RS')")
      case False      
      with af have af: "af_regarded_pos \<mu> t (hole_pos C)" by auto
      from False obtain u where "(s,u) \<in> qrstep nfs Q RS'" unfolding NF_def by auto
      {
        from qrstepE[OF this] obtain l r C \<sigma> where lr: "(l,r) \<in> RS'" and s: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" .
        from lr RS' have lr: "(l,r) \<in> RS" by auto
        from DP have "wf_trs RS" and RS: "lhss RS \<subseteq> Fsharp_terms FS F" by (cases, auto)+
        from wf_trs_imp_lhs_Fun[OF this(1) lr] obtain f ls where l: "l = Fun f ls" by auto
        from RS lr have "l \<in> Fsharp_terms FS F" by auto
        from this[unfolded l] have "(f,length ls) \<in> FS" by (cases, auto)
        with s l have "funas_term s \<inter> FS \<noteq> {}" by simp
      } note s_FS = this
      from DP have FS_F: "FS \<inter> F = {}" by (cases, auto)
      with s_FS have sF: "\<not> funas_term s \<subseteq> F" by auto
      from t ts af have "af_regarded_pos \<mu>_com t (hole_pos C)"
      proof (induct t arbitrary: C)
        case (compound f ts C)
        show ?case
        proof (cases C)
          case (More g bef D aft) note C = this
          let ?i = "length bef"
          let ?n = "length ts"
          from compound(1) have "(f,?n) \<in> Cp" by simp
          note \<mu>_com = \<mu>_com[OF this]
          from compound(5)[unfolded C]
          have i[simp]: "?i < ?n"
            and mu: "?i \<in> \<mu> (f, ?n)"
            and af: "af_regarded_pos \<mu> (ts ! ?i) (hole_pos D)" by auto
          from mu \<mu>_com have [simp]: "?i \<in> \<mu>_com (f,?n)" by auto
          have [simp]: "af_regarded_pos \<mu>_com (ts ! ?i) (hole_pos D)"
            by (rule compound(3)[OF _ _ af], insert i compound(4) C, auto)
          show ?thesis unfolding C by simp
        qed simp
      next
        case (base t C)
        with sF have False by auto
        thus ?case by simp
      next
        case (sharp t C)
        show ?case
        proof (cases C)
          case (More g bef D aft)
          from sharp(1)[unfolded sharp(2) More]
          have "funas_term s \<subseteq> F" by (cases, force)
          with sF have False by simp
          thus ?thesis by simp
        qed simp
      qed
      thus ?thesis ..
    qed simp
  qed
qed

end

