theory Usable_Replacement_Map_Impl
imports 
  Usable_Replacement_Map
  Icap_Impl
  QDP_Framework_Impl
  Term_Order_Impl
  Map_Choice
  Complexity_Framework_Impl
begin

definition "full_empty fs = (let fs' = filter (\<lambda> (f,n). n \<noteq> 0) fs in (fs', \<lambda> f. if f \<in> set fs' then full_af f else {}, ''full AF''))"

locale usable_replacement_map =
  fixes R :: "('f :: {show,key},string)rules"
  and Q :: "('f,string)term list"
  and ecap :: "('f,string)term list \<Rightarrow> ('f, string) term \<Rightarrow> ('f, unit + string) term"
begin

definition innermost_repl_map_impl where 
  "innermost_repl_map_impl \<equiv> remdups [(f,i) . (l,r) <- R, u <- supteq_list r, is_Fun u, rs <- [args u], f <- [the (root u)],
  n <- [snd f], i <- [0..< n], Inl () \<in> vars_term (ecap (args l) (rs ! i)) ]"

definition \<mu>_i_impl :: "(('f \<times> nat)list \<times> 'f af \<times> string)" where 
  "\<mu>_i_impl \<equiv> let fis = innermost_repl_map_impl;
    fs = remdups (map fst fis);
    \<mu> = (\<lambda> f. set (map snd (filter (\<lambda> (g,i). g = f) fis)))
    in (fs,precompute_fun \<mu> fs, ''innermost URM'')"

definition "default_fs \<equiv> funas_trs_list R"

lemma full_empty: assumes fe: "full_empty fs = (fs',\<mu>,info)"
  and defa: "set default_fs \<subseteq> set fs"
  and sig: "set (get_signature_of_cm cm) \<subseteq> set fs"
  and wf: "wf_trs (set R)"
  shows "\<And> f. f \<notin> set fs' \<Longrightarrow> \<mu> f = {}"
  and "usable_replacement_map \<mu> (terms_of cm) nfs (set R) (set Q) (set R)"
proof -
  let ?fs = "filter (\<lambda> (f,n). n \<noteq> 0) fs"
  from fe[unfolded full_empty_def, simplified]
  have \<mu>: "\<mu> = (\<lambda>f. if f \<in> set ?fs then full_af f else {})" and fs: "fs' = ?fs" by auto
  show "\<And> f. f \<notin> set fs' \<Longrightarrow> \<mu> f = {}" unfolding fs \<mu> by auto
  let ?T = "terms_of cm"
  let ?R = "qrstep nfs (set Q) (set R)"
  from defa[unfolded default_fs_def] have R: "funas_trs (set R) \<subseteq> set fs" by auto
  show "usable_replacement_map \<mu> ?T nfs (set R) (set Q) (set R)"
    unfolding usable_replacement_map_def
  proof
    fix t
    assume "t \<in> ?R^* `` ?T"
    then obtain u where steps: "(u,t) \<in> ?R^*" and u: "u \<in> ?T" by auto
    then obtain n where "u \<in> terms_of_nat cm n" by (auto simp: terms_of)
    from set_mp[OF get_signature_of_cm this] 
    have "funas_term u \<subseteq> set (get_signature_of_cm cm)" by auto
    with sig have "funas_term u \<subseteq> set fs" by auto
    from rsteps_preserve_funas_terms[OF R this set_mp[OF rtrancl_mono steps] wf]
    have t: "funas_term t \<subseteq> set fs" by auto
    {
      fix C s
      assume id: "t = C\<langle>s\<rangle>"
      with t have "funas_ctxt C \<subseteq> set fs" by auto
      hence "af_regarded_pos \<mu> t (hole_pos C)" unfolding id
      proof (induct C)
        case (More f bef C aft)
        let ?i = "length bef"
        let ?n = "Suc (?i + length aft)"
        from More
        have IH: "af_regarded_pos \<mu> C\<langle>s\<rangle> (hole_pos C)" and 
          f: "(f,?n) \<in> set fs" by auto
        from f have "?i \<in> \<mu> (f,?n)" unfolding \<mu> full_af_def by auto
        with IH show ?case by auto
      qed simp
    }
    thus  "t \<in> af_nf_compatible_terms \<mu> ?R" unfolding af_nf_compatible_terms_def by auto
  qed
qed

fun get_fs_\<mu> :: "bool \<Rightarrow> ('f,string)complexity_measure \<Rightarrow> (('f \<times> nat)list \<times> 'f af \<times> string)" where
  "get_fs_\<mu> inn (Derivational_Complexity F) = (full_empty (remdups (F @ default_fs)))"
| "get_fs_\<mu> inn (Runtime_Complexity C D) = 
    (if inn \<and> set C \<inter> set (defined_list R) \<subseteq> {} then \<mu>_i_impl 
      else full_empty (remdups (C @ D @ default_fs))
    )"

definition get_fs_\<mu>_DP :: "bool \<Rightarrow> ('f,string)rules \<Rightarrow> ('f,string)complexity_measure \<Rightarrow> (('f \<times> nat)list \<times> 'f af \<times> string)" where
  "get_fs_\<mu>_DP inn S cm = (
   let (fs,\<mu>,info) = get_fs_\<mu> inn cm   
   in (case check_DP_complexity R cm of 
     Inr (RS, R', Cp, FS, F) \<Rightarrow> if set S \<subseteq> set RS then (list_inter fs Cp, \<lambda> f. if f \<in> set Cp then \<mu> f else {}, info @ '' with DPs'') else (fs,\<mu>,info)
     | _ \<Rightarrow> (fs,\<mu>,info)))"

context
  assumes ecap: "ecap = icap_impl' R Q"
begin

lemma innermost_repl_map_impl[simp]: "set innermost_repl_map_impl = innermost_repl_map icap' (set R) (set Q)"
  (is "?l = ?r")
proof -
  note d = innermost_repl_map_impl_def
  have "?r = { ((f,length rs),i) | l C f rs i. (l,C\<langle>Fun f rs\<rangle>) \<in> (set R) \<and>
    i < length rs \<and> Inl () \<in> vars_term (ecap (args l) (rs ! i))}" (is "_ = ?m")    
    unfolding innermost_repl_map_def
    unfolding icap' icap_mv_def[symmetric] icap_impl'_sound[symmetric] ecap[symmetric]
    by (rule refl)
  also have "?m = ?l" 
  proof -
    {
      fix f n i
      assume "((f,n),i) \<in> ?m"
      then obtain l C rs where lr: "(l,C\<langle>Fun f rs\<rangle>) \<in> (set R)"
       and *: "i < length rs \<and> Inl () \<in> vars_term (ecap (args l) (rs ! i))" 
       and n: "n = length rs" by auto
      have "((f,n),i) \<in> ?l" unfolding d n
        by (clarsimp, rule bexI[OF _ lr], rule exI, rule exI, rule conjI[OF refl],
        rule bexI[of _ "Fun f rs"], insert *, auto)
    }
    hence one: "?m \<subseteq> ?l" by force
    {
      fix fn i
      assume "(fn,i) \<in> ?l"
      from this[unfolded d]
      obtain l r u where 
        lr: "(l,r) \<in>set R"
        and supt: "r \<unrhd> u" and u: "is_Fun u"
        and f: "fn = the (root u)"
        and i: "i < snd (the (root u))"
        and cap: "Inl () \<in> vars_term (ecap (args l) (args u ! i))" by auto
      from supt obtain C where r: "r = C \<langle> u \<rangle>" by auto
      from u obtain f rs where u: "u = Fun f rs" by auto
      have "(fn,i) \<in> ?m" using lr i cap unfolding u r f by auto
    } 
    hence "?l \<subseteq> ?m" by auto
    with one show ?thesis by blast
  qed
  finally show ?thesis by simp
qed

lemma \<mu>_i_impl: assumes impl: "\<mu>_i_impl = (fs,\<mu>,info)"
  shows "\<mu> = \<mu>_i icap' (set R) (set Q) \<and> (\<forall> f. f \<notin> set fs \<longrightarrow> \<mu> f = {})"
proof -
  note d = \<mu>_i_impl_def Let_def
  show ?thesis
  proof (intro conjI allI impI)
    fix f
    assume "f \<notin> set fs"
    thus "\<mu> f = {}"
    using impl by (auto simp: d)
  next
    show "\<mu> = \<mu>_i icap' (set R) (set Q)" using impl unfolding d
      by (intro ext, auto simp: \<mu>_i_def)
  qed
qed  

lemma get_fs_\<mu>: assumes split: "get_fs_\<mu> inn cm = (fs,\<mu>,info)"
  and wf: "wf_trs (set R)"
  and inn: "inn = (NF_terms (set Q) \<subseteq> NF_trs (set R))"
  shows "\<And> f. f \<notin> set fs \<Longrightarrow> \<mu> f = {}"
  and "usable_replacement_map \<mu> (terms_of cm) nfs (set R) (set Q) (set R)"
proof -
  have "(\<forall> f. f \<notin> set fs \<longrightarrow> \<mu> f = {}) \<and> usable_replacement_map \<mu> (terms_of cm) nfs (set R) (set Q) (set R)"
  proof (cases cm)
    case (Derivational_Complexity F) note cm = this
    from split[unfolded cm] 
    have "full_empty (remdups (F @ default_fs)) = (fs, \<mu>, info)" by simp
    from full_empty[OF this _ _ wf, of cm]
    show ?thesis unfolding cm by auto
  next
    case (Runtime_Complexity C D) note cm = this
    let ?cond = "inn \<and> set C \<inter> {fn. defined (set R) fn} = {}"
    show ?thesis
    proof (cases ?cond)
      case False
      with split[unfolded cm]
      obtain info where "full_empty (remdups (C @ D @ default_fs)) = (fs, \<mu>, info)" by auto
      from full_empty[OF this _ _ wf, of cm]
      show ?thesis unfolding cm by auto
    next
      case True
      with split[unfolded cm]
      have "\<mu>_i_impl = (fs,\<mu>,info)" by auto
      from \<mu>_i_impl[OF this] have \<mu>: "\<mu> = \<mu>_i icap' (set R) (set Q)" and 
        fs: "(\<forall>f. f \<notin> set fs \<longrightarrow> \<mu> f = {})" by auto
      show ?thesis
        by (rule conjI[OF fs], unfold \<mu> cm, rule hirokawa_moser_4_8_1[OF _ icap],
        insert True[unfolded inn] wf, auto)
    qed
  qed
  thus "\<And> f. f \<notin> set fs \<Longrightarrow> \<mu> f = {}"
  and "usable_replacement_map \<mu> (terms_of cm) nfs (set R) (set Q) (set R)"
    by blast+
qed

lemma get_fs_\<mu>_DP: assumes get: "get_fs_\<mu>_DP inn S cm = (fs,\<mu>,info)"
  and wf: "wf_trs (set R)"
  and S: "set S \<subseteq> set R"
  and inn: "inn = (NF_terms (set Q) \<subseteq> NF_trs (set R))"
  shows "\<And> f. f \<notin> set fs \<Longrightarrow> \<mu> f = {}"
  and "usable_replacement_map \<mu> (terms_of cm) nfs (set R) (set Q) (set S)"
proof -
  note d = get_fs_\<mu>_DP_def
  obtain fs' \<mu>' info where get': "get_fs_\<mu> inn cm = (fs',\<mu>',info)" by (cases "get_fs_\<mu> inn cm")
  from get_fs_\<mu>[OF this wf inn] have f: "\<And> f. f \<notin> set fs' \<Longrightarrow> \<mu>' f = {}" 
    and urm: "usable_replacement_map \<mu>' (terms_of cm) nfs (set R) (set Q) (set R)" by auto
  from usable_replacement_map_mono[OF _ _ S urm]
  have urm: "usable_replacement_map \<mu>' (terms_of cm) nfs (set R) (set Q) (set S)" by auto
  have "(\<forall> f. f \<notin> set fs \<longrightarrow> \<mu> f = {}) \<and> usable_replacement_map \<mu> (terms_of cm) nfs (set R) (set Q) (set S)"
  proof (cases "(fs,\<mu>,info) = (fs',\<mu>',info)")
    case True
    with f urm show ?thesis by auto
  next
    case False
    note get = get[unfolded d get' Let_def split]
    from get False obtain res where check: "check_DP_complexity R cm = Inr res" by (cases "check_DP_complexity R cm", auto)
    obtain RS R' Cp FS F where res: "res = (RS, R', Cp, FS, F)" by (cases res, auto)
    from get[unfolded check res, simplified] False have S: "set S \<subseteq> set RS"
    and fs: "fs = list_inter fs' Cp" and \<mu>: "\<mu> = (\<lambda>f. if f \<in> set Cp then \<mu>' f else {})" 
      by (auto split: if_splits)
    from check_DP_complexity[OF check[unfolded res] wf]
    have R: "set R = set RS \<union> set R'" and dp: "is_DP_complexity (set Cp) (set FS) (set F) (set RS) (set R') cm"
      by auto
    have urm: "usable_replacement_map \<mu> (terms_of cm) nfs (set R) (set Q) (set S)"
      by (rule avanzini_14_34[OF urm[unfolded R] dp S, folded R], auto simp: \<mu>)
    show ?thesis
      by (rule conjI[OF _ urm], auto simp: fs \<mu> f)
  qed
  thus "\<And> f. f \<notin> set fs \<Longrightarrow> \<mu> f = {}" and "usable_replacement_map \<mu> (terms_of cm) nfs (set R) (set Q) (set S)" 
    by auto
qed

end
end

declare usable_replacement_map.get_fs_\<mu>.simps[code]
declare usable_replacement_map.get_fs_\<mu>_DP_def[code]
declare usable_replacement_map.\<mu>_i_impl_def[code]
declare usable_replacement_map.default_fs_def[code]
declare usable_replacement_map.innermost_repl_map_impl_def[code]

definition show_position_set :: "'f \<times> nat \<Rightarrow> nat set \<Rightarrow> shows" where
  "show_position_set f s = shows_list [Suc i . i <- [0 ..< snd f], i \<in> s]"

definition
  rule_shift_complexity_urm_tt ::
    "('tp, 'f, string) tp_ops \<Rightarrow> ('f::{show,key}, string) redtriple \<Rightarrow> ('f,string)rules \<Rightarrow>
    ('f,string)complexity_measure \<Rightarrow> complexity_class \<Rightarrow> 'tp proc"
where
  "rule_shift_complexity_urm_tt I rp Rdelete cm cc tp \<equiv> let 
      Rb = tp_ops.rules I tp;
      R = tp_ops.R I tp;
      Rw = tp_ops.Rw I tp;
      R2 = ceta_list_diff R Rdelete ;
      Q = tp_ops.Q I tp;
      inn = tp_ops.NFQ_subset_NF_rules I tp in
 check_return (do {
     check_subseteq Rdelete Rb 
        <+? (\<lambda> lr. ''rule '' +#+ shows_rule lr +@+ 
          shows '' should be deleted, but does not occur in problem'');
     check_wf_trs Rb;
     let (fs,\<mu>,info) = usable_replacement_map.get_fs_\<mu>_DP Rb (icap_impl' Rb Q) inn Rdelete cm;
     redtriple.valid rp;
     (check_allm (\<lambda> f. check (\<mu> f \<subseteq> redtriple.mono_af rp f) 
       (shows ''error in monotonicity: strict order for '' +@+ shows f
       +@+ shows '' ensures monotonicity in positions '' +@+ show_position_set f (redtriple.mono_af rp f)
       +@+ shows_nl +@+ shows ''but usable replacement map is '' +@+
       show_position_set f (\<mu> f))) fs) <+? 
     (\<lambda> s. s +@+ shows_nl +@+ shows ''the computed usable replacement map ('' +@+ shows info +@+ shows '') is'' +@+ shows_nl +@+
       shows_sep (\<lambda> f. shows ''mu('' +@+ shows f +@+ shows '') = '' +@+ show_position_set f (\<mu> f)) shows_nl fs
       +@+ shows_nl +@+ shows ''and mu(f) = {} for all other symbols f'');
     check_allm (redtriple.s rp) Rdelete
       <+? (\<lambda>s. ''problem when orienting strict TRS'' +#+ shows_nl +@+ s);
     check_allm (redtriple.ns rp) (Rw @ R2)
       <+? (\<lambda>s. ''problem when orienting non-strict TRS'' +#+ shows_nl +@+ s);
     cc' <- redtriple.cpx rp cm;
     check (cc' \<le> cc) (shows ''could only derive complexity '' +@+ shows cc')
   } <+? (\<lambda>s. ''could not derive the intended complexity '' +#+ shows cc +@+ shows '' from the following'' +@+ shows_nl +@+
     (redtriple.desc rp) +@+ shows_nl +@+ s))
     (tp_ops.mk I (tp_ops.nfs I tp) (tp_ops.Q I tp) R2 (list_union Rw Rdelete))"

lemma rule_shift_complexity_urm_tt:
  assumes "tp_spec I"
  and grp: "generic_redtriple_impl grp"
  and res: "rule_shift_complexity_urm_tt I (grp rp) Rdelete cm cc tp = return tp'"
  and cpx: "deriv_bound_measure_class (rel_qrstep (tp_ops.qreltrs I tp')) cm cc"
  shows "deriv_bound_measure_class (rel_qrstep (tp_ops.qreltrs I tp)) cm cc"
proof -
  interpret tp_spec I by fact
  interpret generic_redtriple_impl grp by (rule grp)
  note res = res[unfolded rule_shift_complexity_urm_tt_def Let_def, simplified]
  let ?rp = "grp rp"
  let ?R = "set (R tp)"
  let ?Rw = "set (Rw tp)"
  let ?D = "ceta_list_diff (R tp) Rdelete"
  let ?RwD = "Rw tp @ ?D"
  let ?nfs = "NFS tp"
  let ?Q = "qrstep ?nfs (set (Q tp))"
  let ?us = "usable_replacement_map.get_fs_\<mu>_DP (rules tp) (icap_impl' (rules tp) (Q tp))
         (NF_terms (set (Q tp)) \<subseteq> NF_trs (set (R tp) \<union> set (Rw tp))) Rdelete cm"
  let ?pi = "mono_af (grp rp)"
  obtain fs \<mu> info where us: "?us = (fs,\<mu>,info)" by (cases ?us)
  note res = res[unfolded us, simplified]
  from res have valid: "isOK (redtriple.valid ?rp)"
    and S: "isOK(check_allm (redtriple.s ?rp) Rdelete)" 
    and NS: "isOK(check_allm (redtriple.ns ?rp) ?RwD)"
    and af: "\<And> f. f \<in> set fs \<Longrightarrow> \<mu> f \<subseteq> ?pi f"
    and wf: "wf_trs (set (rules tp))"
    and subset: "set Rdelete \<union> (?Rw \<union> set ?D) \<subseteq> ?R \<union> ?Rw" "set Rdelete \<subseteq> ?R \<union> ?Rw"
    by auto 
  have rules: "set (rules tp) = ?R \<union> ?Rw" by auto
  note urm = usable_replacement_map.get_fs_\<mu>_DP[OF refl us wf, unfolded rules, OF subset(2) refl]
  have af: "af_subset \<mu> ?pi" unfolding af_subset_def
  proof 
    fix f
    show "\<mu> f \<subseteq> ?pi f"
      by (cases "f \<in> set fs", rule af, insert urm(1), auto)
  qed
  let ?cpx = "redtriple.cpx (grp rp) cm"
  from res have tp': "tp' = mk ?nfs (Q tp) ?D (list_union (Rw tp) Rdelete)" by simp
  from cpx[unfolded tp', simplified] 
  have bound: "deriv_bound_measure_class (relto (?Q (?R - set Rdelete)) (?Q (?Rw \<union> set Rdelete))) cm cc" .
  from res obtain cc' where cpx: "?cpx = Inr cc'" by (cases ?cpx, auto)
  from res cpx have le: "O_of cc' \<subseteq> O_of cc" by simp
  from generate_redtriple[OF valid S NS, of  Nil]
  obtain S NS NST where "cpx_ce_af_redtriple_order S NS NST (redtriple.af ?rp) (redtriple.mono_af ?rp) (redtriple.cpx ?rp)" 
  and S: "set Rdelete \<subseteq> S" and NS: "?Rw \<union> set ?D \<subseteq> NS" 
    by auto
  interpret cpx_ce_af_redtriple_order S NS NST "redtriple.af ?rp" "(redtriple.mono_af ?rp)" "redtriple.cpx ?rp" by fact
  have bound1: "deriv_bound_measure_class (relto (?Q (set Rdelete)) (?Q (?Rw \<union> set ?D))) cm cc'"
    by (rule avanzini_14_10[OF S NS usable_replacement_map_mono[OF _ subset(1) _ urm(2)] af cpx], auto)
  have bound1: "deriv_bound_measure_class (relto (?Q (set Rdelete)) (?Q (?Rw \<union> set ?D))) cm cc"
    by (rule deriv_bound_measure_class_mono[OF _ _ le bound1], auto)
  have bound: "deriv_bound_measure_class (relto (?Q (set Rdelete \<union> set ?D)) (?Q ?Rw)) cm cc"
    unfolding qrstep_union
    by (rule deriv_bound_relto_measure_class_union, insert bound bound1, auto simp: qrstep_union)
  have bound: "deriv_bound_measure_class (relto (?Q ?R) (?Q ?Rw)) cm cc" 
    by (rule deriv_bound_measure_class_mono[OF relto_mono[OF qrstep_mono[OF _ subset_refl] subset_refl] subset_refl subset_refl bound], auto)
  thus ?thesis by simp
qed

end
