(*  Title:       Formalization of Decreasing Diagrams for Term Rewrite Systems
    Author:      Harald Zankl  <Harald.Zankl at uibk.ac.at>, 2014
    Maintainer:  Harald Zankl  <Harald.Zankl at uibk.ac.at>, 2014
*)

(*
Copyright 2014 Harald Zankl 

This formalization is free software: you can redistribute it and/or modify it under
the terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later version.

This formalization is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with the formalization. If not, see <http://www.gnu.org/licenses/>.
*)

header "Decreasing Diagrams2"

theory Decreasing_Diagrams2 
imports 
  "$AFP/Decreasing-Diagrams/Decreasing_Diagrams" 
  Critical_Pairs
begin

(* abbreviations for hidden consts from Decreasing_Diagrams  *)
abbreviation "DD_D \<equiv> Decreasing_Diagrams.D"
abbreviation "DD_seq \<equiv> Decreasing_Diagrams.seq"
abbreviation "DD_measure \<equiv> Decreasing_Diagrams.measure"
abbreviation "DD_labels \<equiv> Decreasing_Diagrams.labels"

(*Section 1: decreasingness for families of relations *)
text {*ARSs whose relation is the union of an indexed set of relations.*}

definition fars :: "('a \<Rightarrow> 'b rel) \<Rightarrow> 'a set \<Rightarrow> 'b rel"
 where "fars r I = (\<Union>a\<in>I. r a)"

definition fars_ld :: "'a rel \<Rightarrow> ('a \<Rightarrow> 'b rel) \<Rightarrow> 'a set \<Rightarrow> bool"
 where "fars_ld gt r I = (\<forall> a \<in> I. \<forall> b \<in> I. \<forall> x y z. (x,y) \<in> r a \<and> (x,z) \<in> r b \<longrightarrow> (\<exists> y1 y2 z1 z2 v. 
  ( (y,y1) \<in> (\<Union>c\<in>ds gt {a}. r c)\<^sup>* \<and> ((y1,y2) \<in> (r b)\<^sup>=) \<and> (y2,v) \<in> (\<Union>c\<in>ds gt {a,b}. r c)\<^sup>* \<and>
    (z,z1) \<in> (\<Union>c\<in>ds gt {b}. r c)\<^sup>* \<and> ((z1,z2) \<in> (r a)\<^sup>=) \<and> (z2,v) \<in> (\<Union>c\<in>ds gt {a,b}. r c)\<^sup>*)))"

definition fars_eld :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> ('a \<Rightarrow> 'b rel) \<Rightarrow> bool"
 where "fars_eld gt ge r = ((\<forall> x y z a b. (x,y) \<in> r a \<and> (x,z) \<in> r b \<longrightarrow> (\<exists> y1 y2 z1 z2 v. 
  ( (y,y1) \<in> (\<Union>c\<in>ds gt {a}. r c)\<^sup>* \<and> (y1,y2) \<in> (\<Union>c\<in>ds ge {b}. r c)\<^sup>= \<and> (y2,v) \<in> (\<Union>c\<in>ds gt {a,b}. r c)\<^sup>* \<and>
    (z,z1) \<in> (\<Union>c\<in>ds gt {b}. r c)\<^sup>* \<and> (z1,z2) \<in> (\<Union>c\<in>ds ge {a}. r c)\<^sup>= \<and> (z2,v) \<in> (\<Union>c\<in>ds gt {a,b}. r c)\<^sup>*))))"

definition C_idx :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a \<Rightarrow> 'a set"
 where "C_idx gt ge a = ds ge {a} - ds gt {a}"

definition C_I :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a set set"
 where "C_I gt ge = {C. \<exists>a. (C = C_idx gt ge a)}"

definition C_r :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> ('a \<Rightarrow> 'b rel) \<Rightarrow> 'a set \<Rightarrow> 'b rel"
 where "C_r gt ge r C = (\<Union>c \<in> C. r c)"

definition C_rel :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> ('a \<Rightarrow> 'b rel) \<Rightarrow> ('a set \<Rightarrow> 'b rel)"
 where "C_rel gt ge r = (\<lambda>C. C_r gt ge r C)"

definition C_gt :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> 'a set rel"
 where "C_gt gt ge  = {(C,D). (\<exists> (a,b)\<in> gt. (C = C_idx gt ge a) \<and> (D = C_idx gt ge b))}"

definition rel_props :: "'a rel \<Rightarrow> 'a rel \<Rightarrow> bool"
 where "rel_props gt ge = (trans gt \<and> trans ge \<and> wf gt \<and> refl ge \<and> (ge O gt O ge \<subseteq> gt))"

(* interaction of gt and ge *)
lemma ds_gt_closed:
  assumes "rel_props gt ge" and "(a,b) \<in> ge" and "c \<in> ds gt {a}"
  shows "c \<in> ds gt {b}" proof -
  have "(c,c) \<in> ge" using assms unfolding rel_props_def by (metis UNIV_I refl_onD)
  thus ?thesis using assms unfolding rel_props_def ds_def by auto
qed

lemma ds_gt_closedX:
  assumes "rel_props gt ge" and "(a,b) \<in> ge" and "(a',b') \<in> ge" and "c \<in> ds gt {a,a'}"
  shows "c \<in> ds gt {b,b'}" proof -
  have gt: "(c,a) \<in> gt \<or> (c,a') \<in> gt" using assms unfolding ds_def by auto
  have "(c,c) \<in> ge" using assms unfolding rel_props_def by (metis UNIV_I refl_onD)
  hence "(c,b) \<in> gt \<or> (c,b') \<in> gt" using assms(1) unfolding rel_props_def using gt assms(2,3) by auto
  thus ?thesis unfolding ds_def by auto
qed

(* properties of the sets and order on the sets *)
lemma C_idx_mem: assumes "rel_props gt ge" shows "a \<in> C_idx gt ge a" proof -
 have ge: "(a,a) \<in> ge" using assms unfolding rel_props_def by (metis UNIV_I refl_onD)
 have gt: "(a,a) \<notin> gt" using assms unfolding rel_props_def by (metis wf_not_refl)
 show ?thesis using assms gt ge unfolding C_idx_def ds_def by auto 
qed

lemma rel_imp_C_r: assumes "rel_props gt ge" and "(x,y) \<in> r a" shows "(x,y) \<in> C_r gt ge r (C_idx gt ge a)" proof -
 have "a \<in> C_idx gt ge a" using assms C_idx_mem by fast
 thus ?thesis using assms unfolding C_r_def C_idx_def ds_def by auto
qed

lemma gt_imp_C_gt: assumes  "(a,b) \<in> gt" shows "(C_idx gt ge a,C_idx gt ge b) \<in> C_gt gt ge" using assms unfolding C_idx_def C_gt_def ds_def by auto

lemma C_gt_imp_gt: assumes "rel_props gt ge" and "(C_idx gt ge a,C_idx gt ge b) \<in> C_gt gt ge" shows "(a,b) \<in> gt" proof -
 from assms obtain a' b' where gt:"(a',b') \<in> gt" and "C_idx gt ge a = C_idx gt ge a'"and "C_idx gt ge b = C_idx gt ge b'"  unfolding C_gt_def by auto
 have "a \<in> C_idx gt ge a'"by (metis C_idx_mem `C_idx gt ge a = C_idx gt ge a'` assms(1))
 hence a:"(a,a') \<in> ge" unfolding C_idx_def ds_def by auto 
 have "b' \<in> C_idx gt ge b" by (metis C_idx_mem `C_idx gt ge b = C_idx gt ge b'` assms(1))
 hence b:"(b',b) \<in> ge" unfolding C_idx_def ds_def by auto
 show "(a,b) \<in> gt" using assms(1) a b gt unfolding rel_props_def by auto
qed

lemma ds_imp_C_ds: 
 assumes "rel_props gt ge" and "a \<in> ds gt {b,c}" 
 shows "C_idx gt ge a \<in> ds (C_gt gt ge) {C_idx gt ge b,C_idx gt ge c}" proof - 
 from  assms have gt: "(a,b) \<in> gt \<or> (a,c) \<in> gt" unfolding rel_props_def ds_def by auto
 thus ?thesis proof (cases "(a,b) \<in> gt") 
  case True thus ?thesis using gt_imp_C_gt unfolding ds_def by auto
 next
  case False hence "a \<in> ds gt {c}" using gt unfolding ds_def by auto 
  thus ?thesis using gt_imp_C_gt unfolding ds_def by fast
 qed
qed

lemma r_eq_C_rel:
assumes "rel_props gt ge" 
shows "(\<Union>c\<in>C_I gt ge. C_rel gt ge r c) = (\<Union>c\<in>UNIV. r c)" (is "?l = ?r") proof 
 show "?l \<subseteq> ?r" proof
  fix x y assume A: "(x, y) \<in> ?l" show "(x, y) \<in> ?r" proof -
   from A obtain C where "C \<in> C_I gt ge" and mem: "(x,y) \<in> C_rel gt ge r C" by auto
   then obtain c where C_dec: "C = C_idx gt ge c" unfolding C_I_def C_rel_def C_r_def  by auto
   have "(x, y) \<in> UNION (C_idx gt ge c) r"  using mem unfolding C_rel_def C_r_def C_dec by auto
   then obtain d where "d \<in> C_idx gt ge c" and mem: "(x,y) \<in> r d" unfolding C_idx_def by auto
   thus ?thesis using mem by auto
  qed
 qed
 next
 show "?r \<subseteq> ?l" proof 
  fix x y assume A: "(x,y) \<in> ?r" show "(x,y) \<in> ?l" proof -
   from A obtain c where mem2:"(x,y) \<in> r c" by auto
   hence 0:"c \<in> C_idx gt ge c" using C_idx_mem assms by fast
   hence 1: "C_idx gt ge c \<in> C_I gt ge"  unfolding C_I_def by auto
   have 2: "(x,y) \<in> C_rel gt ge r (C_idx gt ge c)" unfolding C_rel_def C_r_def using mem2 0 by auto
   show ?thesis using 1 2 by auto
  qed
 qed
qed

lemma C_trans: 
assumes "rel_props gt ge" 
shows "trans (C_gt gt ge)" (is "trans ?gt") proof -
{ fix C D E assume A: "(C,D) \<in> ?gt" and B: "(D,E) \<in> ?gt" 
  from A obtain c d where A3:"C = C_idx gt ge c" and A4:"D = C_idx gt ge d" and A5:"(c,d) \<in> gt" unfolding  C_gt_def by auto
  from B obtain d' e where  B3:"D = C_idx gt ge d'" and B4: "E = C_idx gt ge e" and B5:"(d',e) \<in> gt"  unfolding C_gt_def by auto
  have cc: "(c,c) \<in> ge" using assms(1) unfolding rel_props_def by (metis UNIV_I refl_onD)  
  have "(d,d') \<in> ge" using A4 B3 C_idx_mem[OF assms] unfolding C_idx_def ds_def by auto 
  hence "(c,d') \<in> gt" using cc A5 assms(1) unfolding rel_props_def by auto
  hence "(c,e) \<in> gt" using B5 assms(1) unfolding rel_props_def by (metis transE)
  hence "(C,E) \<in> ?gt" unfolding A3 B4 using gt_imp_C_gt by fast
}
 hence "\<forall> C D E. (C,D) \<in> ?gt \<and> (D,E) \<in> ?gt \<longrightarrow> (C,E) \<in> ?gt" by auto
 thus ?thesis by (metis (poly_guards_query) trans_def)
qed

lemma C_eq_imp_ge: assumes "rel_props gt ge" and "C_idx gt ge c = C_idx gt ge d" shows "(c,d) \<in> ge" proof -
 from C_idx_mem assms have "c \<in> C_idx gt ge d"  by fast
 thus ?thesis unfolding C_idx_def ds_def by auto
qed

lemma C_wf_inv: assumes "rel_props gt ge" and  "\<not> (wf (C_gt gt ge))" (is "\<not> (wf ?gt)") shows "\<not> (wf gt)" proof -
 from assms have "\<not> (SN (?gt\<inverse>))" using assms unfolding rel_props_def using  SN_iff_wf converse_converse by metis
 from this obtain s where fact: "\<forall>i. (s i, s (Suc i)) \<in> ?gt\<inverse>" unfolding SN_def SN_on_def by auto
 let ?t = "\<lambda> i. (SOME x. ((s i) = C_idx gt ge x))"
 have fact:"\<forall>i. (?t i, ?t (Suc i)) \<in> gt\<inverse>" proof 
  fix i have gt: "(s i, s (Suc i)) \<in> ?gt\<inverse>" using fact by auto
  from gt have "s i \<in> C_I gt ge"  using assms(1) unfolding rel_props_def unfolding C_gt_def  unfolding C_I_def by auto
  then obtain c where si: "s i = C_idx gt ge c" unfolding C_I_def by auto
  have 1: "(c,SOME x. s i = C_idx gt ge x) \<in> ge" using C_eq_imp_ge[OF assms(1)]  si by (metis (mono_tags) tfl_some) 
  from gt have "s (Suc i) \<in> C_I gt ge" using assms(1) unfolding rel_props_def unfolding C_gt_def  unfolding C_I_def by auto
  then obtain d where sii: "s (Suc i) = C_idx gt ge d" unfolding C_I_def by auto
  have 3: "(SOME x. s (Suc i) = C_idx gt ge x,d) \<in> ge" using C_eq_imp_ge[OF assms(1)] sii  by (metis (mono_tags) tfl_some) 
  have 2:"(c,d) \<in> gt\<inverse>" using gt unfolding si sii using C_gt_imp_gt assms by auto
  thus "(SOME x. s i = C_idx gt ge x,SOME x. s (Suc i) = C_idx gt ge x) \<in> gt\<inverse>" using assms(1) 1 2 3 unfolding rel_props_def by auto
 qed
 hence "\<not> (SN (gt\<inverse>))"  using chain_imp_not_SN_on fact unfolding SN_on_def by auto
 thus ?thesis unfolding SN_iff_wf converse_converse by auto
qed

lemma C_wf: assumes "rel_props gt ge" shows "wf (C_gt gt ge)" using C_wf_inv assms unfolding rel_props_def by auto

(*relating gt to C_gt *)
lemma key1: 
assumes "rel_props gt ge" and "(y, y1) \<in> (UNION (gt \<down>s {a}) r)\<^sup>*" 
shows "(y, y1) \<in> (UNION ((C_gt gt ge) \<down>s {C_idx gt ge a}) (C_rel gt ge r))\<^sup>*" proof -
 from assms obtain n where "(y, y1) \<in> (UNION (gt \<down>s {a}) r)^^n" by auto
 thus ?thesis proof (induct "n" arbitrary: y y1)
 case 0 thus ?case by auto
 next
 case (Suc n) 
  then obtain z where "(y, z) \<in> (\<Union>a\<in>gt \<down>s {a}. r a)^^n" and step: "(z, y1) \<in> (\<Union>a\<in>gt \<down>s {a}. r a)" by auto
  hence steps: "(y, z) \<in> (\<Union>a\<in>C_gt gt ge \<down>s {C_idx gt ge a}. C_rel gt ge r a)\<^sup>*" using Suc by auto
  from step obtain \<alpha> where 1: "(z,y1) \<in> r \<alpha>" and 2:"\<alpha> \<in> ds gt {a}" by auto
  from 2 have x: "(C_idx gt ge \<alpha>) \<in> ds (C_gt gt ge) {C_idx gt ge a}" using gt_imp_C_gt unfolding ds_def by auto 
  from 1 have "(z, y1) \<in> C_r gt ge r (C_idx gt ge \<alpha>)" using rel_imp_C_r[OF assms(1)] 1 by auto 
  hence "(z,y1) \<in> (\<Union>a \<in> ds (C_gt gt ge) {C_idx gt ge a}. C_rel gt ge r a)" using 2 x unfolding C_gt_def C_idx_def C_rel_def C_r_def  ds_def by auto 
  thus ?case using steps by (metis (erased, hide_lams) rtrancl.rtrancl_into_rtrancl)
 qed
qed

lemma key2: 
assumes "rel_props gt ge" and "(y, y1) \<in> (UNION (ge \<down>s {a}) r)\<^sup>=" 
shows  "(y, y1) \<in> (C_rel gt ge r (C_idx gt ge a))\<^sup>= \<or> (y,y1) \<in> (UNION (ds (C_gt gt ge) {C_idx gt ge a}) (C_rel gt ge r))" proof (cases "y = y1")
 case True thus ?thesis by auto
next
 case False hence "(y, y1) \<in> (UNION (ge \<down>s {a}) r)" using assms by auto
 then obtain \<alpha> where step: "(y,y1) \<in> r \<alpha>" and ge: "(\<alpha>,a) \<in> ge" unfolding ds_def by auto
 show ?thesis proof (cases "(\<alpha>,a) \<in> gt")
  case True 
  have "(y, y1) \<in> C_r gt ge r (C_idx gt ge \<alpha>)" using step True rel_imp_C_r[OF assms(1)] step  by fast
  hence "(y,y1) \<in> (UNION (ds (C_gt gt ge) {C_idx gt ge a}) (C_rel gt ge r))" using True by (metis (erased, lifting) C_rel_def UN_iff assms(1) ds_def gt_imp_C_gt mem_Collect_eq singletonI)
  thus ?thesis by auto
 next
  case False 
  have "(y, y1) \<in> (C_rel gt ge r (C_idx gt ge \<alpha>))" using rel_imp_C_r[OF assms(1)] step unfolding C_rel_def by metis
  have "(y, y1) \<in> (UNION (C_idx gt ge a) r)" unfolding C_idx_def using ge False step unfolding ds_def by auto
  thus ?thesis unfolding C_rel_def C_r_def  by auto
 qed
qed

(*generalizes key1*)
lemma key3:
assumes "rel_props gt ge" and "(y, y1) \<in> (UNION (gt \<down>s {a,b}) r)\<^sup>*" 
shows "(y, y1) \<in> (UNION ((C_gt gt ge) \<down>s {C_idx gt ge a,C_idx gt ge b}) (C_rel gt ge r))\<^sup>*" proof -
 from assms obtain n where "(y, y1) \<in> (UNION (gt \<down>s {a,b}) r)^^n" by auto
 thus ?thesis proof (induct "n" arbitrary: y y1)
 case 0 thus ?case by auto
 next
 case (Suc n) 
  then obtain z where "(y, z) \<in> (\<Union>a\<in>gt \<down>s {a,b}. r a)^^n" and step: "(z, y1) \<in> (\<Union>a\<in>gt \<down>s {a,b}. r a)" by auto
  hence steps: "(y, z) \<in> (\<Union>a\<in>C_gt gt ge \<down>s {C_idx gt ge a,C_idx gt ge b}. C_rel gt ge r a)\<^sup>*" using Suc by auto
  from step obtain \<alpha> where 1: "(z,y1) \<in> r \<alpha>" and 2:"\<alpha> \<in> ds gt {a,b}" by auto
  from 2 have x: "(C_idx gt ge \<alpha>) \<in> ds (C_gt gt ge) {C_idx gt ge a,C_idx gt ge b}" using ds_imp_C_ds[OF assms(1)] by fast
  from 1 have "(z, y1) \<in> C_r gt ge r (C_idx gt ge \<alpha>)" using rel_imp_C_r[OF assms(1)] 1 by auto 
  hence "(z,y1) \<in> (\<Union>a \<in> ds (C_gt gt ge) {C_idx gt ge a,C_idx gt ge b}. C_rel gt ge r a)" by (metis (erased, hide_lams) C_rel_def UN_iff x)
  thus ?case using steps by (metis (erased, hide_lams) rtrancl.rtrancl_into_rtrancl)
 qed
qed

lemma ds_gt_closed2: assumes "rel_props gt ge" and "(a,b) \<in> ge" "(x,y) \<in> (\<Union>c\<in>ds gt {a}. r c)^*" shows "(x,y) \<in> (\<Union>c \<in> ds gt {b}. r c)^*" proof -
 from assms(3) obtain n where "(x,y) \<in> (\<Union>c\<in>ds gt {a}. r c)^^n" by auto
 thus ?thesis proof (induct "n" arbitrary: x y)
 case 0 thus ?case by auto
 next
 case (Suc n) 
 then obtain z where "(x, z) \<in> (\<Union>a\<in>gt \<down>s {a}. r a) ^^ n" and step: "(z,y) \<in> (\<Union>a\<in>ds gt {a}. r a)" using Suc by auto
 hence steps: "(x, z) \<in> (\<Union>a\<in>gt \<down>s {b}. r a) ^*" using Suc by auto
 from step obtain \<alpha> where "(z,y) \<in> r \<alpha>" and "\<alpha> \<in> ds gt {a}" by auto
 hence "(z,y) \<in> (\<Union>a\<in>ds gt {b}. r a)" using ds_gt_closed[OF assms(1,2)] by auto
 thus ?case using steps by (metis (erased, hide_lams) rtrancl.rtrancl_into_rtrancl)
qed
qed

lemma ds_gt_closed3: assumes "rel_props gt ge" and "(a,b) \<in> ge" and "(a',b') \<in> ge" and "(x,y) \<in> (\<Union>c\<in>ds gt {a,a'}. r c)^*" shows "(x,y) \<in> (\<Union>c \<in> ds gt {b,b'}. r c)^*" proof -
 from assms(4) obtain n where "(x,y) \<in> (\<Union>c\<in>ds gt {a,a'}. r c)^^n" by auto
 thus ?thesis proof (induct "n" arbitrary: x y)
 case 0 thus ?case by auto
 next
 case (Suc n) 
 then obtain z where "(x, z) \<in> (\<Union>a\<in>gt \<down>s {a,a'}. r a) ^^ n" and step: "(z,y) \<in> (\<Union>a\<in>ds gt {a,a'}. r a)" using Suc by auto
 hence steps: "(x, z) \<in> (\<Union>a\<in>gt \<down>s {b,b'}. r a) ^*" using Suc by auto
 from step obtain \<alpha> where "(z,y) \<in> r \<alpha>" and "\<alpha> \<in> ds gt {a,a'}" by auto
 hence "(z,y) \<in> (\<Union>a\<in>ds gt {b,b'}. r a)" using ds_gt_closedX[OF assms(1,2,3)] by auto
 thus ?case using steps by (metis (erased, hide_lams) rtrancl.rtrancl_into_rtrancl)
qed
qed

lemma ds_ge_closed2: assumes "rel_props gt ge" and "(a,b) \<in> ge" "(x,y) \<in> (\<Union>c\<in>ds ge {a}. r c)^=" shows "(x,y) \<in> (\<Union>c \<in> ds ge {b}. r c)^=" proof -
 from assms have k: "(x = y) \<or> (\<exists>c \<in> ds ge {a}. (x,y) \<in> r c)" by auto
 show ?thesis proof (cases "x = y")
  case True thus ?thesis by auto
 next 
 case False then obtain c where "c \<in> ds ge {a}" and mem: "(x,y) \<in> r c" using k by auto
  hence "c \<in> ds ge {b}" using assms(1,2) unfolding rel_props_def ds_def apply auto by (metis transD)
  thus ?thesis using mem by auto
qed
qed

lemma fars_eld_imp_fars_ld:
assumes rel_props: "rel_props gt ge" and eld: "fars_eld gt ge r"
shows "fars_ld (C_gt gt ge) (C_rel gt ge r) (C_I gt ge)" unfolding fars_ld_def apply auto proof -
 fix Ca Cb x y z 
 assume A: "Ca \<in> C_I gt ge"
        "Cb \<in> C_I gt ge"
        "(x, y) \<in> C_rel gt ge r Ca"
        "(x, z) \<in> C_rel gt ge r Cb"
 show   "\<exists>y1. (y, y1) \<in> (\<Union>x\<in>C_gt gt ge \<down>s {Ca}. C_rel gt ge r x)\<^sup>* \<and>
         (\<exists>y2. ((y1, y2) \<in> C_rel gt ge r Cb \<or> y1 = y2) \<and>
         (\<exists>z1 z2 v. (y2, v) \<in> (\<Union>x\<in>C_gt gt ge \<down>s {Ca, Cb}. C_rel gt ge r x)\<^sup>* \<and>
         (z, z1) \<in> (\<Union>x\<in>C_gt gt ge \<down>s {Cb}. C_rel gt ge r x)\<^sup>* \<and> ((z1, z2) \<in> C_rel gt ge r Ca \<or> z1 = z2) \<and> (z2, v) \<in> (\<Union>x\<in>C_gt gt ge \<down>s {Ca, Cb}. C_rel gt ge r x)\<^sup>*))" proof -
 
 from A(1,3) obtain a a' where decA: "Ca = C_idx gt ge a" "(x,y) \<in> r a'" "(a',a) \<in> ge" using assms unfolding rel_props_def C_rel_def C_idx_def C_I_def C_r_def ds_def by auto
 from A(2,4) obtain b b' where decB: "Cb = C_idx gt ge b" "(x, z) \<in> r b'" "(b',b) \<in> ge" using assms unfolding rel_props_def C_rel_def C_idx_def C_I_def C_r_def ds_def by auto 
 from decA decB obtain y1 y2 z1 z2 v where
  k: "(y, y1) \<in> (UNION (gt \<down>s {a'}) r)\<^sup>*" 
     "(y1, y2) \<in> (UNION (ge \<down>s {b'}) r)\<^sup>=" 
     "(y2, v) \<in> (UNION (gt \<down>s {a', b'}) r)\<^sup>*"
     "(z, z1) \<in> (UNION (gt \<down>s {b'}) r)\<^sup>*" 
     "(z1, z2) \<in> (UNION (ge \<down>s {a'}) r)\<^sup>=" 
     "(z2, v) \<in> (UNION (gt \<down>s {a', b'}) r)\<^sup>*" using assms unfolding fars_eld_def by metis
 have k1:"(y, y1) \<in> (UNION (gt \<down>s {a}) r)\<^sup>*" using k(1) ds_gt_closed2[OF rel_props decA(3)] by fast
 have k2: "(y1, y2) \<in> (UNION (ge \<down>s {b}) r)\<^sup>=" using k(2) ds_ge_closed2[OF rel_props decB(3)] by metis
 have k3:"(y2, v) \<in> (UNION (gt \<down>s {a, b}) r)\<^sup>*" using k(3) ds_gt_closed3[OF rel_props decA(3) decB(3)] by metis
 have k4:"(z, z1) \<in> (UNION (gt \<down>s {b}) r)\<^sup>*" using k(4) ds_gt_closed2[OF rel_props decB(3)] by fast
 have k5:"(z1, z2) \<in> (UNION (ge \<down>s {a}) r)\<^sup>=" using k(5) ds_ge_closed2[OF rel_props decA(3)] by metis
 have k6:"(z2, v) \<in> (UNION (gt \<down>s {a, b}) r)\<^sup>*" using k(6) ds_gt_closed3[OF rel_props decA(3) decB(3)] by fast

 have l1:"(y, y1) \<in> (\<Union>x\<in>C_gt gt ge \<down>s {Ca}. C_rel gt ge r x)\<^sup>*" using k1 key1[OF assms(1)] unfolding decA by auto 
 have l2:"(y1, y2) \<in> (C_rel gt ge r Cb)\<^sup>= \<or> (y1, y2) \<in> UNION (C_gt gt ge \<down>s {Cb}) (C_rel gt ge r)" using key2[OF assms(1) k2] decA unfolding decB by fast
 have l3:"(y2, v) \<in> (\<Union>x\<in>C_gt gt ge \<down>s {Ca, Cb}. C_rel gt ge r x)\<^sup>*" (is "_ \<in> ?l3") using k3 key3[OF assms(1)] unfolding decA decB by auto
 from l2 have "(y1, y2) \<in> (C_rel gt ge r Cb)\<^sup>= \<or> (y1, y2) \<in> UNION (C_gt gt ge \<down>s {Ca,Cb}) (C_rel gt ge r)" unfolding ds_def by auto
 hence l23: "(((y1,y2) \<in> (C_rel gt ge r Cb)^= \<and> (y2,v) \<in> ?l3)) \<or> (y1,v) \<in> ?l3" (is "?l2 \<or> ?l3") using l3 by (metis (erased, hide_lams) converse_rtrancl_into_rtrancl)

 have r1:"(z, z1) \<in> (\<Union>x\<in>C_gt gt ge \<down>s {Cb}. C_rel gt ge r x)\<^sup>*" using k4 key1[OF assms(1)] unfolding decA decB by auto
 have r2:"(z1, z2) \<in> (C_rel gt ge r Ca)\<^sup>= \<or> (z1, z2) \<in> UNION (C_gt gt ge \<down>s {Ca}) (C_rel gt ge r)" using key2[OF assms(1) k5] decB unfolding decA by fast
 have r3:"(z2, v) \<in> (\<Union>x\<in>C_gt gt ge \<down>s {Ca, Cb}. C_rel gt ge r x)\<^sup>*" (is "_ \<in> ?r3") using k6 key3[OF assms(1)] unfolding decA decB by metis
 from r2 have "(z1, z2) \<in> (C_rel gt ge r Ca)\<^sup>= \<or> (z1, z2) \<in> UNION (C_gt gt ge \<down>s {Ca,Cb}) (C_rel gt ge r)" unfolding ds_def by auto
 hence r23: "(((z1, z2) \<in> (C_rel gt ge r Ca)\<^sup>= \<and> (z2,v) \<in> ?r3)) \<or> (z1,v) \<in> ?r3" (is "?r2 \<or> ?r3") using r3  by (metis (erased, hide_lams) converse_rtrancl_into_rtrancl)
 show ?thesis using l1 l23 r1 r23 proof (cases ?l2)
  case True hence l2: "?l2" by auto 
  thus ?thesis proof (cases ?r2)
   case True thus ?thesis using l1 r1 l2 apply auto apply metis apply metis apply metis by metis
  next
   case False hence r3: "?r3" using r23 by auto
   thus ?thesis using l1 l2 r1 r3 apply auto apply metis by metis 
  qed
 next
  case False hence l3: "?l3" using l23 by auto
  thus ?thesis proof (cases ?r2)
   case True thus ?thesis using l1 r1 l3 apply auto apply metis by metis
  next
   case False hence r3: "?r3" using r23 by auto
   thus ?thesis using l1 l2 r1 l3 apply auto apply metis apply metis by metis
 qed
qed
qed
qed

(* should go to Decreasing_Diagrams.thy *)
lemma local_peak_dec: assumes "local_peak lars P"
shows "\<exists> x y z a b. (P = ((x,[(a,y)]),(x,[(b,z)])) \<and> (x,a,y) \<in> lars \<and> (x,b,z) \<in> lars)" proof -
 from assms obtain \<sigma> \<tau> where k: "P = (\<sigma>,\<tau>)" and ts:"{\<tau>, \<sigma>} \<subseteq> DD_seq lars" and kend:" (fst \<tau> = fst \<sigma>) \<and> length (snd \<tau>) = 1 \<and> length (snd \<sigma>) = 1" unfolding local_peak_def peak_def by auto
 from k ts kend obtain a y where 1: "snd \<sigma> = [(a,y)]"  by (metis drop_1_Cons drop_eq_Nil impossible_Cons order_refl pair_collapse remdups_adj.cases)
 from this obtain x where sigma: "\<sigma> = (x,[(a,y)])" by (metis surjective_pairing)
 hence xy: "(x,a,y) \<in> lars" using ts by (metis fst_conv insert_subset seq_tail1(2) snd_conv)
 from k ts kend obtain b z where 2: "snd \<tau> = [(b,z)]"  by (metis drop_1_Cons drop_eq_Nil impossible_Cons order_refl pair_collapse remdups_adj.cases)
    from this obtain x' where tau: "\<tau> = (x',[(b,z)])" by (metis surjective_pairing)
    hence xz: "(x',b,z) \<in> lars" using ts by (metis fst_conv insert_subset seq_tail1(2) snd_conv)
 show ?thesis using xy xz tau sigma k kend 1 2 by auto
qed

lemma rel_imp_seq:
assumes  "ds gt S \<subseteq> I" and "(x, y) \<in> (UNION (gt \<down>s S) r)\<^sup>*" and "lars = {(a,c,b). (a,b) \<in> r c \<and> c \<in> I}"
shows "\<exists> \<sigma>. \<sigma> \<in> DD_seq lars \<and> fst \<sigma> = x \<and> lst \<sigma> = y \<and> set (labels \<sigma>) \<subseteq> ds gt S" proof -
 from assms obtain n where "(x, y) \<in> (UNION (gt \<down>s S) r)^^n"  using assms(1) by auto 
 thus ?thesis proof (induct n arbitrary: x y)
  case 0 
  hence "(x,[]) \<in> DD_seq lars" (is "?t2 \<in> _") using Decreasing_Diagrams.seq.intros assms by auto
  thus ?case  unfolding labels_def lst_def fst_def using 0 assms by auto
 next
  case  (Suc n) 
  then obtain z where "(x, z) \<in> (\<Union>a\<in>gt \<down>s S. r a) ^^ n" and step: "(z,y) \<in>(\<Union>a\<in>gt \<down>s S. r a)" by auto
  then obtain \<sigma> where s1: "\<sigma> \<in> DD_seq lars" and s1ex:"fst \<sigma> = x \<and> lst \<sigma> = z \<and> set (labels \<sigma>) \<subseteq> gt \<down>s S" using Suc by fast
   from step obtain a  where zy: "(z,y) \<in> r a" and mem: "a \<in> ds gt S" by auto
   hence "a \<in> I" using assms unfolding rel_props_def ds_def by auto
   hence  step': "(z,a,y) \<in> lars"  using assms zy by auto 
   have pre: "(z,[(a,y)]) \<in> DD_seq lars" (is "?\<sigma>2 \<in> _") using step' Decreasing_Diagrams.seq.intros by metis
   moreover have "fst ?\<sigma>2 = z" unfolding fst_def by auto
   moreover have "lst ?\<sigma>2 = y" unfolding lst_def by auto
   moreover have "set (labels ?\<sigma>2) \<subseteq> ds gt S" using pre mem unfolding labels_def by auto
   ultimately obtain \<sigma>2 where s2: "\<sigma>2 \<in> DD_seq lars" and s2ex: "fst \<sigma>2 = z \<and> lst \<sigma>2 = y \<and> set (labels \<sigma>2) \<subseteq> ds gt S" by auto 
  have "(fst \<sigma>, snd \<sigma> @ snd \<sigma>2) \<in> DD_seq lars" (is "?\<sigma> \<in> _") using s1 seq_concat[OF s1 s2] s1ex s2ex by metis
  moreover have pre: "fst ?\<sigma> = x" using s1ex unfolding fst_def by auto 
  moreover have "lst ?\<sigma> = y" using s2ex pre by (metis append_Nil2 append_is_Nil_conv last_appendR lst_def s1ex snd_conv)
  moreover have "set (map fst (snd ?\<sigma>)) \<subseteq> gt \<down>s S" using s1ex s2ex unfolding labels_def by auto
  ultimately show ?case unfolding labels_def by fast
 qed
qed

lemma rel_eq_imp_step_eq:
assumes "(z1, z2) \<in> (r c)\<^sup>=" and "c \<in> I" and "lars = {(a,c,b). (a,b) \<in> r c \<and> c \<in> I}"
shows "\<exists>t2. t2 \<in> DD_seq lars \<and> fst t2 = z1 \<and> lst t2 = z2 \<and> length (labels t2) \<le> 1 \<and> set (labels t2) \<subseteq> {c}" proof (cases "(z1, z2) \<in> (r c)")
 case True hence "(z1,c,z2) \<in> lars" using assms by auto
  hence "(z1,[(c,z2)]) \<in> DD_seq lars" (is "?t2 \<in> _") using Decreasing_Diagrams.seq.intros by metis
  thus ?thesis  unfolding labels_def lst_def fst_def  by force
 next
 case False 
 hence "(z1,[]) \<in> DD_seq lars" (is "?t2 \<in> _") using Decreasing_Diagrams.seq.intros assms by auto
 thus ?thesis unfolding labels_def lst_def fst_def using False assms by auto
qed

lemma fars_ld_imp_cr:
 assumes "gt \<subseteq> I \<times> I" and t: "trans gt" and w: "wf gt" and "fars_ld gt r I"
 shows "CR (\<Union>c \<in> I. r c)" proof - 
 from assms(2) obtain lars where lars_dec: "lars = {(a,c,b). (a,b) \<in> r c \<and> c \<in> I}" unfolding fars_ld_def by auto
 then obtain ars where ars_dec: "ars = unlabel lars" by auto
 have ars_dec2: "ars = (\<Union>c \<in> I. r c)" using lars_dec unfolding ars_dec unlabel_def by auto
 have p:"\<forall>P. local_peak lars P \<longrightarrow> (\<exists>\<sigma>' \<tau>'. DD lars gt (fst P, snd P, \<sigma>', \<tau>'))" proof 
  fix P show "local_peak lars P \<longrightarrow> (\<exists>\<sigma>' \<tau>'. DD lars gt (fst P, snd P, \<sigma>', \<tau>'))" proof
   assume A: "local_peak lars P" show "\<exists>\<sigma>' \<tau>'. DD lars gt (fst P, snd P, \<sigma>', \<tau>')" proof -
   from A obtain x y z a b where P_dec: "P = ((x,[(a,y)]),(x,[(b,z)]))" and  "(x,a,y) \<in> lars" and "(x,b,z) \<in> lars" using local_peak_dec by fast
   hence a: "a \<in> I"  and b: "b \<in> I" and  "(x,y) \<in> r a" and "(x,z) \<in> r b" unfolding lars_dec by auto
   then obtain y1 y2 z1 z2 v where 
    j: "(y, y1) \<in> (UNION (gt \<down>s {a}) r)\<^sup>* \<and> (y1, y2) \<in> (r b)\<^sup>= \<and> (y2, v) \<in> (UNION (gt \<down>s {a, b}) r)\<^sup>* \<and> (z, z1) \<in> (UNION (gt \<down>s {b}) r)\<^sup>* \<and> (z1, z2) \<in> (r a)\<^sup>= \<and> (z2, v) \<in> (UNION (gt \<down>s {a, b}) r)\<^sup>*"
    using assms(4) unfolding fars_ld_def by blast
    have ds1: "ds gt {a} \<subseteq> I" and ds2: "ds gt {a,b} \<subseteq> I" and ds3: "ds gt {b} \<subseteq> I" using assms(1) unfolding rel_props_def ds_def by auto
    then obtain \<sigma>1 where "\<sigma>1 \<in> DD_seq lars" "fst \<sigma>1 = y" "lst \<sigma>1 = y1" "set (labels \<sigma>1) \<subseteq> ds gt {a}" using j rel_imp_seq[OF ds1 _ lars_dec] by metis
   moreover obtain \<sigma>2 where "\<sigma>2 \<in> DD_seq lars" "fst \<sigma>2 = y1" "lst \<sigma>2 = y2" "length (labels \<sigma>2) \<le> 1" "set (labels \<sigma>2) \<subseteq> {b}"
     using j rel_eq_imp_step_eq[OF _ b lars_dec ]  by metis
   moreover obtain \<sigma>3 where "\<sigma>3 \<in> DD_seq lars" "fst \<sigma>3 = y2" "lst \<sigma>3 = v" "set (labels \<sigma>3) \<subseteq> ds gt {a,b}" using j rel_imp_seq[OF ds2 _ lars_dec] by metis
   ultimately have 1: "LDD1 lars gt (fst P) (snd P) \<sigma>1 \<sigma>2 \<sigma>3" using rel_imp_seq[OF ds3 _ lars_dec] A
    unfolding LDD1_def local_diagram1_def LD_1'_def P_dec local_peak_def peak_def lst_def labels_def ds_def by auto 
    obtain t1 where "t1 \<in> DD_seq lars" "fst t1 = z" "lst t1 = z1" "set (labels t1) \<subseteq> ds gt {b}" using j rel_imp_seq[OF ds3 _ lars_dec] by metis
   moreover obtain t2 where "t2 \<in> DD_seq lars" "fst t2 = z1" "lst t2 = z2" "length (labels t2) \<le> 1" "set (labels t2) \<subseteq> {a}" 
    using j rel_eq_imp_step_eq[OF _ a lars_dec] by metis
   moreover obtain t3 where "t3 \<in> DD_seq lars" "fst t3 = z2" "lst t3 = v" "set (labels t3) \<subseteq> ds gt {a,b}" using j rel_imp_seq[OF ds2 _ lars_dec] by metis
   ultimately have 2: "LDD1 lars gt (snd P) (fst P) t1 t2 t3" using rel_imp_seq[OF ds1 _ lars_dec] A
    unfolding LDD1_def local_diagram1_def LD_1'_def P_dec local_peak_def peak_def lst_def labels_def ds_def by auto
   have lst : "lst \<sigma>3 = lst t3" by (metis `lst \<sigma>3 = v` `lst t3 = v`) 
   have "LDD lars gt (fst P,snd P,\<sigma>1,\<sigma>2,\<sigma>3,t1,t2,t3)" using 1 2 lst unfolding LDD_def by auto
   moreover have i: "irrefl gt" using w by (metis irrefl_def wf_not_refl)
   ultimately show ?thesis using LDD_imp_DD[OF t i] by fast
   qed
  qed
 qed  
 show ?thesis using sound p t w ars_dec unfolding LD_def unfolding ars_dec2 by auto
qed

lemma fars_eld_imp_cr:
 assumes "rel_props gt ge" and "fars_eld gt ge r"
 shows "CR (\<Union>c \<in> UNIV. r c)" proof -
 from fars_eld_imp_fars_ld[OF assms] have ld: "fars_ld (C_gt gt ge) (C_rel gt ge r) (C_I gt ge)" (is "fars_ld ?gt ?r ?I") by auto
 have "CR(\<Union>c \<in> ?I. ?r c)" using fars_ld_imp_cr[OF _ C_trans[OF assms(1)] C_wf[OF assms(1)] ld] unfolding C_gt_def C_I_def by auto
 thus ?thesis using r_eq_C_rel[OF assms(1)] by metis
qed

(*Section 2: Closing local peaks in TRSs *)
type_synonym ('f,'v) step = "('f,'v) term \<times> ('f,'v) rule \<times> pos \<times> ('f,'v) subst \<times> ('f,'v) term"
type_synonym ('f,'v) lpeak = "('f,'v) step \<times> ('f,'v) step"
type_synonym ('f,'v) join = "('f,'v) step list \<times> ('f,'v) step list"
(*a labeling is a function from rewrite steps to a set that respects an ordering*)
type_synonym ('f,'v,'a) labeling = "('f,'v) step \<Rightarrow> 'a"

definition get_source :: "('f,'v) step \<Rightarrow> ('f,'v) term"
  where "get_source s = fst s"

definition get_rule :: "('f,'v) step \<Rightarrow> ('f,'v) rule"
 where "get_rule s = fst (snd s)"

definition get_pos :: "('f,'v) step \<Rightarrow> pos"
 where "get_pos s = fst (snd (snd s))"

definition get_subst :: "('f,'v) step \<Rightarrow> ('f,'v) subst"
 where "get_subst s = fst (snd (snd (snd s)))"

definition get_target :: "('f,'v) step \<Rightarrow> ('f,'v) term"
 where "get_target s = snd (snd (snd (snd s)))"

lemma rstep_r_p_s_ctxt_closed: 
  assumes "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" and "q \<in> poss u"
  shows "((ctxt_of_pos_term q u)\<langle>s\<rangle>,(ctxt_of_pos_term q u)\<langle>t\<rangle>) \<in> rstep_r_p_s R rl (q<#>p) \<sigma>"
 using assms unfolding rstep_r_p_s_def'
 apply auto
 apply (metis hole_pos_ctxt_of_pos_term hole_pos_poss)
 apply (simp add: replace_at_subt_at)
 apply (metis hole_pos_ctxt_of_pos_term hole_pos_poss_conv replace_at_subt_at subterm_poss_conv)
 by (metis assms(2) ctxt_compose_apply_distrib ctxt_of_pos_term_append ctxt_of_pos_term_hole_pos 
  hole_pos_ctxt_of_pos_term hole_pos_poss replace_at_subt_at)

lemma rstep_r_p_s_ctxt_closed2: 
  assumes "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" 
  shows "(C\<langle>s\<rangle>,C\<langle>t\<rangle>) \<in> rstep_r_p_s R rl (hole_pos C<#>p) \<sigma>"
using rstep_r_p_s_ctxt_closed[OF assms] ctxt_of_pos_term_hole_pos hole_pos_poss
by fastforce

lemma rstep_r_p_s_subst_closed:
  assumes "(s,t) \<in> rstep_r_p_s R rl p \<tau>" 
  shows "(s\<cdot>\<sigma>,t\<cdot>\<sigma>) \<in> rstep_r_p_s R rl p (\<tau>\<circ>s \<sigma>)" 
 using assms unfolding rstep_r_p_s_def' 
 by (auto simp add: ctxt_of_pos_term_subst)

fun step_close:: "('f,'v) step \<Rightarrow> ('f,'v) ctxt \<Rightarrow> ('f,'v) subst \<Rightarrow> ('f,'v) step"
  where "step_close (s, rl, p, \<sigma>, t) C \<tau> = (C\<langle>s \<cdot> \<tau>\<rangle>, rl, hole_pos C <#> p, \<sigma> \<circ>s \<tau>, C\<langle>t \<cdot> \<tau>\<rangle>)"

(* rewrite sequences *)
type_synonym ('f,'v) seq = "(('f,'v) term \<times> (('f,'v) step) list)"

inductive_set seq :: "('f,'v) trs \<Rightarrow> ('f,'v) seq set" 
for R where 
  "(s,[]) \<in> seq R"
| "(s,t) \<in> rstep_r_p_s R r p \<sigma> \<Longrightarrow> (t,ts) \<in> seq R \<Longrightarrow> (s,(s,r,p,\<sigma>,t)#ts) \<in> seq R"

lemma rstep_is_seq: 
  assumes "(s,t) \<in> rstep_r_p_s R r p \<sigma>"
  shows "(s,[(s,r,p,\<sigma>,t)]) \<in> seq R" 
using assms seq.intros by blast

fun first :: "('f,'v) seq \<Rightarrow> ('f,'v) term"
  where "first (s,_) = s"

fun last :: "('f,'v) seq \<Rightarrow> ('f,'v) term"
where 
  "last (s,[]) = s"
| "last (s,step # ss) = last (get_target step,ss)"

lemma seq_chop_first: 
  assumes "(s,(s,rl,p,\<sigma>,t) # ss) \<in> seq R"
  shows "(t,ss) \<in> seq R" and "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" 
using assms seq.cases by auto  

fun seq_close :: "('f,'v) seq \<Rightarrow> ('f,'v) ctxt \<Rightarrow> ('f,'v) subst \<Rightarrow> ('f,'v) seq"
 where "seq_close ss C \<tau> = (C\<langle>fst ss\<cdot>\<tau>\<rangle>,map (\<lambda>step. step_close step C \<tau>) (snd ss) )"

definition seq_concat :: "('f,'v) seq \<Rightarrow> ('f,'v) seq \<Rightarrow> ('f,'v) seq"
 where "seq_concat \<sigma>1 \<sigma>2 = (first \<sigma>1,snd \<sigma>1 @ snd \<sigma>2)"

lemma seq_concat_seq_close: 
  shows "seq_close (seq_concat ts us) C \<tau> = seq_concat (seq_close ts C \<tau>) (seq_close us C \<tau>)"
proof -
  obtain t tq u uq where ts_dec: "ts = (t,tq)" and us_dec: "us = (u,uq)"
    by (cases ts, cases us, auto)
  thus ?thesis unfolding seq_concat_def by auto
qed

lemma seq_concat_labels:
  shows "map l (snd (seq_concat ts us)) = (map l (snd ts)) @ (map l (snd us))"
proof -
  from assms obtain t tq u uq where ts_dec: "ts = (t,tq)" and us_dec: "us = (u,uq)"
    by (cases ts, cases us, auto)
  show ?thesis unfolding seq_concat_def by auto
qed

lemma seq_dec:
  assumes "ss \<in> seq R"
  and "map l (snd ss) = l1 @ l2"
  shows "\<exists> L1 L2. ss = seq_concat L1 L2 \<and> L1 \<in> seq R \<and> L2 \<in> seq R \<and> map l (snd L1) = l1 \<and> 
    (map l (snd L2) = l2) \<and> first ss = first L1 \<and> last L1 = first L2 \<and> last L2 = last ss"
using assms 
proof (induct l1 arbitrary: ss)
  case Nil
  hence L1: "(first ss,[]) \<in> seq R" (is "?L1 \<in> _") using seq.intros by auto
  hence ss_dec: "ss = seq_concat ?L1 ss"
    unfolding seq_concat_def apply auto 
    by (metis first.simps surjective_pairing)
  thus ?case using Nil L1 apply auto 
    by (metis first.simps surjective_pairing)
next
  case (Cons a ls)
  from Cons obtain s step ts where ss_dec: "ss = (s,step#ts)" by (cases ss, auto)
  with Cons(2) have "(s,step#ts) \<in> seq R" by auto
  then obtain rl p \<sigma> t where step_dec: "step = (s,rl,p,\<sigma>,t)" by (cases, auto)
  hence step: "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" using Cons(2)
    unfolding ss_dec step_dec using seq.cases by auto
  have i1:"(t,ts) \<in> seq R" using Cons(2) seq_chop_first(1) unfolding ss_dec step_dec by fast
  have i2:"map l (snd (t,ts)) = ls@l2" using Cons(3) Cons unfolding ss_dec step_dec by auto 
  from Cons(1)[OF i1 i2] obtain L1 L2 where 
    IH: "(t, ts) = seq_concat L1 L2" "L1 \<in> seq R" "L2 \<in> seq R"
    "map l (snd L1) = ls" "map l (snd L2) = l2" 
    "first (t, ts) = first L1" "last L1 = first L2" "last L2 = last (t, ts)" by fast
  have t: "first L1 = t" using IH(1) unfolding seq_concat_def by auto
  have f1: "ss = seq_concat (s,step # snd L1) L2" using IH(1)
    unfolding ss_dec step_dec seq_concat_def by auto
  have f2: "(s,step # snd L1) \<in> seq R"
    unfolding step_dec using seq.intros(2)[OF step] IH(2) t 
    by (metis first.cases first.simps snd_conv step_dec)
  have f4: "map l (snd (s,step# snd L1)) = a # ls"
    using IH(4) Cons(3) unfolding ss_dec step_dec by auto
  have e1: "first ss = first ((s, step # snd L1))" unfolding ss_dec by auto
  have e2: "last (s,step#snd L1) = first L2"
    using IH step_dec apply (auto simp add: get_target_def)
    by (metis first.simps surjective_pairing)
  have e3: "last L2 = last ss"
    unfolding ss_dec step_dec last.simps get_target_def using IH by auto
  show ?case using f1 f2 IH(3) f4 IH(5) using e1 e2 e3 by metis
qed

lemma seq_close_length:
  "length (snd ss) = length (snd (seq_close ss C \<tau>))"
by auto

lemma step_super_closed1: 
  fixes C \<sigma>
  shows "get_target (step_close step C \<sigma>) = C\<langle>get_target step \<cdot> \<sigma>\<rangle>"
unfolding get_target_def Let_def by (cases step, simp)

lemma seq_super_closed1:
  fixes C \<sigma>
  assumes seq:"ss \<in> seq R"
  shows "seq_close ss C \<sigma> \<in> seq R \<and> (first (seq_close ss C \<sigma>)) = C\<langle>first ss \<cdot> \<sigma>\<rangle> \<and> 
    last (seq_close ss C \<sigma>) = C\<langle>last ss \<cdot> \<sigma>\<rangle>" 
using seq
proof (induct "snd ss" arbitrary: ss)
  case Nil then obtain s where ss_dec: "ss = (s,[])" by (cases ss, auto)
  thus ?case using seq.intros unfolding ss_dec by auto
next
  case (Cons step ts)
  from Cons(2) obtain s where ss_dec: "ss=(s,step#ts)" by (cases ss, auto)
  with Cons(3) obtain rl p \<tau> t where step_dec: "step=(s,rl,p,\<tau>,t)"
    using seq.cases by blast
  hence step: "(s,t) \<in> rstep_r_p_s R rl p \<tau>" and  i:"(t,ts) \<in> seq R" (is "?ts \<in> _") 
    using Cons(3) seq.cases unfolding ss_dec step_dec by auto
  from Cons(1)[OF _ i] have ih: 
    "seq_close ?ts C \<sigma> \<in> seq R" "(first (seq_close ?ts C \<sigma>)) = (C\<langle>first ?ts\<cdot>\<sigma>\<rangle>)"
    "last (seq_close ?ts C \<sigma>) = C\<langle>last ?ts\<cdot>\<sigma>\<rangle>" 
    by auto
  have C_step: "(C\<langle>s\<cdot>\<sigma>\<rangle>,C\<langle>t\<cdot>\<sigma>\<rangle>) \<in> rstep_r_p_s R rl (hole_pos C<#>p) (\<tau>\<circ>s \<sigma>)" 
    using rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step]] by auto
  hence C_seq: "(seq_close ss C \<sigma>) \<in> seq R" 
    using ih(1) seq.intros(2)[OF C_step] unfolding ss_dec step_dec by auto
  hence C_first: "first (seq_close ss C \<sigma>) = C\<langle>first ss\<cdot>\<sigma>\<rangle>" unfolding ss_dec step_dec by auto
  have "last (seq_close ?ts C \<sigma>) = C\<langle>last ?ts\<cdot>\<sigma>\<rangle>" using ih by auto
  hence C_last: "last (seq_close ss C \<sigma>) = C\<langle>last ss\<cdot>\<sigma>\<rangle>"
    unfolding ss_dec seq_close.simps last.simps fst_conv snd_conv  
      list.map get_target_def step_dec by auto
  show ?case using C_seq C_first C_last by auto
qed

(*peaks*)
definition local_peaks :: "('f,string) trs \<Rightarrow> (('f,string) lpeak) set"
  where "local_peaks R = {((s,r1,p1,\<sigma>1,t),(s,r2,p2,\<sigma>2,u)) | s t u r1 r2 p1 p2 \<sigma>1 \<sigma>2. 
  (s,t) \<in> rstep_r_p_s R r1 p1 \<sigma>1 \<and> (s,u) \<in> rstep_r_p_s R r2 p2 \<sigma>2}"

definition critical_peaks :: "('f,string) trs \<Rightarrow> (bool \<times> ('f,string) lpeak) set"
  where "critical_peaks R = {(C = \<box>, (l \<cdot> \<sigma>, (l, r), \<epsilon>, \<sigma>, r \<cdot> \<sigma>), (l \<cdot> \<sigma>, (l', r'), hole_pos C, \<tau>, (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>)) | l r l' r' l'' C \<sigma> \<tau>.  
    (l, r) \<in> R \<and> (l', r') \<in> R \<and> l = C\<langle>l''\<rangle> \<and> is_Fun l'' \<and>
    mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>) }"

lemma critical_peak_is_local_peak:
  assumes "(b,s1,s2) \<in> critical_peaks R"
  shows "(s1,s2) \<in> local_peaks R"
using assms
unfolding critical_peaks_def local_peaks_def rstep_r_p_s_def
by auto (metis ctxt_of_pos_term_hole_pos hole_pos_subst mgu_var_disjoint_generic_sound)+

lemma critical_peaksI:
  assumes "(l, r) \<in> R" and "(l', r') \<in> R" and "l = C\<langle>l''\<rangle>"
  and "is_Fun l''" and "mgu_var_disjoint_string l'' l' = Some (\<sigma>, \<tau>)" and "s = r \<cdot> \<sigma>"
  and "t =  (C \<cdot>c \<sigma>)\<langle>r' \<cdot> \<tau>\<rangle>" and "b = (C = \<box>)"
  shows "(b, (l \<cdot> \<sigma>, (l, r), \<epsilon>, \<sigma>, s), (l \<cdot> \<sigma>, (l', r'), hole_pos C, \<tau>, t)) \<in> critical_peaks R"
using assms unfolding critical_peaks_def by blast

lemma critical_peak_to_critical_pair:
  assumes "(b,s1,s2) \<in> critical_peaks R"
  shows "(b, get_target s1,get_target s2) \<in> critical_pairs R R"
using assms 
unfolding critical_peaks_def critical_pairs_def get_target_def 
by fastforce

definition parallel_peak :: "('f,string) trs \<Rightarrow> (('f,string) lpeak) \<Rightarrow> bool"
 where "parallel_peak R p = (
  p \<in> local_peaks R \<and>
  (let ((s,r1,p1,\<sigma>1,t),(s,r2,p2,\<sigma>2,u)) = p in p1 \<bottom> p2)
 )"

definition variable_peak :: "('f,string) trs \<Rightarrow> (('f,string) lpeak) \<Rightarrow> bool"
 where "variable_peak R p = (
  p \<in> local_peaks R \<and>
  (let ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u)) = p in
  \<exists> r. ((p1 <#> r = p2) \<and> \<not> (r \<in> poss (fst rl1) \<and> is_Fun ((fst rl1) |_ r)))
 ))"

definition function_peak :: "('f,string) trs \<Rightarrow> ('f,string) lpeak \<Rightarrow> bool"
 where "function_peak R p = (
  p \<in> local_peaks R \<and>
  (let ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u)) = p in
  \<exists> r. ((p1 <#> r = p2) \<and> r \<in> poss (fst rl1) \<and> is_Fun ((fst rl1) |_ r)))
 )"

lemma function_peak_inst_of_critical_peak:
  assumes "function_peak R p"
  shows "\<exists> C \<delta> b cp. (b,cp) \<in> critical_peaks R \<and> p = (step_close (fst cp) C \<delta>, step_close (snd cp) C \<delta>)"
proof -
  from assms obtain s l1 r1 p1 \<sigma>1 t l2 r2 p2 \<sigma>2 u v where
    p: "p = ((s, (l1, r1), p1, \<sigma>1, t), (s, (l2, r2), p2, \<sigma>2, u))" and
    st:"(s,t) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1" and
    su:"(s,u) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2" and
    v:"p1 <#> v = p2" and vl1:"v \<in> poss l1" and l1v:"is_Fun (l1 |_ v)"
    unfolding function_peak_def local_peaks_def by auto
  from st su have
    p1: "p1 \<in> poss s" and \<sigma>1: "s |_ p1 = l1 \<cdot> \<sigma>1" and
    t: "t = replace_at s p1 (r1 \<cdot> \<sigma>1)" and lr1: "(l1,r1) \<in> R" and
    p2: "p2 \<in> poss s" and \<sigma>2: "s |_ p2 = l2 \<cdot> \<sigma>2" and
    u: "u = replace_at s p2 (r2 \<cdot> \<sigma>2)" and lr2: "(l2,r2) \<in> R"
    unfolding rstep_r_p_s_def' by auto
  from v have l2v: "(l1 \<cdot> \<sigma>1) |_ v = l2 \<cdot> \<sigma>2" using \<sigma>1 \<sigma>2 p1 by auto
  from l2v subt_at_subst[OF vl1] have "l1 |_ v \<cdot> \<sigma>1 = l2 \<cdot> \<sigma>2" by metis
  from mgu_var_disjoint_string_complete[OF this]
  obtain \<mu>1 \<mu>2 \<delta> where mgu: "mgu_var_disjoint_string (l1 |_ v) l2 = Some (\<mu>1, \<mu>2)"
    and d1:"\<sigma>1 = \<mu>1 \<circ>s \<delta>" and d2:"\<sigma>2 = \<mu>2 \<circ>s \<delta>" and mu12:"l1 |_v \<cdot> \<mu>1 = l2 \<cdot> \<mu>2"
    by fast
  let ?s' = "l1 \<cdot> \<mu>1"
  let ?t' = "r1 \<cdot> \<mu>1"
  let ?C' = "ctxt_of_pos_term v l1"
  let ?u' = "(?C' \<cdot>c \<mu>1)\<langle>r2 \<cdot> \<mu>2\<rangle>"
  let ?C = "ctxt_of_pos_term p1 s"
  from p1 \<sigma>1 d1 have s_inst:"?C\<langle>?s' \<cdot> \<delta>\<rangle> = s" using replace_at_ident by simp
  from d1 t have t_inst:"t = ?C\<langle>?t' \<cdot> \<delta>\<rangle>" using replace_at_ident by simp
  from u v[symmetric] d2 ctxt_of_pos_term_append[OF p1] \<sigma>1 d1
  have u_inst: "?C\<langle>?u' \<cdot> \<delta>\<rangle> = u" apply simp
    unfolding ctxt_of_pos_term_subst[OF vl1] subst_compose_apply_term_distrib[symmetric]
    by simp
  have "v = hole_pos ?C'" by (metis hole_pos_ctxt_of_pos_term vl1)
  with critical_peaksI[OF lr1 lr2 _ l1v mgu, of ?C' ?t' ?u' "?C' = \<box>"] ctxt_supt_id[OF vl1]
  have "(?C' = \<box>, (?s', (l1, r1), \<epsilon>, \<mu>1, ?t') ,(?s', (l2, r2), v, \<mu>2, ?u')) \<in> critical_peaks R"
    (is "(?b, ?cp) \<in> _") by auto
  moreover have "p = (step_close (fst ?cp) ?C \<delta>, step_close (snd ?cp) ?C \<delta>)"
    unfolding p using s_inst p1 d1 d2 v u_inst t_inst by simp
  ultimately show ?thesis by fast
qed

lemma mirror: 
  assumes "p \<in> local_peaks R" 
  shows "(snd p,fst p) \<in> local_peaks R" 
using assms unfolding local_peaks_def by auto

lemma local_peaks_helper:
  assumes pos: "p1 \<le> p2" 
  and p: "((s,r1,p1,\<sigma>1,t),(s,r2,p2,\<sigma>2,u)) \<in> local_peaks R" (is "?p \<in> _")
  shows "variable_peak R ?p \<or> function_peak R ?p" 
proof -
  from pos obtain r where "(p1 <#> r) = p2" using prefix_pos_diff by fast
  thus ?thesis using p unfolding variable_peak_def function_peak_def by auto
qed

lemma local_peaks_cases:
  assumes p: "p \<in> local_peaks R"
  shows "parallel_peak R p \<or> variable_peak R p \<or> variable_peak R (snd p, fst p) \<or> function_peak R p \<or> function_peak R (snd p, fst p)" 
using assms 
proof -
  from assms obtain  s t u r1 r2 p1 p2 \<sigma>1 \<sigma>2 where "(s,t) \<in> rstep_r_p_s R r1 p1 \<sigma>1 \<and> (s,u) \<in> rstep_r_p_s R r2 p2 \<sigma>2" 
    and p_dec: "p = ((s,r1,p1,\<sigma>1,t),(s,r2,p2,\<sigma>2,u))" unfolding local_peaks_def by fast
  thus ?thesis 
  proof (cases "p1 \<bottom> p2")
    case True 
    hence "parallel_peak R p" 
      using p unfolding parallel_peak_def p_dec Let_def by auto
    thus ?thesis by auto
  next
    case False note F = this
    thus ?thesis 
    proof (cases "p1 \<le> p2") 
      case True thus ?thesis using local_peaks_helper p p_dec by fast 
    next
      case False 
      hence "p2 \<le> p1" using F parallel_pos by auto 
      thus ?thesis using local_peaks_helper mirror p p_dec by (metis fst_conv snd_conv)
    qed
  qed  
qed

lemma variable_peak_decompose: 
  assumes "variable_peak R p" 
  shows "\<exists> s t u l1 r1 p1 \<sigma>1 l2 r2 p2 \<sigma>2 q q1 q2 w. (
   p = ((s,(l1,r1),p1,\<sigma>1,t),(s,(l2,r2),p2,\<sigma>2,u)) \<and>
   (s,t) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1 \<and>
   (s,u) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2 \<and>
   p1 <#> q = p2 \<and>
   (q1<#> q2) = q \<and>
   q1 \<in> poss l1 \<and>
   l1 |_ q1 = Var w)" 
proof -
  from assms obtain s rl1 p1 s1 t rl2 p2 s2 u q where p: "p \<in> local_peaks R" and p_dec: "p = ((s,rl1,p1,s1,t),(s,rl2,p2,s2,u))"
    and q:"p1 <#> q = p2" and px: "\<not> (q \<in> poss (fst rl1) \<and> is_Fun ((fst rl1) |_ q))" 
    unfolding variable_peak_def local_peaks_def Let_def  by blast 
  then obtain l1 r1 l2 r2 where rl1: "(l1,r1) = rl1" and rl2: "(l2,r2) = rl2" using surjective_pairing by metis
  hence s1: "(s,t) \<in> rstep_r_p_s R (l1,r1) p1 s1" and s2:"(s,u) \<in> rstep_r_p_s R (l2,r2) p2 s2" 
    using p p_dec unfolding local_peaks_def by auto
  have \<sigma>1: "s |_ p1 = l1 \<cdot> s1" using s1 unfolding rl1[symmetric] rstep_r_p_s_def' by auto
  have p2: "p2 \<in> poss s" using s2 unfolding rl2[symmetric] rstep_r_p_s_def' by auto
  from px have py: "\<not>(q \<in> poss l1 \<and> is_Fun (l1 |_ q))" using rl1 by auto 
  from py obtain q1 q2 w where k: "q = q1<#>q2" and q1:"q1 \<in> poss l1" and w:"l1 |_ q1 =  Var w" 
    using \<sigma>1 p2 pos_into_subst[of l1 s1 "s |_ p1" q] poss_append_poss q[symmetric] by auto 
  thus ?thesis using s1 s2 q p_dec rl1 rl2 by fast
qed

(* close variable peaks *)
lemma variable_steps_lin:
  assumes xy:"(x, y) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1"
  and xz:"(x, z) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2"
  and "p1 <#> v = p2"
  and "(v1 <#> v2) = v" and "v1 \<in> poss l1" and "l1 |_ v1 = Var w"
  and linl1: "linear_term l1"
  defines "\<tau> \<equiv> (\<lambda>y. if y = w then replace_at (\<sigma>1 y) v2 (r2 \<cdot> \<sigma>2) else \<sigma>1 y)"
  defines "t \<equiv> (ctxt_of_pos_term p1 x)\<langle>r1 \<cdot> \<tau>\<rangle>"
  shows "(z, t) \<in> (rstep_r_p_s R (l1, r1) p1 \<tau>)" (is "?g1")
    "linear_term r1 \<longrightarrow> (
     (y = t)  \<or> 
      (\<exists>q. (q \<in> poss r1 \<and>  (r1 |_ q = Var w) \<and> (y,t) \<in> (rstep_r_p_s R (l2,r2) (p1<#>q<#>v2) \<sigma>2))))" (is "?g2")
proof -
  from xy[unfolded rstep_r_p_s_def'] 
  have p1: "p1 \<in> poss x" and lr1: "(l1,r1) \<in> R" and \<sigma>1: "x |_ p1 = l1 \<cdot> \<sigma>1"
    and y: "y = replace_at x p1 (r1 \<cdot> \<sigma>1)" by auto
  from xz[unfolded rstep_r_p_s_def'] 
  have p2: "p2 \<in> poss x" and lr2: "(l2,r2) \<in> R" and \<sigma>2: "x |_ p2 = l2 \<cdot> \<sigma>2"
    and z: "z = replace_at x p2 (r2 \<cdot> \<sigma>2)" by auto
  from assms(3) have v:"p2 = p1 <#> v" using less_eq_pos_def by auto
  hence l2v:"(l1 \<cdot> \<sigma>1) |_ v = l2 \<cdot> \<sigma>2" using \<sigma>1 \<sigma>2 p1 p2 by auto
  from v z \<sigma>1 have zp1:"z = replace_at x p1 (replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2))" 
    using ctxt_of_pos_term_append[OF p1] ctxt_compose_apply_distrib by force
  from assms have  v12: "v = v1 <#> v2" by auto
  from assms have  w: "l1 |_ v1 = Var w" by auto
  from assms p2 have ths: "v \<in> poss (x |_ p1)" by auto
  have v1: "v1 \<in> poss l1" using assms by auto
  from assms have  wv2: "\<sigma>1 w |_ v2 = l2 \<cdot> \<sigma>2"
    by (metis \<sigma>1 l2v p2 poss_append_poss subst_apply_term.simps(1) subt_at_append subt_at_subst v1)
  have v1l1\<sigma>1:"v1 \<in> poss (l1 \<cdot> \<sigma>1)" using v1 by simp
  let ?cr = "replace_at x p1 (r1 \<cdot> \<tau>)"
  from linear_term_replace_in_subst[OF linl1 v1 w, of \<sigma>1 \<tau>, simplified]
  have "l1 \<cdot> \<tau> = replace_at (l1 \<cdot> \<sigma>1) v1 (\<tau> w)" using \<tau>_def by auto
  also have "... = replace_at (l1 \<cdot> \<sigma>1) v1 (replace_at (\<sigma>1 w) v2 (r2 \<cdot> \<sigma>2))" using \<tau>_def by auto
  also have "... = replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2)" 
    using v12 v1 w
      ctxt_of_pos_term_append[OF v1l1\<sigma>1] 
      ctxt_compose_apply_distrib by auto
  finally have "l1 \<cdot> \<tau> = replace_at (l1 \<cdot> \<sigma>1) v (r2 \<cdot> \<sigma>2)" .
  with zp1 have zp2: "z = replace_at x p1 (l1 \<cdot> \<tau>)" by auto
  hence zcr:"(z, ?cr) \<in> rstep R" using subset_rstep lr1 by auto
  have p1z: "p1 \<in> poss z" using zp1 by (metis less_eq_simps(1) p2 replace_at_below_poss v z)
  have k: "ctxt_of_pos_term p1 x = ctxt_of_pos_term p1 z" 
    using ctxt_of_pos_term_replace_at_below[OF p1] zp2 by auto
  have zcr: "(z,?cr) \<in> rstep_r_p_s R (l1,r1) p1 \<tau>" 
    unfolding rstep_r_p_s_def Let_def
    using p1z lr1 zp2 k by force
  thus "?g1" unfolding t_def by auto
  (*second goal*)
  have g2a: "(w \<notin> vars_term r1) \<longrightarrow> (y = ?cr)" 
  proof
    assume A: "w \<notin> vars_term r1" 
    show "y = ?cr" 
    proof -
      from A have "r1 \<cdot> \<tau> = r1 \<cdot> \<sigma>1" unfolding \<tau>_def by (induct r1, auto)
      hence "?cr = y" using y by auto
      thus ?thesis by auto
    qed
  qed       
 (*third goal*)
  have g2b: "w \<in> vars_term r1 \<longrightarrow> (linear_term r1 \<longrightarrow> (\<exists>q. (q \<in> poss r1 \<and>  (r1 |_ q = Var w) \<and> 
    (y, ?cr) \<in> (rstep_r_p_s R (l2,r2) (p1<#>q<#>v2) \<sigma>2))))" (is "_ \<longrightarrow> (?l \<longrightarrow> ?s)")
  proof
    assume W: "w \<in> vars_term r1" 
    show "?l \<longrightarrow> ?s"
    proof 
      assume linr1: "linear_term r1"
      show "?s" proof -
      from W obtain q where q: "q \<in> poss r1" and qw: "r1 |_ q = Var w" using supteq_imp_subt_at[OF supteq_Var[OF W]] by auto
      have qr1\<sigma>1:"q \<in> poss (r1 \<cdot> \<sigma>1)" using q by simp
      from linear_term_replace_in_subst[OF linr1 q qw, of \<sigma>1 \<tau>, simplified]
      have "r1 \<cdot> \<tau> = replace_at (r1 \<cdot> \<sigma>1) q (\<tau> w)" using \<tau>_def by auto
      also have "... = replace_at (r1 \<cdot> \<sigma>1) q (replace_at (\<sigma>1 w) v2 (r2 \<cdot> \<sigma>2))" using \<tau>_def by auto
      also have "... = replace_at (r1 \<cdot> \<sigma>1) (q <#> v2) (r2 \<cdot> \<sigma>2)" 
        using q qw ctxt_of_pos_term_append[OF qr1\<sigma>1] ctxt_compose_apply_distrib by auto
      finally have s3:"r1 \<cdot> \<tau> = replace_at (r1 \<cdot> \<sigma>1) (q <#> v2) (r2 \<cdot> \<sigma>2)" .
      from wv2 qw have "(r1 \<cdot> \<sigma>1) |_ (q <#> v2) = l2 \<cdot> \<sigma>2" 
        using q subt_at_append[OF qr1\<sigma>1, of v2] by auto
      hence ths: "r1 \<cdot> \<sigma>1 = (ctxt_of_pos_term (q <#> v2) (r1 \<cdot> \<sigma>1))\<langle>l2 \<cdot> \<sigma>2\<rangle>"
        using qr1\<sigma>1 ctxt_supt_id \<sigma>1 p2 poss_append_poss q qw subt_at_subst v v1 v12 w using y by metis
      have k: "(q<#>v2) \<in> poss (r1 \<cdot> \<sigma>1)" using q 
        by (metis v12 w v1 \<sigma>1 p2 poss_append_poss qr1\<sigma>1 qw subst_apply_term.simps(1) subt_at_subst v)
      from ths k lr2 have key: "(r1 \<cdot> \<sigma>1, r1 \<cdot> \<tau>) \<in> rstep_r_p_s R (l2,r2) (q<#>v2) \<sigma>2" 
        unfolding s3 rstep_r_p_s_def Let_def by auto 
      hence r: "((ctxt_of_pos_term p1 x)\<langle>r1\<cdot>\<sigma>1\<rangle>,(ctxt_of_pos_term p1 x)\<langle>r1 \<cdot> \<tau>\<rangle>) \<in> rstep_r_p_s R (l2,r2) (p1<#>q<#>v2) \<sigma>2"
        using rstep_r_p_s_ctxt_closed p1 by auto
      show ?thesis using q qw r unfolding y by auto
    qed
   qed
  qed
  show "?g2" using g2a g2b unfolding t_def by auto
qed

(* towards left-linear variable peaks *)

(* auxiliary functions and lemmas for parallel rewriting *)

fun mapi :: "(nat \<Rightarrow> 'a \<Rightarrow> 'b) \<Rightarrow> nat \<Rightarrow> 'a list \<Rightarrow> 'b list" where
  "mapi f i [] = []"
| "mapi f i (x # xs) = f i x # mapi f (i + 1) xs"

lemma append_mapi:
  "mapi f n (xs @ ys) = mapi f n xs @ mapi f (n + length xs) ys"
by (induct xs arbitrary: n) simp_all

lemma length_mapi:
  "length (mapi f n xs) = length xs"
by (induct xs arbitrary: n) simp_all

lemma length_concat_mapi:
  "length (concat (mapi (\<lambda>i. map (f i)) n xs)) = length (concat xs)"
by (induct xs arbitrary: n) simp_all

lemma concat_replicate_empty:
  "concat (mapi (\<lambda>i. map (f i)) n (replicate m [])) = []"
by (induct m arbitrary: n) simp_all

lemma mapi_nth:
  assumes "i < length xs"
  shows "(mapi f j xs) ! i = f (j + i) (xs ! i)"
using assms
apply (induct i arbitrary: xs j)
apply (case_tac[!] xs)
apply auto
done

lemma mapi_drop:
  shows "drop i (mapi f j xs) = mapi f (j + i) (drop i xs)"
apply (induct xs arbitrary: j i)
apply auto
apply (case_tac i)
apply auto
done

lemma mapi_map_replicate_trivial:
  shows "mapi (\<lambda>i. map (f i)) j (replicate i []) = replicate i []"
by (induct i arbitrary: j, auto)

lemma fst_concat_mapi1:
  "map fst (concat xs) = map fst (concat (mapi (\<lambda>i. map (\<lambda>(rl,p,\<sigma>). (rl,i<#p,\<sigma>))) n xs))"
by (induct xs arbitrary: n) auto

lemma fst_concat_mapi2:
  assumes "\<forall>i < length (concat steps). fst ((concat steps) ! i) = (l,r)"
  shows "\<forall>i < length (concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) n steps)). 
           fst (concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) n steps) ! i) = (l, r)"
proof (intro allI impI)
  fix i
  assume A: "i < length (concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p,
 \<sigma>))) n steps))"
  have l: "i < length (concat steps)" using A length_concat_mapi by metis
  have "map fst (concat steps) ! i = (l, r)" using nth_map l assms by auto
  hence "map fst (concat (mapi (\<lambda>i. map (\<lambda>(rl,p,\<sigma>). (rl, i <# p,\<sigma>))) n steps)) ! i = (l, r)"
    using fst_concat_mapi1 l by metis
  thus "fst (concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) n steps) ! i) = (l, r)"
    using nth_map A by auto
qed

lemma fst_concat_mapi3:
  assumes "\<forall>j < length steps. \<forall>i < length (steps ! j). fst (steps ! j ! i) = (l, r)"
  shows "\<forall>i < length (concat steps). fst ((concat steps) ! i) = (l, r)"
using assms
proof (induct steps)
  case (Cons s ss)
  show ?case
  proof (intro allI impI)
    fix i 
    assume A: "i < length (concat (s # ss))"
    show "fst (concat (s # ss) ! i) = (l, r)"
    proof (cases "i < length s") 
      case True
      have "\<forall>i < length s. fst (s ! i) = (l, r)" using Cons by auto
      hence elt: "fst (s ! i) = (l, r)" using True by auto
      from True have "s ! i = concat (s # ss) ! i" using nth_append[of s] by simp
      thus ?thesis using elt by auto
    next
      case False 
      have k: "\<forall>j < length ss. \<forall>i < length (ss ! j). fst (ss ! j ! i) = (l, r)" using Cons(2) by auto
      have k2: "\<forall>i < length s. fst (s ! i) = (l, r)" using Cons(2) by auto 
      hence k': "\<forall>i < length (concat ss). fst (concat ss ! i) = (l, r)" using Cons(1)[OF k] by auto
      have "i < length s + length (concat ss)" using k' False A by auto
      hence "fst (concat ss ! (i - length s)) = (l, r)" using False k' by auto
      thus ?thesis using k2 False using nth_append[of s] by simp
    qed
  qed
qed auto

lemma fst_concat_mapi: 
  assumes "\<forall>j<length steps. \<forall>i<length (steps ! j). fst (steps ! j ! i) = (l, r)"
  shows "\<forall>i<length (concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) n steps)). 
           fst (concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) n steps) ! i) = (l, r)"
using fst_concat_mapi2[OF fst_concat_mapi3[OF assms]] .

lemma concat_singleton:
  assumes "concat xs = [z]"
  shows "(\<exists>i < length xs. xs ! i = [z] \<and> (\<forall>j < length xs. j \<noteq> i \<longrightarrow> (xs ! j) = []))"
using assms
proof (induct xs)
  case (Cons x xs)
  show ?case
  proof (cases x)
    case Nil
    with Cons have "concat xs = [z]" by auto
    with Cons obtain i where len:"i<length xs" and z:"xs ! i = [z]" and nil:"\<forall>j<length xs. j \<noteq> i \<longrightarrow> (xs ! j) = []" by auto
    let ?i = "Suc i"
    from len have "?i<length (x # xs)"
      by auto
    moreover from z have "(x # xs) ! ?i = [z]" by auto
    moreover have "\<forall>j<length (x # xs). j \<noteq> ?i \<longrightarrow> (x # xs) ! j = []" using Nil nil 
      by (metis gr0_conv_Suc length_Cons linorder_neqE_nat not_less_eq nth_Cons' nth_Cons_Suc)
    ultimately show ?thesis by auto  
  qed (insert Cons, auto)
qed auto

lemma concat_Cons:
  assumes "concat xss = x # xs"
  shows "\<exists>i < length xss. \<exists>ys. (\<forall>j < i. xss ! j = []) \<and> 
    xss ! i = x # ys \<and> xs = ys @ concat (drop (Suc i) xss)"
using assms
proof (induct xss arbitrary: x xs)
  case (Cons z zs)
  show ?case
  proof (cases "z = []")
    case False
    with Cons have x:"x = hd z" by (metis concat.simps(2) hd_append2 list.sel(1))
    have "0 < length (z # zs)" by auto
    moreover have "\<forall>j < 0. (z # zs) ! j = []" by auto
    moreover have "(z # zs) ! 0 = x # tl z" using x nth_Cons_0 False by simp
    moreover have "xs = tl z @ concat (drop (Suc 0) (z # zs))" using Cons False    
      by (metis concat.simps(2) drop_0 drop_Suc_Cons list.sel(3) tl_append2)
    ultimately show ?thesis by auto
  next
    case True
    with Cons have "concat zs = x # xs" by auto
    from Cons(1)[OF this] obtain i ys where IH:
      "i < length zs"
      "\<forall>j<i. zs ! j = []"
      "zs ! i = x # ys"
      "xs = ys @ concat (drop (Suc i) zs)"
      by auto
    hence "(Suc i) < length (z # zs)" by auto
    moreover have "\<forall>j < Suc i. (z # zs) ! j = []" using IH True less_Suc_eq_0_disj by force
    moreover have "(z # zs) ! (Suc i) = x # ys" using IH by auto
    moreover have "xs = ys @ concat (drop (Suc (Suc i)) (z # zs))" using IH  by auto
    ultimately show ?thesis by auto
  qed
qed auto

lemma update_eq:
  assumes "\<forall>j < length ss. j \<noteq> i \<longrightarrow> ss ! j = ts ! j"
  and "length ss = length ts"
  shows "ss = ts[i:=ss ! i]"
proof -
  have "\<forall>j < length ss. ss ! j = ts[i:=ss ! i] ! j"
    apply clarsimp using assms apply (case_tac "j = i") by auto
  moreover have "length ss = length (ts[i:=ss ! i])" using assms by auto
  ultimately show ?thesis using nth_equalityI by blast
qed

lemma concat_steps_empty: 
  assumes "[] = concat (mapi (\<lambda>i. map (f i)) n steps)"
  shows "\<forall> i < length steps. steps ! i = []"
using assms
proof (induct steps arbitrary: n)
  case Nil thus ?case by auto
next
  case (Cons x xs) 
  from Cons have x:"x = []" by auto
  from Cons have "[] = concat (mapi (\<lambda>i. map (f i)) (Suc n) xs)" by auto
  hence "\<forall>i<length xs. xs ! i = []" using Cons by auto
  thus ?case using x by (metis Suc_less_eq gr0_conv_Suc length_Cons nth_Cons_Suc nth_non_equal_first_eq)
qed

(*parallel step with r/p/s information*)
inductive_set
  par_rstep2 :: "('f,'v) trs \<Rightarrow> (('f,'v)term \<times> (('f,'v) rule \<times> pos \<times> ('f,'v) subst)list \<times> ('f,'v) term) set"
  for R :: "('f,'v)trs"
where
  root_step2[intro]: "(s,t) \<in> R \<Longrightarrow> (s \<cdot> \<sigma>, [((s, t), \<epsilon>, \<sigma>)], t \<cdot> \<sigma>) \<in> par_rstep2 R"
| par_step_fun2[intro]: "\<lbrakk>\<And>i. i < length ts \<Longrightarrow> (ss ! i, steps ! i, ts ! i) \<in> par_rstep2 R\<rbrakk> \<Longrightarrow> 
    length ss = length ts \<Longrightarrow> length ss = length steps \<Longrightarrow>
    (Fun f  ss, concat (mapi (\<lambda>i si. (map (\<lambda> (rl, p, \<sigma>). (rl, i <# p,\<sigma>)) si)) 0 steps), Fun f ts) \<in> par_rstep2 R"
| par_step_var2[intro]: "(Var x, [], Var x) \<in> par_rstep2 R"

lemma par_rstep2_refl:
  shows "(s, [], s) \<in> par_rstep2 R"
proof (induct s)
  case (Fun f ts) 
  let ?steps = "replicate (length ts) []"
  {
    fix i
    assume "i < length ts"
    hence "(ts ! i, replicate (length ts) [] ! i, ts ! i) \<in> par_rstep2 R" using Fun by auto
  }
  from par_rstep2.intros(2)[OF this]
  show ?case using concat_replicate_empty length_replicate by metis
qed auto

lemma par_rstep2_empty:
  assumes "(s, [], t) \<in> par_rstep2 R"
  shows "s = t"
using assms
proof (induct s arbitrary: t)
  case (Var x) show ?case using par_rstep2.cases[OF Var] by auto
next
  case (Fun f ss)  obtain ts steps where
    steps_empty: "[] = concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) 0 steps)" and t_dec: "t = Fun f ts"
    and steps: "(\<forall>x<length ts. (ss ! x, steps ! x, ts ! x) \<in> par_rstep2 R)" and l:"length ss = length ts" "length ss = length steps" 
    using Fun(2) unfolding  par_rstep2.simps[of "Fun f ss" "[]" t R] by auto
  have steps_i: "\<forall>i < length steps. (steps ! i = [])" using concat_steps_empty steps_empty l by fast
  hence "\<forall>i<length ts. (ss ! i, steps ! i, ts ! i) \<in> par_rstep2 R" using steps l by auto 
  have "\<forall>i < length ts. (ss ! i = ts ! i)"
  proof (intro allI impI)
    fix i
    assume A: "i < length ts" 
    have "(ss ! i, steps ! i, ts ! i) \<in> par_rstep2 R" using steps A by auto
    thus "ss ! i = ts ! i" using Fun(1) by (metis A steps_i l(1) l(2) nth_mem) 
  qed
  hence "ss = ts" using l by (metis nth_equalityI)
  thus ?case unfolding t_dec by auto
qed

lemma par_rstep2_append:
  assumes "length ss = length ts" and "length ts = length steps"
  and "\<forall>i < length ss. (ss ! i, steps ! i, ts ! i) \<in> par_rstep2 R"
  and "\<forall>i < length ss'. (ss' ! i, steps' ! i, ts' ! i) \<in> par_rstep2 R"
  shows "\<forall> i < length (ss @ ss'). ((ss @ ss') ! i, (steps @ steps') ! i, (ts @ ts') ! i) \<in> par_rstep2 R"
using assms
proof (induct rule: list_induct3)
  case (Cons s ss t ts step steps)
  show ?case 
  proof (intro allI impI)
    fix i
    assume i: "i < length ((s # ss) @ ss')"
    show "(((s # ss) @ ss') ! i, ((step # steps) @ steps') ! i, ((t # ts) @ ts') ! i) \<in> par_rstep2 R"
    proof (cases i)
      case 0
      thus ?thesis using Cons by auto
    next
      case (Suc j)
      from Cons have "\<forall>i<length (ss @ ss'). ((ss @ ss') ! i, (steps @ steps') ! i, (ts @ ts') ! i) \<in> par_rstep2 R" by fastforce
      with Suc i show ?thesis by simp
    qed
  qed
qed auto

lemma par_rstep2_ctxt_closed:
  assumes "(s,steps,t) \<in> par_rstep2 R"
  shows "(C\<langle>s\<rangle>, map (\<lambda> (rl, p, \<sigma>). (rl, hole_pos C <#> p, \<sigma>)) steps, C\<langle>t\<rangle>) \<in> par_rstep2 R"
using assms
proof (induct C arbitrary:s t steps)
  case (More f ls C' rs) 
  hence step: "(C'\<langle>s\<rangle>, map (\<lambda> (rl, p, \<sigma>). (rl, hole_pos C' <#> p, \<sigma>)) steps, C'\<langle>t\<rangle>) \<in> par_rstep2 R" (is "(_,?steps,_) \<in> _")
    using More by auto
  have ls: "\<forall> i < length ls. (ls ! i, [], ls ! i) \<in> par_rstep2 R"
    using par_rstep2_refl by auto
  have rs: "\<forall> i < length rs. (rs ! i,[], rs ! i) \<in> par_rstep2 R"
    using par_rstep2_refl  by auto
  let ?steps' = "replicate (length ls) [] @ (?steps # replicate (length rs) [])"
  from par_rstep2_append[of "[C'\<langle>s\<rangle>]" "[C'\<langle>t\<rangle>]" "[?steps]"] step rs
  have "\<forall>i < length (C'\<langle>s\<rangle> # rs). ((C'\<langle>s\<rangle> # rs) ! i, (?steps # replicate (length rs) []) ! i, (C'\<langle>t\<rangle> # rs) ! i) \<in> par_rstep2 R"
    by simp
  hence x: "\<forall>i < length (ls @ C'\<langle>s\<rangle> # rs). ((ls @ C'\<langle>s\<rangle> # rs) ! i, ?steps' ! i, (ls @ C'\<langle>t\<rangle> # rs) ! i) \<in> par_rstep2 R"
    (is "\<forall> i < length _ . (?ss ! i, _, ?ts ! i) \<in> _")
    using par_rstep2_append[of ls ls] ls by simp
  have l: "length ?ss = length ?ts" by auto
  have lx: "length (replicate (length ls) []) = length ls" by auto
  have "(Fun f ?ss, concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) 0 ?steps'), Fun f ?ts) \<in> par_rstep2 R"
    using x l by auto
  hence "(Fun f ?ss,concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) (length ls) ([?steps] @ replicate (length rs) [])), Fun f ?ts)  \<in> par_rstep2 R" 
    unfolding append_mapi concat_append concat_replicate_empty[where ?n=0] by auto
  hence "(Fun f ?ss,
    concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) (length ls) [?steps]) @
    concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) (Suc (length ls)) (replicate (length rs) [])), Fun f ?ts) \<in> par_rstep2 R"
    unfolding append_mapi concat_append by auto
  hence *: "(Fun f ?ss, concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) (length ls) [?steps]), Fun f ?ts)  \<in> par_rstep2 R"
    unfolding length_append[symmetric] lx using concat_replicate_empty append_Nil2 by metis
  have f_comp: "((\<lambda>(rl, p, \<sigma>). (rl, length ls <# p, \<sigma>)) \<circ> (\<lambda>(rl, p, \<sigma>). (rl, hole_pos C' <#> p, \<sigma>))) = 
    ((\<lambda>(rl, p, \<sigma>). (rl, length ls <# hole_pos C' <#> p, \<sigma>)))" by auto 
   from * show ?case apply auto using f_comp by metis
qed auto

lemma rstep_r_p_s_imp_par_rstep2: 
  assumes "(s,t) \<in> rstep_r_p_s R (l,r) p \<sigma>"
  shows "(s, [((l, r), p, \<sigma>)], t) \<in> par_rstep2 R"
proof -
  from assms obtain C  where lr: "(l,r) \<in> R" and C_dec: "C = ctxt_of_pos_term p s" and p: "p \<in> poss s" 
    and  s_dec: "s = C\<langle>l \<cdot> \<sigma>\<rangle>" and t_dec: "t = C\<langle>r \<cdot> \<sigma>\<rangle>"  unfolding rstep_r_p_s_def by fastforce
  have h: "hole_pos C = p" unfolding C_dec using p by auto 
  hence s:"(l \<cdot> \<sigma>, [((l, r), \<epsilon>, \<sigma>)], r \<cdot> \<sigma>) \<in> par_rstep2 R" using par_rstep2.intros(1)[OF lr] by auto
  have "(C\<langle>l \<cdot> \<sigma>\<rangle>, [((l, r), p, \<sigma>)], C\<langle>r \<cdot> \<sigma>\<rangle>) \<in> par_rstep2 R" using par_rstep2_ctxt_closed[OF s] using h by auto
  thus ?thesis unfolding s_dec t_dec using par_rstep2_ctxt_closed C_dec by auto
qed

lemma par_rstep2_singleton:
  assumes "(s, steps, t) \<in> par_rstep2 R"
  and "steps = [(lr, p, \<sigma>)]"
  shows "(s, t) \<in> rstep_r_p_s R lr p \<sigma>"
  using assms
proof (induct arbitrary: lr p \<sigma>)
  case (par_step_fun2 ts ss steps f)
  let ?msteps = "mapi (\<lambda>i. map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>))) 0 steps"
  from par_step_fun2 obtain i where
    im:"i < length ?msteps" and mstepsi:"?msteps ! i = [(lr, p, \<sigma>)]"
    and mstepsnil:"\<forall>j < length ?msteps. j \<noteq> i \<longrightarrow> ?msteps ! j = []"
    using concat_singleton[of ?msteps "(lr, p, \<sigma>)"] by auto
  hence i: "i < length steps" using length_mapi by metis
  from i par_step_fun2 have its:"i < length ts" by auto
  {
    fix j
    assume j:"j < length ss"  and "j \<noteq> i"
    with par_step_fun2 mstepsnil im length_mapi have nil:"?msteps ! j = []" by metis
    from j have *:"j < length steps" by (metis par_step_fun2(4))
    from nil have nil:"steps ! j = []" unfolding mapi_nth[OF *] by simp
    have "j < length ts" using j par_step_fun2(3) by auto
    from par_step_fun2(1)[OF this, unfolded nil] have "ss ! j = ts ! j"
      using par_rstep2_empty by auto
  }
  hence  "\<forall>j < length ss. j \<noteq> i \<longrightarrow> ss ! j = ts ! j" by blast
  hence "ss = ts[i := ss ! i]" using i par_step_fun2 update_eq by blast 
  hence ts:"ts = take i ss @ ts ! i # drop (Suc i) ss"
    by (metis append_eq_conv_conj append_take_drop_id its length_take par_step_fun2.hyps(3) take_drop_update_first upd_conv_take_nth_drop)
  from mapi_nth[OF i,of _ 0] have
    stepsi:"?msteps ! i  = (\<lambda>i. map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>))) i (steps ! i)" by auto
  with mstepsi obtain q where q:"p = i<#q" by auto
  with stepsi mstepsi have stepsi:"steps ! i = [(lr, q, \<sigma>)]" by auto
  from par_step_fun2(2)[OF its stepsi] have
    "q \<in> poss (ss ! i)" "lr \<in> R"
    "(ss ! i) |_ q = fst lr \<cdot> \<sigma>" "(ts ! i) = (ctxt_of_pos_term q (ss ! i))\<langle>snd lr \<cdot> \<sigma>\<rangle>"
    unfolding rstep_r_p_s_def' by auto
  hence "(Fun f ss, Fun f ts) \<in> rstep_r_p_s R lr p \<sigma>" using q
    its par_step_fun2(3) unfolding rstep_r_p_s_def' apply auto
    using id_take_nth_drop[of i ts] ts by metis
  thus ?case by auto
next
  case (root_step2 l r \<tau>)
  thus ?case unfolding rstep_r_p_s_def by auto
qed auto


(*
lemma par_rstep2_subst_closed:
 assumes "(s,steps,t) \<in> par_rstep2 R"
 shows "(s\<cdot>\<tau>,map (\<lambda> (rl,p,\<sigma>). (rl,hole_pos C<#>p,\<sigma>\<circ>s\<tau>)) steps,t\<cdot>\<tau>) \<in> par_rstep2 R"
sorry
*)

(*
lemma assumes "\<forall>x.\<exists>y. P(x,y)" shows "\<exists>f.\<forall>x. P(x,f(x))" using assms by metis
lemma y: assumes "\<forall>j < length (s # ss). P((s # ss)!j)" shows "\<forall>j < length ss. P(ss!j)" proof
 fix j show "j < length ss \<longrightarrow> P (ss!j)" using assms by auto
qed
*)

lemma par_step_merge:
  assumes  "length ss = length ts" 
  and "\<forall> j < length ss. \<exists> steps. (\<forall>i<length steps. fst (steps ! i) = (l, r)) \<and> (ss ! j, steps, ts ! j) \<in> par_rstep2 R"
  shows "\<exists> steps. (\<forall>i<length steps. fst (steps ! i) = (l, r)) \<and> (Fun f ss, steps, Fun f ts) \<in> par_rstep2 R"
proof -
  from assms have a:"\<exists> steps. (length steps = length ss) \<and> 
    (\<forall> j < length ss. (\<forall>i<length (steps ! j). fst ((steps ! j) ! i) = (l, r)) \<and> (ss ! j, steps ! j, ts ! j) \<in> par_rstep2 R)"
  proof (induct ss arbitrary: ts)
    case (Cons s ss) 
    obtain u us where ts_dec: "ts = u # us" using Cons by (cases ts, auto)
    hence l: "length ss = length us" using Cons by auto
    from Cons(3) obtain steps1 where step1: "(\<forall>i<length steps1. fst (steps1 ! i) = (l,r)) \<and> (s, steps1, u) \<in> par_rstep2 R" 
      using ts_dec by auto
    have y: "\<forall>j<length ss. \<exists>steps. (\<forall>i<length steps. fst (steps ! i) = (l, r)) \<and> (ss ! j, steps, us ! j) \<in> par_rstep2 R"
    proof (intro allI impI)
      fix j
      assume A:"j < length ss"
      from Cons(3) A obtain steps where
        "(\<forall>i<length steps. fst (steps ! i) = (l, r)) \<and> ((s # ss) ! (j + 1), steps, (u # us) ! (j + 1)) \<in> par_rstep2 R"
        unfolding ts_dec by auto
      thus "\<exists> steps. (\<forall>i<length steps. ((fst (steps!i) = (l,r)))) \<and> (ss ! j, steps, us ! j) \<in> par_rstep2 R"
        using Cons by auto
    qed
    obtain steps2 where l1: "length steps2 = length ss" and
      steps2: "\<forall>j<length ss. (\<forall>i<length (steps2 ! j). fst (steps2 ! j ! i) = (l, r)) \<and> (ss ! j, steps2 ! j, us ! j) \<in> par_rstep2 R"
      using Cons(1)[OF l y] unfolding ts_dec by auto
    have l: "length (steps1 # steps2) = length (s # ss)" using l1 by auto
    have "\<forall>j<length (s # ss). 
      (\<forall>i<length ((steps1 # steps2) ! j). fst ((steps1 # steps2) ! j ! i) = (l,r)) \<and>  
      ((s # ss) ! j, (steps1 # steps2) ! j, ts ! j) \<in> par_rstep2 R"
    proof (intro allI impI)
      fix j
      assume j: "j < length (s # ss)"
      show "(\<forall>i<length ((steps1 # steps2) ! j). fst ((steps1#steps2)!j!i) = (l,r)) \<and> 
        ((s # ss) ! j, (steps1 # steps2) ! j, ts ! j) \<in> par_rstep2 R"
      proof (cases j) 
        case 0
        thus ?thesis using step1 unfolding ts_dec by auto
      next
        case (Suc j')
        thus ?thesis unfolding ts_dec using steps2 j by (metis length_Cons not_less_eq nth_Cons_Suc)
      qed
    qed
    thus ?case using l by metis
  qed auto
  then obtain steps where l: "length steps = length ss" and
    k: "\<forall>j<length ts. (\<forall>i<length (steps ! j). fst (steps ! j ! i) = (l, r)) \<and> (ss ! j, steps ! j, ts ! j) \<in> par_rstep2 R"
    using assms(1) by auto
  hence two: "(Fun f ss, concat (mapi (\<lambda>i. map (\<lambda>(rl, p, \<sigma>). (rl, i <# p, \<sigma>))) 0 steps), Fun f ts) \<in> par_rstep2 R"
    (is "(_,?steps,_) \<in>  _")
    using assms by auto
  have j: "\<forall>j<length steps. \<forall>i<length (steps ! j). fst (steps ! j ! i) = (l, r)" using k l assms(1) by auto
  have one: "\<forall>i<length ?steps. fst (?steps ! i) = (l, r)" using fst_concat_mapi[OF j] by auto
  show ?thesis using one two by auto
qed

lemma subst_step_imp_par_rstep2:
  assumes "(\<sigma>1(w), \<tau>(w)) \<in> rstep_r_p_s R (l, r) v2 \<sigma>2"
  and "\<forall>x. (x = w) \<or> (\<sigma>1(x) = \<tau>(x))" 
  shows "\<exists> steps. (\<forall>i < length steps. fst (steps ! i) = (l, r)) \<and> (t \<cdot> \<sigma>1, steps, t \<cdot> \<tau>) \<in> par_rstep2 R"
proof (induct t)
  case (Var y)
  thus ?case
  proof (cases "y = w")
    case True 
    let ?steps = "[((l, r), v2, \<sigma>2)]"
    have len: "\<forall>i < length ?steps. fst (?steps ! i) = (l, r)" by auto
    from True have "(\<sigma>1 y, \<tau> y) \<in> rstep_r_p_s R (l, r) v2 \<sigma>2"
      using assms unfolding par_rstep2_def by auto
    hence "(\<sigma>1 y, ?steps, \<tau> y) \<in> par_rstep2 R" using rstep_r_p_s_imp_par_rstep2 by fast
    hence "(Var y \<cdot> \<sigma>1, ?steps, Var y \<cdot> \<tau>) \<in> par_rstep2 R" by auto
    thus ?thesis using len by blast 
  next
    case False 
    have len: "\<forall>i < length []. fst ([] ! i) = (l, r)" by auto
    have id: "(Var y \<cdot> \<sigma>1, [], Var y \<cdot> \<sigma>1) \<in> par_rstep2 R" using par_rstep2_refl by auto
    have "\<sigma>1(y) = \<tau>(y)" using assms False by auto
    hence "(Var y \<cdot> \<sigma>1, [], Var y \<cdot> \<tau>) \<in> par_rstep2 R" using id by auto
    thus ?thesis using len by blast
  qed
next
  case (Fun f ts) 
  let ?ts\<sigma> = "map (\<lambda>t. t \<cdot> \<sigma>1) ts"
  let ?ts\<tau> = "map (\<lambda>t. t \<cdot> \<tau>) ts"
  from Fun have "\<forall> j < length ts. \<exists> steps. 
    (\<forall>i<length steps. fst (steps ! i) = (l, r)) \<and> (ts ! j \<cdot> \<sigma>1, steps, ts ! j \<cdot> \<tau>) \<in> par_rstep2 R" by auto
  hence x: "\<forall>j < length ?ts\<sigma>. \<exists> steps. 
    (\<forall>i<length steps. fst (steps ! i) = (l, r)) \<and> (?ts\<sigma> ! j, steps, ?ts\<tau> ! j) \<in> par_rstep2 R" by auto
  have "\<exists>steps. (\<forall>i<length steps. fst (steps ! i) = (l, r)) \<and> (Fun f ?ts\<sigma>, steps, Fun f ?ts\<tau>) \<in> par_rstep2 R"
    using par_step_merge[OF _ x] by auto
  thus ?case by auto 
qed

lemma par_rstep2_decompose:
  assumes "(s, steps, t) \<in> par_rstep2 R"
  and "steps = x # xs"
  shows "\<exists> s'. (s, [x], s') \<in> par_rstep2 R \<and> (s', xs, t) \<in> par_rstep2 R"
using assms
proof (induct  arbitrary: x xs rule: par_rstep2.induct)
  case (root_step2 s t \<sigma>)
  hence "xs = []" by auto
  hence "(t \<cdot> \<sigma>, xs, t \<cdot> \<sigma>) \<in> par_rstep2 R" using par_rstep2_refl by auto
  moreover from root_step2 have "(s \<cdot> \<sigma>, [x], t \<cdot> \<sigma>) \<in> par_rstep2 R" by auto
  ultimately show ?case by auto
next
  case (par_step_fun2 ts ss steps f)
  let ?msteps = "mapi (\<lambda>i. map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>))) 0 steps"
  from concat_Cons[OF par_step_fun2(5)] obtain i ys where
    i: "i < length ?msteps" and 
    nil: "\<forall>j < i. ?msteps ! j = []" and
    x: "?msteps ! i = x # ys" and
    xs: "xs = ys @ concat (drop (Suc i) ?msteps)" by auto
  hence ist: "i < length steps" using length_mapi by metis
  from x have "x # ys = map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>)) (steps ! i)"
    unfolding mapi_nth[OF ist, of _ 0]  by auto
  then obtain lr p \<sigma> zs where x:"x = (lr, i<#p, \<sigma>)" and zs:"steps ! i = (lr, p, \<sigma>) # zs"
    and ys:"ys = map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>)) zs" by auto
  with par_step_fun2 ist obtain s' where
    s': "(ss ! i, [(lr, p, \<sigma>)], s') \<in> par_rstep2 R \<and> (s', zs, ts ! i) \<in> par_rstep2 R"
    by fastforce
  let ?z = "(lr, p, \<sigma>)"
  let ?us = "take i ss @ s' # drop (Suc i) ss"
  let ?steps' = "replicate i [] @ [[?z]] @ replicate (length steps - (Suc i)) []"
  have steps':"length ss = length ?steps'" using ist par_step_fun2(4) by auto
  have us: "length ss = length ?us" using ist par_step_fun2(4) by auto
  {
    fix j
    assume j:"j < length ?us"
    {
      assume "j < i"
      hence "?steps' ! j = []" 
        by (metis (erased, hide_lams) length_replicate nth_append nth_replicate)
      moreover have "?us ! j = ss ! j"         
        by (metis `j < i` j nat_le_linear nth_append_take_is_nth_conv take_all us)
      ultimately have "(ss ! j, ?steps' ! j, ?us ! j) \<in> par_rstep2 R"
        using par_rstep2_refl by auto
    } moreover {
      assume "j = i"
      hence "?us ! j = s'" by (metis j nat_less_le nth_append_take us)
      moreover have "?steps' ! j = [?z]"
        by (metis `j = i` append_Cons length_replicate nth_append_length)
      ultimately have "(ss ! j, ?steps' ! j, ?us ! j) \<in> par_rstep2 R"
        using s' `j = i` by auto
    } moreover {
      assume *:"j > i"
      have "j < length steps" unfolding par_step_fun2(4)[symmetric] us using j .
      hence "j - Suc i < length steps - Suc i" using * by auto
      hence "?steps' ! j = []" using * unfolding nth_append by auto
      moreover from * have "?us ! j = ss ! j" 
        by (metis j less_imp_le_nat nat_le_linear nth_append nth_append_drop_is_nth_conv take_all us)
      ultimately have "(ss ! j, ?steps' ! j, ?us ! j) \<in> par_rstep2 R"
        using par_rstep2_refl by auto
    }
    ultimately
    have "(ss ! j, ?steps' ! j, ?us ! j) \<in> par_rstep2 R" by fastforce
  } note args = this
  have conc: "concat (mapi (\<lambda>i. map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>))) 0 ?steps') = [x]"
    apply (subst append_mapi)+
    apply (subst concat_append)+
    apply (subst mapi_map_replicate_trivial)+
    apply (subst concat_replicate_trivial)+
    using x by auto
  from par_rstep2.intros(2)[OF args us steps', unfolded conc]
  have first:"(Fun f ss, [x], Fun f ?us) \<in> par_rstep2 R" by auto
  let ?steps'' = "take i steps @ zs # drop (Suc i) steps"
  have steps'': "length ?us = length ?steps''" 
    by (metis ist length_list_update par_step_fun2(4) upd_conv_take_nth_drop)
  have ts: "length ?us = length ts"by (metis par_step_fun2(3) us)
  {
    fix j
    assume j: "j < length ts"
    have "(?us ! j, ?steps'' ! j, ts ! j) \<in> par_rstep2 R"
      by (metis ist j less_imp_le_nat nth_append_take nth_list_update_neq par_step_fun2.hyps(1) par_step_fun2.hyps(4) s' upd_conv_take_nth_drop)
  } note args = this
  let ?isteps = "mapi (\<lambda>i. map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>))) 0 (take i steps)"
  {
    fix xs
    assume "xs \<in> set ?isteps"
    then obtain j where j:"j < length (take i steps)" and xs: "?isteps ! j = xs"
      by (metis (erased, lifting) in_set_idx length_mapi)
    hence eq: "steps ! j = (take i steps) ! j" by auto
    from j have j2:"j < length steps" by auto
    from  xs eq have "?msteps ! j = xs"
      unfolding mapi_nth[OF j2, of _ 0] mapi_nth[OF j, of _ 0] by auto
    moreover have "j < i" using j by auto
    ultimately have "xs = []" using nil by blast
  }   
  hence *:"concat (mapi (\<lambda>i. map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>))) 0 (take i steps)) = []"
    by auto
  from ist have min:"(min (length steps) i) = i" by auto
  have conc: "concat (mapi (\<lambda>i. map (\<lambda>(lr, p, \<sigma>). (lr, i <# p, \<sigma>))) 0 ?steps'') = xs"
    apply (subst append_mapi)
    apply (subst concat_append)
    apply (subst *)
    apply (simp add: min)
    unfolding xs ys zs
    apply (subst mapi_drop[of "Suc i" _ 0 steps])
    by auto
  from par_rstep2.intros(2)[OF args ts steps'', unfolded conc]
  have "(Fun f ?us, xs, Fun f ts) \<in> par_rstep2 R" by auto
  with first show ?case by auto
qed auto

lemma par_rstep2_imp_seq:
  assumes "(t, steps, v) \<in> par_rstep2 R"
  shows "\<exists> ss \<in> seq R. first ss = t \<and> last ss = v \<and> length steps = length (snd ss) \<and> 
  (\<forall> i < length steps. steps ! i = (get_rule (snd ss ! i), get_pos (snd ss ! i), get_subst (snd ss ! i)))"
using assms
proof (induct steps arbitrary: t)
  case Nil hence eq:"t = v" using par_rstep2_empty by auto
  hence "(t,[]) \<in> seq R" (is "?ss") using seq.intros by auto
  thus ?case unfolding eq apply auto
    by (metis Decreasing_Diagrams2.last.simps(1) first.simps snd_conv)
next
  case (Cons x xs)
  then obtain t' where step: "(t, [x], t') \<in> par_rstep2 R" and steps: "(t', xs, v) \<in> par_rstep2 R"
    using par_rstep2_decompose by metis
  then obtain rl p \<sigma> where x_dec: "x = (rl,p,\<sigma>)" using prod_cases3 by blast
  with step have step: "(t, t') \<in> rstep_r_p_s R rl p \<sigma>"
    using par_rstep2_singleton by metis
  then obtain ss' where key: "ss' \<in> seq R \<and> first ss' = t' \<and> last ss' = v \<and> length xs = length (snd ss') \<and> 
    (\<forall>i<length xs. xs ! i = (get_rule (snd ss' ! i), get_pos (snd ss' ! i), get_subst (snd ss' ! i)))"
    using Cons steps by metis
  hence "(t, (t, rl, p, \<sigma>, t') # snd ss') \<in> seq R" (is "?ss \<in> _")
    using step seq.intros(2) by (metis first.elims snd_conv)
  moreover have "first ?ss = t" by auto
  moreover have "last ?ss = v" using key unfolding last.simps get_target_def
    apply auto by (metis first.simps surjective_pairing)
  moreover have "length (x # xs) = length (snd ?ss)" using key by auto
  moreover have "\<forall>i<length (x # xs). (x # xs) ! i = (get_rule (snd ?ss ! i), get_pos (snd ?ss ! i), get_subst (snd ?ss ! i))"
    apply (intro allI impI) apply (case_tac i)
    using key unfolding x_dec get_rule_def get_pos_def get_subst_def by auto
  ultimately show ?case by fast
qed

(* close variable peaks (left-linear case) *)
lemma variable_steps_left_lin:
  assumes xy:"(s, t) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1"
  and     xz:"(s, u) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2"
  and "p1 <#> q = p2"
  and "(v1 <#> v2) = q" and "v1 \<in> poss l1" and "l1 |_ v1 = Var w"
  and lin: "linear_term l1"
  defines "\<tau> \<equiv> (\<lambda>y. if y = w then replace_at (\<sigma>1 y) v2 (r2 \<cdot> \<sigma>2) else \<sigma>1 y)"
  defines "v \<equiv> (ctxt_of_pos_term p1 s)\<langle>r1 \<cdot> \<tau>\<rangle>"
  shows "(u, v) \<in> (rstep_r_p_s R (l1, r1) p1 \<tau>)" (is "?g1")
    "\<exists>steps. (\<forall> i < length steps. ((fst (steps!i)) = (l2,r2))) \<and> 
    (t,steps,v) \<in> par_rstep2 R" (is "?g2")
proof -
  from assms show ?g1 using variable_steps_lin(1)[OF xy xz] by fast
  (*begin*)
  show ?g2
  proof - 
    have k1: "p1<#>v1 \<in> poss s" using assms unfolding rstep_r_p_s_def Let_def by auto
    have k2: "(p1<#>v1)<#>v2 \<in> poss s" using assms unfolding rstep_r_p_s_def Let_def by auto 
    have s_dec1: "s = (ctxt_of_pos_term p1 s)\<langle>l1\<cdot>\<sigma>1\<rangle>" using xy unfolding rstep_r_p_s_def Let_def by auto
    hence "s|_p1 = l1\<cdot>\<sigma>1" by (metis (erased, lifting) mem_Collect_eq old.prod.case rstep_r_p_s_def subt_at_ctxt_of_pos_term xy)
    hence "(s|_p1)|_v1 = \<sigma>1(w)" using assms by (metis subst_apply_term.simps(1) subt_at_subst)
    hence "s|_(p1<#>v1) = \<sigma>1(w)" using k1 by (metis subterm_poss_conv)
    hence "(s|_(p1<#>v1))|_v2 = (\<sigma>1(w))|_v2" by metis
    hence "(s|_(p1<#>v1<#>v2)) = (\<sigma>1(w))|_v2"  using k2 by (metis Position.append_assoc k1 subt_at_append)
    hence "(s|_p2) = (\<sigma>1(w))|_v2"  using assms by auto    
    have s_dec2: "s = (ctxt_of_pos_term p2 s)\<langle>l2\<cdot>\<sigma>2\<rangle>" using xz unfolding rstep_r_p_s_def Let_def by auto
    hence "s|_p2 = l2\<cdot>\<sigma>2" by (metis (erased, lifting) mem_Collect_eq old.prod.case rstep_r_p_s_def subt_at_ctxt_of_pos_term xz)
    hence s1:"\<sigma>1(w) = replace_at (\<sigma>1 w) v2 (l2\<cdot>\<sigma>2)" using assms
      by (metis `s |_ p1 <#> v1 = \<sigma>1 w` `s |_ p2 = \<sigma>1 w |_ v2` ctxt_supt_id k2 subterm_poss_conv)
    have s2:"\<tau>(w)  = replace_at (\<sigma>1 w) v2 (r2\<cdot>\<sigma>2)" using assms by auto
    have "(l2\<cdot>\<sigma>2,r2\<cdot>\<sigma>2) \<in> rstep_r_p_s R (l2,r2) \<epsilon> \<sigma>2" using assms unfolding rstep_r_p_s_def Let_def by auto
    hence "(replace_at (\<sigma>1 w) v2 (l2\<cdot>\<sigma>2),replace_at (\<sigma>1 w) v2 (r2\<cdot>\<sigma>2)) \<in> rstep_r_p_s R (l2,r2) v2 \<sigma>2"
      unfolding rstep_r_p_s_def Let_def apply auto
      apply (metis `s |_ p1 <#> v1 = \<sigma>1 w` k2 poss_append_poss s1) apply (metis s1) by (metis s1)
    hence subst_step: "(\<sigma>1(w),\<tau>(w)) \<in> rstep_r_p_s R (l2,r2) v2 \<sigma>2" using s1 s2 by auto 
    have aux: "\<forall>x. (x = w) \<or> (\<sigma>1(x) = \<tau>(x))" using assms unfolding \<tau>_def by auto
    then obtain steps where A1: "\<forall>i < length steps. (fst (steps!i) = (l2,r2))" and B1: "(r1\<cdot>\<sigma>1,steps,r1\<cdot>\<tau>) \<in> par_rstep2 R"
      using subst_step_imp_par_rstep2[OF subst_step aux] by fast
  (*end*)
    have B2: "((ctxt_of_pos_term p1 s)\<langle>r1 \<cdot> \<sigma>1\<rangle>, map (\<lambda>(rl, p, \<sigma>). (rl, hole_pos (ctxt_of_pos_term p1 s) <#> p, \<sigma>)) steps, (ctxt_of_pos_term p1 s)\<langle>r1 \<cdot> \<tau>\<rangle>) \<in> par_rstep2 R" (is "(_,?steps,_) \<in> _")
      using par_rstep2_ctxt_closed[OF B1] by auto
    have A2: "\<forall>i < length ?steps. ((fst (?steps!i)) = (l2,r2))"
    proof (intro allI impI)
      fix i
      assume "i < length ?steps"
      hence l:"i < length steps" by auto
      hence "fst (steps!i) = (l2,r2)" using A1 by auto
      then obtain p1 \<sigma>1 where step_dec: "steps!i = ((l2,r2),p1,\<sigma>1)" by (cases "steps ! i", auto)
      have "(\<lambda>(rl, p, \<sigma>). (rl, hole_pos (ctxt_of_pos_term p1 s) <#> p, \<sigma>)) (steps ! i) = 
            ((l2, r2),hole_pos (ctxt_of_pos_term p1 s) <#>p1,\<sigma>1)"
        using step_dec by auto
      thus "fst (?steps!i) = (l2,r2)" unfolding nth_map[OF l] using step_dec by auto
    qed
    have t_dec: "t = (ctxt_of_pos_term p1 s)\<langle>r1\<cdot>\<sigma>1\<rangle>" using xy unfolding rstep_r_p_s_def Let_def by auto
    from A2 B2 show ?thesis unfolding t_dec v_def  by fast
  qed
qed

(*Section 3: decreasingness for TRSs *)
(*diagrams*)
definition ld_trs :: "('f,char list) trs \<Rightarrow> ('f,char list) lpeak \<Rightarrow> ('f,char list) seq \<Rightarrow> ('f,char list) seq \<Rightarrow> bool"
 where "ld_trs R p jl jr = (p \<in> local_peaks R \<and> jl \<in> seq R \<and> jr \<in> seq R \<and> get_target (fst p) = first jl \<and> get_target (snd p) = first jr \<and> last jl = last jr)"

definition diamond_trs :: "('f,char list) trs \<Rightarrow> ('f,char list) lpeak \<Rightarrow> ('f,char list) seq \<Rightarrow> ('f,char list) seq \<Rightarrow> bool"
 where "diamond_trs R p jl jr = (ld_trs R p jl jr \<and> length (snd jl) \<le> 1 \<and> length (snd jr) \<le> 1)"

definition ld2_trs :: "('f,char list) trs \<Rightarrow> ('f,char list) lpeak \<Rightarrow> ('f,char list) seq \<Rightarrow> ('f,char list) seq \<Rightarrow> bool"
 where "ld2_trs R p jl jr = (ld_trs R p jl jr \<and> (length (snd jr) \<le> 1 \<and> (length (snd jl) \<le> 1 \<or> (\<not> linear_term (snd (get_rule (fst p)))))))"
 
type_synonym 'a relp = "'a rel \<times> 'a rel"

definition ELD_1' :: "'a relp \<Rightarrow> 'a \<Rightarrow> 'a \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> 'a list \<Rightarrow> bool"
 where "ELD_1' r \<beta> \<alpha> \<sigma>1 \<sigma>2 \<sigma>3 = (set \<sigma>1 \<subseteq> ds (fst r) {\<beta>} \<and> length \<sigma>2 \<le> 1 \<and> set \<sigma>2 \<subseteq> ds (snd r) {\<alpha>} \<and> set \<sigma>3 \<subseteq> ds (fst r) {\<alpha>,\<beta>})"

definition eld_seq :: "(('f,'v,'a) labeling) \<Rightarrow> 'a relp \<Rightarrow> ('f,'v) lpeak \<Rightarrow> ('f,'v) seq \<Rightarrow> ('f,'v) seq \<Rightarrow> bool"
 where "eld_seq l r p jl jr = (\<exists> \<sigma>1 \<sigma>2 \<sigma>3 \<tau>1 \<tau>2 \<tau>3. ((map l (snd jl) = \<sigma>1@\<sigma>2@\<sigma>3) \<and> (map l (snd jr) = \<tau>1@\<tau>2@\<tau>3) \<and>
 (ELD_1' r (l (fst p)) (l (snd p)) \<sigma>1 \<sigma>2 \<sigma>3) \<and>
 (ELD_1' r (l (snd p)) (l (fst p)) \<tau>1 \<tau>2 \<tau>3)
))"

definition eld :: "('f,char list) trs \<Rightarrow> ('f,char list,'a) labeling \<Rightarrow> 'a relp \<Rightarrow> ('f,char list) lpeak \<Rightarrow> bool"
where "eld R l r p = (\<exists> jl jr. (ld_trs R p jl jr \<and> eld_seq l r p jl jr))"

(* abstract labelings *)
definition is_labeling :: "('f,char list) trs \<Rightarrow> ('f,char list,'a) labeling \<Rightarrow> 'a relp \<Rightarrow> bool"
where "is_labeling R l r = (\<forall> s rl1 t u rl2 v C \<rho> \<sigma>1 \<sigma>2 p1 p2.
  (s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1 \<and> (u,v) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<longrightarrow> (
  let st1 = (s,rl1,p1,\<sigma>1,t) in
  let st2 = (u,rl2,p2,\<sigma>2,v) in
    ((l st1, l st2) \<in> (fst r) \<longrightarrow> (l (step_close st1 C \<rho>), l (step_close st2 C \<rho>)) \<in> (fst r)) \<and>
    ((l st1, l st2) \<in> (snd r) \<longrightarrow> (l (step_close st1 C \<rho>), l (step_close st2 C \<rho>)) \<in> (snd r))))"

definition labels :: "('f,'v,'a) labeling \<Rightarrow> ('f,'v) seq \<Rightarrow> 'a set"
 where "labels l ss = set (map l (snd ss))"

definition weld_seq :: "('f,'v,'a) labeling \<Rightarrow> 'a relp \<Rightarrow> ('f,'v) lpeak \<Rightarrow> ('f,'v) seq \<Rightarrow> ('f,'v) seq \<Rightarrow> bool"
where "weld_seq l r p jl jr = (
  (\<forall> a \<in> labels l jl. a \<in> ds (snd r) {l (snd p)}) \<and> 
  (\<forall> b \<in> labels l jr. b \<in> ds (snd r) {l (fst p)}))"

definition weld :: "('f,char list) trs \<Rightarrow> ('f,char list,'a) labeling \<Rightarrow> 'a relp \<Rightarrow> ('f,char list) lpeak \<Rightarrow> bool"
 where "weld R l r p = (\<exists> jl jr. (ld2_trs R p jl jr \<and> weld_seq l r p jl jr))"

definition compatible :: "('f,char list) trs \<Rightarrow> ('f,char list,'a) labeling \<Rightarrow> 'a relp \<Rightarrow> bool"
where "compatible R l r = (is_labeling R l r \<and> 
  (\<forall> p. variable_peak R p \<or> parallel_peak R p \<longrightarrow> eld R l r p))"
 
definition weakly_compatible :: "('f,char list) trs \<Rightarrow> ('f,char list,'a) labeling \<Rightarrow> 'a relp \<Rightarrow> bool"
where "weakly_compatible R l r = (is_labeling R l r \<and> 
  (left_linear_trs R \<longrightarrow> (\<forall> p. variable_peak R p \<or> parallel_peak R p \<longrightarrow> weld R l r p)))"

lemma linear_and_weld_imp_eld:
  assumes "linear_trs R" and "weld R l r p"
  shows "eld R l r p"
proof - 
  from assms obtain jl jr where ld2_trs: "ld2_trs R p jl jr" and weld_seq: "weld_seq l r p jl jr"
    unfolding weld_def by auto
  then obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u where
    p_dec: "p = ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u))" and lp: "p \<in> local_peaks R" 
    unfolding ld2_trs_def ld_trs_def local_peaks_def by fast
  hence ld: "ld_trs R p jl jr" using ld2_trs unfolding ld2_trs_def by auto
  from lp have "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" unfolding local_peaks_def p_dec by auto 
  hence "rl1 \<in> R" unfolding rstep_r_p_s_def apply auto by metis
  hence "linear_term (snd (get_rule (fst p)))" unfolding p_dec get_rule_def apply auto using assms unfolding linear_trs_def by (metis assms(1) linear_trsE prod.collapse)
  hence jl: "length (snd jl) \<le> 1"  using ld2_trs unfolding weld_def ld_trs_def ld2_trs_def weld_seq_def eld_seq_def linear_trs_def  using assms(1) by auto
  have jr: "length (snd jr) \<le> 1" using ld2_trs unfolding ld2_trs_def by auto
  have "ELD_1' r (l (fst p)) (l (snd p)) [] (map l (snd jl))  []" using weld_seq jl unfolding ELD_1'_def ds_def apply auto unfolding weld_seq_def labels_def ds_def by auto 
  moreover have "ELD_1' r (l (snd p)) (l (fst p)) [] (map l (snd jr))  []" using weld_seq jr unfolding ELD_1'_def ds_def apply auto unfolding weld_seq_def labels_def ds_def by auto
  ultimately have "eld_seq l r p jl jr" unfolding eld_seq_def by (metis append_Nil append_Nil2)
  thus ?thesis using ld unfolding eld_def by fast
qed

lemma linear_and_weakly_compatible_is_compatible:
 assumes "linear_trs R" and "weakly_compatible R l r"
 shows "compatible R l r" proof -
  from assms have ll: "left_linear_trs R" unfolding linear_trs_def left_linear_trs_def by auto
  have 1: "is_labeling R l r" using assms unfolding weakly_compatible_def by auto
  {
    fix p assume A: "variable_peak R p \<or> parallel_peak R p" 
    hence " weld R l r p" using assms(2) ll unfolding weakly_compatible_def by metis 
    hence "eld R l r p" using linear_and_weld_imp_eld[OF assms(1)] by fast
  }
  thus ?thesis using 1 unfolding compatible_def by auto
qed
  
lemma ld_trs_closed:
  fixes C \<tau>
  assumes "ld_trs R p jl jr" 
  shows "ld_trs R ((step_close (fst p) C \<tau>),(step_close (snd p) C \<tau>)) (seq_close jl C \<tau>) (seq_close jr C \<tau>)" (is "ld_trs R ?P _ _") 
proof -
  from assms obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u where p_dec: "p = ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u))" and "p \<in> local_peaks R" 
    unfolding ld_trs_def local_peaks_def by fast
  hence l:"(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" and r:"(s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2" unfolding local_peaks_def by auto
  from l have P1: "(C\<langle>s\<cdot>\<tau>\<rangle>,C\<langle>t\<cdot>\<tau>\<rangle>) \<in> rstep_r_p_s R rl1 (hole_pos C<#> p1) (\<sigma>1\<circ>s \<tau>)" 
    using rstep_r_p_s_ctxt_closed2 rstep_r_p_s_subst_closed by metis
  from r have P2: "(C\<langle>s\<cdot>\<tau>\<rangle>,C\<langle>u\<cdot>\<tau>\<rangle>) \<in> rstep_r_p_s R rl2 (hole_pos C<#> p2) (\<sigma>2\<circ>s \<tau>)" 
    using rstep_r_p_s_ctxt_closed2 rstep_r_p_s_subst_closed by metis
  from P1 P2 have P: "?P \<in> local_peaks R" unfolding local_peaks_def p_dec step_close.simps fst_conv snd_conv Let_def by fast
  from assms have jl:"jl \<in> seq R" and jr:"jr \<in> seq R" and l: "get_target (fst p) = first jl" 
    and r: "get_target (snd p) = first jr" and b: "last jl = last jr" unfolding ld_trs_def by auto
  have JL: "seq_close jl C \<tau> \<in>  seq R" using seq_super_closed1[OF jl] by fast
  have JR: "seq_close jr C \<tau> \<in>  seq R" using seq_super_closed1[OF jr] by fast
  have L: "get_target (step_close (fst (p)) C \<tau>) = first (seq_close jl C \<tau>)" 
    using step_super_closed1 l seq_super_closed1[OF jl] by metis
  have R: "get_target (step_close (snd (p)) C \<tau>) = first (seq_close jr C \<tau>)" 
    using step_super_closed1 r seq_super_closed1[OF jr] by metis 
  have "last (seq_close jl C \<tau>) = last (seq_close jr C \<tau>)" using seq_super_closed1 jl jr b by metis 
  thus ?thesis using P JL JR L R unfolding ld_trs_def by auto
qed

lemma el_closed:
  fixes C \<tau>
  assumes lab: "is_labeling R l r"
  and s1: "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1"
  and ss: "ss \<in> seq R"
  and dec: "\<forall> b \<in> set (map l (snd ss)). (b,l (s,rl1,p1,\<sigma>1,t)) \<in> (snd r)"
  shows "\<forall> b \<in> set (map l (snd (seq_close ss C \<tau>))). (b,l (step_close (s,rl1,p1,\<sigma>1,t) C \<tau>)) \<in> (snd r)" 
using ss dec
proof (induct "snd ss" arbitrary:"ss")
 case Nil thus ?case by auto
next
 case (Cons step steps) 
 from Cons(2) obtain u where ss_dec: "ss = (u,step # steps)" by (cases ss, auto)
 with Cons(3) obtain rl2 p2 \<sigma>2 v where step_dec: "step = (u,rl2,p2,\<sigma>2,v)" and s2: "(u,v) \<in> rstep_r_p_s R rl2 p2 \<sigma>2" 
   using seq.cases by blast
 have "(l step, l (s,rl1,p1,\<sigma>1,t)) \<in> (snd r)" using Cons unfolding ss_dec step_dec by auto
 hence step_close: "(l (step_close step C \<tau>), l (step_close (s,rl1,p1,\<sigma>1,t) C \<tau>)) \<in> snd r" 
   using lab s1 s2 unfolding is_labeling_def step_dec Let_def by blast
 hence vs: "(v,steps) \<in> seq R" using Cons unfolding ss_dec step_dec using seq_chop_first by fast
 hence "\<forall>b\<in>set (map l (snd (v,steps))). (b, l (s, rl1, p1, \<sigma>1, t)) \<in> snd r" using Cons unfolding ss_dec step_dec by auto
 hence "\<forall> b \<in> set (map l (snd (seq_close (v,steps) C \<tau>))). (b,l (step_close (s,rl1,p1,\<sigma>1,t) C \<tau>)) \<in> snd r" 
   using Cons(1)[OF _ vs] Cons by auto
 thus ?case using step_close unfolding ss_dec by auto
qed

lemma eseq_super_closed2:
fixes C \<sigma>
assumes step': "(s',t') \<in> rstep_r_p_s R rl' p' \<rho>"
and lab: "is_labeling R l r"
and seq: "ss \<in> seq R"
and dec:"\<forall> b \<in> set (map l (snd ss)). (b,l (s',rl',p',\<rho>,t')) \<in> (fst r)"
shows "(\<forall> b \<in> set (map l (snd (seq_close ss C \<sigma>))). (b,l (step_close (s',rl',p',\<rho>,t') C \<sigma>)) \<in> (fst r))" 
 using seq dec proof (induct "snd ss" arbitrary: ss)
 case Nil then obtain s where ss_dec: "ss = (s,[])" using surjective_pairing by metis 
 thus ?case using seq.intros unfolding ss_dec by auto
next
 case (Cons step ts)
 from Cons(2) obtain s where ss_dec: "ss=(s,step#ts)" by (cases ss, auto)
 with Cons(3) obtain rl p \<tau> t where step_dec: "step=(s,rl,p,\<tau>,t)"  using seq.cases by blast
 hence step: "(s,t) \<in> rstep_r_p_s R rl p \<tau>" and i:"(t,ts) \<in> seq R" (is "?ts \<in> _") using Cons(3) seq.cases unfolding ss_dec step_dec by auto
 have j:"\<forall>b \<in> set (map l (snd ?ts)). (b,l (s',rl',p',\<rho>,t')) \<in> fst r" using Cons unfolding ss_dec snd_conv by auto 
 from Cons(1)[OF _ i j] have
 ih: "(\<forall> b \<in> set (map l (snd (seq_close ?ts C \<sigma>))). (b,l (C\<langle>s'\<cdot>\<sigma>\<rangle>,rl',hole_pos C<#>p',\<rho>\<circ>s \<sigma>,C\<langle>t'\<cdot>\<sigma>\<rangle>)) \<in> fst r)" by auto
 have C_step: "(C\<langle>s\<cdot>\<sigma>\<rangle>,C\<langle>t\<cdot>\<sigma>\<rangle>) \<in> rstep_r_p_s R rl (hole_pos C<#>p) (\<tau>\<circ>s \<sigma>)" using rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step]] by auto 
 have C_label: "(l (s,rl,p,\<tau>,t),l (s',rl',p',\<rho>,t')) \<in> fst r" using Cons unfolding ss_dec step_dec snd_conv  by auto
 have C_label2: "(l (step_close (s,rl,p,\<tau>,t) C \<sigma>),l (step_close (s',rl',p',\<rho>,t') C \<sigma>)) \<in> fst r" 
  using step step' C_label lab dec  unfolding is_labeling_def Let_def by fast
 hence C_labels: "(\<forall>b\<in>set (map l (snd (seq_close ss C \<sigma>))). (b, l (C\<langle>s' \<cdot> \<sigma>\<rangle>, rl', hole_pos C <#> p', \<rho> \<circ>s \<sigma>, C\<langle>t' \<cdot> \<sigma>\<rangle>)) \<in> fst r)" 
  using ih(1) unfolding ss_dec step_dec seq_close.simps snd_conv step_close.simps Let_def by auto  
 show ?case using C_labels by auto
qed

lemma eseq_super_closed3:
fixes C \<sigma>
assumes step': "(s',t') \<in> rstep_r_p_s R rl' p' \<rho>"
and step'': "(s'',t'') \<in> rstep_r_p_s R rl'' p'' \<rho>'"
and lab: "is_labeling R l r"
and seq: "ss \<in> seq R"
and dec:"\<forall> b \<in> set (map l (snd ss)). ((b,l (s',rl',p',\<rho>,t')) \<in> fst r \<or> (b,l (s'',rl'',p'',\<rho>',t'')) \<in> fst r)"
shows "(\<forall> b \<in> set (map l (snd (seq_close ss C \<sigma>))). ((b,l (step_close (s',rl',p',\<rho>,t') C \<sigma>)) \<in> fst r) \<or> (b,l (step_close (s'',rl'',p'',\<rho>',t'') C \<sigma>)) \<in> fst r)"
 using seq dec proof (induct "snd ss" arbitrary: ss)
 case Nil then obtain s where ss_dec: "ss = (s,[])" using surjective_pairing by metis 
 thus ?case using seq.intros unfolding ss_dec by auto
next
 case (Cons step ts)
 from Cons(2) obtain s where ss_dec: "ss=(s,step#ts)" by (cases ss, auto)
 with Cons(3) obtain rl p \<tau> t where step_dec: "step=(s,rl,p,\<tau>,t)"
   using seq.cases by blast
 hence step: "(s,t) \<in> rstep_r_p_s R rl p \<tau>"  and i:"(t,ts) \<in> seq R" (is "?ts \<in> _") using Cons(3) seq.cases unfolding step_dec ss_dec by auto
 have j:"\<forall>b \<in> set (map l (snd ?ts)). ((b, l (s',rl',p',\<rho>,t')) \<in> fst r \<or> (b,l (s'',rl'',p'',\<rho>',t'')) \<in> fst r)" using Cons unfolding ss_dec snd_conv by auto 
 from Cons(1)[OF _ i j] have
 ih: "(\<forall> b \<in> set (map l (snd (seq_close ?ts C \<sigma>))). ((b,l (step_close (s',rl',p',\<rho>,t') C \<sigma>)) \<in> fst r) \<or> (b,l (step_close (s'',rl'',p'',\<rho>',t'') C \<sigma>)) \<in> fst r)" (is "\<forall> b \<in> _. (b,?a') \<in> ?r \<or> (b,?a'') \<in> ?r") by auto
 have C_step: "(C\<langle>s\<cdot>\<sigma>\<rangle>,C\<langle>t\<cdot>\<sigma>\<rangle>) \<in> rstep_r_p_s R rl (hole_pos C<#>p) (\<tau>\<circ>s \<sigma>)" using rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step]] by auto 
 have C_label: "(l (s,rl,p,\<tau>,t),l (s',rl',p',\<rho>,t')) \<in> fst r \<or> (l (s,rl,p,\<tau>,t),l (s'',rl'',p'',\<rho>',t'')) \<in> fst r" using Cons(4) unfolding ss_dec step_dec snd_conv by auto
 have C_label2: "(l (step_close (s,rl,p,\<tau>,t) C \<sigma>),?a') \<in> fst r \<or> (l (step_close (s,rl,p,\<tau>,t) C \<sigma>),?a'') \<in> fst r" (is "(?a,_) \<in> fst r \<or> _") proof(cases "(l (s,rl,p,\<tau>,t),l (s',rl',p',\<rho>,t')) \<in> ?r")
  case True thus ?thesis using step step' lab dec unfolding is_labeling_def Let_def by fast 
  next
  case False hence l: "(l (s,rl,p,\<tau>,t),l (s'',rl'',p'',\<rho>',t'')) \<in> fst r" using C_label by auto
  thus ?thesis using step step'' C_step rstep_r_p_s_ctxt_closed2[OF rstep_r_p_s_subst_closed[OF step''],of "C" "\<sigma>"] l lab dec unfolding is_labeling_def Let_def by fast
 qed
 hence C_labels: "(\<forall>b\<in>set (map l (snd (seq_close ss C \<sigma>))). ((b,?a') \<in> fst r \<or> (b,?a'') \<in> fst r))" 
  using ih unfolding ss_dec step_dec seq_close.simps snd_conv step_close.simps Let_def by auto  
 show ?case using C_labels by auto
 qed

lemma eld_seq_closed: 
  fixes C \<tau> 
  assumes d: "ld_trs R p jl jr" 
  and "eld_seq l r p jl jr" 
  and lab: "is_labeling R l r"
  shows "eld_seq l r ((step_close (fst p) C \<tau>),(step_close (snd p) C \<tau>)) (seq_close jl C \<tau>) (seq_close jr C \<tau>)" 
    (is "eld_seq _ _ ?P ?JL ?JR") 
proof -
  {
    fix s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u l1 l2 l3 jl
    let ?a = "l (s,rl1,p1,\<sigma>1,t)"
    let ?a' = "l (s,rl2,p2,\<sigma>2,u)"
    let ?A = "l (step_close (s, rl1, p1, \<sigma>1, t) C \<tau>)"
    let ?A' = "l (step_close (s, rl2, p2, \<sigma>2, u) C \<tau>)"
    assume
      lstep: "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" and
      lstep': "(s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2" and
      jl_dec2: "map l (snd jl) = (l1@l2)@l3" and
      ELD_l: "ELD_1' r ?a ?a' l1 l2 l3" and
      jl:"jl \<in> seq R"
    obtain L1 L2 L3 where jl_dec: "jl = seq_concat (seq_concat L1 L2) L3"
      and  L1:"L1 \<in> seq R" and L2: "L2 \<in> seq R" and  L3: "L3 \<in> seq R"
      and  L1_l1: "map l (snd L1) = l1" and L2_l2: "map l (snd L2) = l2" and L3_l3:"map l (snd L3) = l3"
      using seq_dec[OF jl jl_dec2] seq_dec by blast
    have "length (snd L2) \<le> 1" using ELD_l L2_l2 unfolding ELD_1'_def by auto
    hence JL: "length (snd (seq_close L2 C \<tau>)) \<le> 1" using seq_close_length by auto
    from ELD_l have
      ELD1: "\<forall> b \<in> set (map l (snd L1)). (b, ?a) \<in> fst r" and
      ELD2: "\<forall> b \<in> set (map l (snd L2)). (b, ?a') \<in> snd r" and
      ELD3: "\<forall> b \<in> set (map l (snd L3)). ((b, ?a) \<in> fst r \<or> (b, ?a') \<in> fst r)"
      unfolding ELD_1'_def ds_def L1_l1 L2_l2 L3_l3 by auto
    let ?sc = "\<lambda>L. seq_close L C \<tau>"
    have L1:"\<forall> b \<in> set (map l (snd (?sc L1))). (b, ?A) \<in> fst r"
      using eseq_super_closed2[OF lstep lab L1] ELD1 unfolding fst_conv by auto
    have  L2:"\<forall> b \<in> set (map l (snd (?sc L2))). (b,?A') \<in> snd r"
      using el_closed[OF lab _ L2] lstep' ELD2 unfolding snd_conv by auto
    have  L3:"\<forall> b \<in> set (map l (snd (?sc L3))). ((b, ?A) \<in> fst r \<or> (b, ?A') \<in> fst r)"
      using eseq_super_closed3[OF lstep lstep' lab L3] ELD3 unfolding fst_conv by simp
    hence L: "ELD_1' r ?A ?A' (map l (snd (?sc L1))) (map l (snd (?sc L2))) (map l (snd (?sc L3)))"
      unfolding ELD_1'_def ds_def using L1 L2 L3 JL by auto
    have "seq_close jl C \<tau> = seq_concat (seq_concat (?sc L1) (?sc L2)) (?sc L3)"
      using seq_concat_seq_close jl_dec by metis
    hence  L2:"map l (snd (?sc jl)) = (map l (snd (?sc L1))) @ (map l (snd (?sc L2))) @ (map l (snd (?sc L3)))"
      using seq_concat_labels map_append List.append_assoc by metis
    with L have "\<exists> L1 L2 L3. ELD_1' r ?A ?A' (map l (snd (?sc L1))) (map l (snd (?sc L2))) (map l (snd (?sc L3))) \<and>
      map l (snd (seq_close jl C \<tau>)) = (map l (snd (?sc L1))) @ (map l (snd (?sc L2))) @ (map l (snd (?sc L3)))" by auto
  } note * = this
  from assms have pl:"p \<in> local_peaks R" unfolding ld_trs_def by auto
  then obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u where p: "p = ((s, rl1, p1, \<sigma>1, t),(s, rl2, p2, \<sigma>2, u))"
    unfolding local_peaks_def by blast
  have lstep: "(s, t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" using pl unfolding local_peaks_def p by auto
  have lstep': "(s, u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2" using pl unfolding local_peaks_def p by auto
  let ?a = "l (s, rl1, p1, \<sigma>1, t)"
  let ?a' = "l (s, rl2, p2, \<sigma>2, u)"
  from assms obtain l1 l2 l3 where jl_dec2: "map l (snd jl) = (l1@l2)@l3" and LD_l: "ELD_1' r ?a ?a' l1 l2 l3"
    unfolding eld_seq_def unfolding p by auto
  have jl:"jl \<in> seq R" using d unfolding ld_trs_def by auto
  from assms obtain r1 r2 r3 where jr_dec2: "map l (snd jr) = (r1@r2)@r3" and LD_r: "ELD_1' r ?a' ?a r1 r2 r3"
    unfolding eld_seq_def unfolding p by auto
  have jr:"jr \<in> seq R" using d unfolding ld_trs_def by auto
  from *[OF lstep lstep' jl_dec2 LD_l jl] *[OF lstep' lstep jr_dec2 LD_r jr] show ?thesis unfolding eld_seq_def p by fastforce
qed

lemma eld_closed:
  fixes C \<tau>
  assumes "eld R l r p"
  and lab: "is_labeling R l r"
  shows "eld R l r (step_close (fst p) C \<tau>,step_close (snd p) C \<tau>)" 
proof -
  from assms(1) obtain jl jr where d: "ld_trs R p jl jr" and D: "eld_seq l r p jl jr" unfolding eld_def by auto
  from d have C_d:"ld_trs R (step_close (fst p) C \<tau>,step_close (snd p) C \<tau>) (seq_close jl C \<tau>) (seq_close jr C \<tau>)" 
    (is "ld_trs R ?p ?jl ?jr") using ld_trs_closed by fast 
  hence "eld_seq l r ?p ?jl ?jr" using eld_seq_closed[OF d D]  lab by fast
  thus ?thesis using C_d unfolding eld_def  by auto 
qed

lemma critical:
  assumes "function_peak R p"
  and cp:"\<forall> (b, p) \<in> critical_peaks R. eld R l r p"
  and lab: "is_labeling R l r"
  shows "eld R l r p" 
proof -
  from function_peak_inst_of_critical_peak assms(1) obtain C \<delta> b cp where
    "(b, cp) \<in> critical_peaks R" and
    inst: "p = (step_close (fst cp) C \<delta>, step_close (snd cp) C \<delta>)"
    by blast
  hence eld: "eld R l r cp" using cp by auto
  thus ?thesis using eld_closed[OF eld] trans eld lab inst by fast
qed

lemma left:
  assumes "variable_peak R p \<or> function_peak R p"
  and cps: "\<forall> (b, p) \<in> critical_peaks R. eld R l r p"
  and lab: "compatible R l r"
  shows "eld R l r p" 
proof (cases "variable_peak R p")
  case True thus ?thesis using lab unfolding compatible_def by fast 
next
  case False hence "function_peak R p" using assms by auto
  thus ?thesis using critical lab cps unfolding eld_def compatible_def by fast
qed

(*
lemma mirror_eld_seq: assumes "eld_seq l r (fst p, snd p) jl jr"
shows "eld_seq l r (snd p, fst p) jr jl" proof -
 from assms obtain \<sigma>1 \<sigma>2 \<sigma>3 \<tau>1 \<tau>2 \<tau>3 where "map l (snd jl) = \<sigma>1 @ \<sigma>2 @ \<sigma>3 \<and> map l (snd jr) = \<tau>1 @ \<tau>2 @ \<tau>3 \<and> ELD_1' r (l (fst (fst p, snd p))) (l (snd (fst p, snd p))) \<sigma>1 \<sigma>2 \<sigma>3 \<and> ELD_1' r (l (snd (fst p, snd p))) (l (fst (fst p, snd p))) \<tau>1 \<tau>2 \<tau>3"
 unfolding eld_seq_def by auto
 hence "map l (snd jr) = \<tau>1 @ \<tau>2 @ \<tau>3 \<and> map l (snd jl) = \<sigma>1 @ \<sigma>2 @ \<sigma>3 \<and> ELD_1' r (l (fst (snd p, fst p))) (l (snd (snd p, fst p))) \<tau>1 \<tau>2 \<tau>3 \<and> ELD_1' r (l (snd (snd p, fst p))) (l (fst (snd p, fst p))) \<sigma>1 \<sigma>2 \<sigma>3" by auto
 thus ?thesis unfolding eld_seq_def by fast
qed
*)

lemma eld_mirror: assumes "eld R l r p" shows "eld R l r (snd p,fst p)" proof -
  from assms obtain jl jr where d: "ld_trs R p jl jr" and D: "eld_seq l r p jl jr" unfolding eld_def by auto 
  from d have  d: "ld_trs R (snd p,fst p) jr jl" unfolding ld_trs_def local_peak_def by (metis fst_conv mirror snd_conv) 
  from D have D: "eld_seq l r (snd p,fst p) jr jl" unfolding eld_seq_def apply auto by metis  
  show ?thesis using d D unfolding eld_def by fast
qed

lemma main_eld:
  assumes cps:"\<forall> (b, p) \<in> critical_peaks R. eld R l r p"
  and lab: "compatible R l r" 
  shows "\<forall> p \<in> local_peaks R. eld R l r p" 
proof
  fix p 
  assume A: "p \<in> local_peaks R" 
  show "eld R l r p" 
  proof -
    show ?thesis using local_peaks_cases[OF A]
    proof (cases "parallel_peak R p")
      case True 
      show ?thesis using True lab unfolding compatible_def by fast
    next
      case False note po = this
      thus ?thesis using local_peaks_cases[OF A] 
      proof (cases "variable_peak R p \<or> function_peak R p")
        case True 
        show ?thesis using left[OF True cps lab] by auto
      next
        case False 
        hence F: "variable_peak R (snd p,fst p) \<or> function_peak R (snd p,fst p)" 
          using local_peaks_cases[OF A] po by auto
        hence "eld R l r (snd p, fst p)" using left[OF F cps lab] by auto
        thus ?thesis using eld_mirror by (metis fst_conv snd_conv surjective_pairing)
      qed
    qed
  qed
qed

lemma trivial_peak_eld:
  assumes eq: "t1 = t2"
  and "((s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2)) \<in> local_peaks R"
  shows "eld R lab r ((s1,r1,p1,\<sigma>1,t1),(s2,r2,p2,\<sigma>2,t2))" (is "eld _ _ _ ?p")
unfolding eld_def
proof (intro exI conjI)
  show "eld_seq lab r ?p (t1, []) (t2, [])" unfolding eld_seq_def ELD_1'_def  by auto
  show "ld_trs R ?p (t1, []) (t2, [])"
    using assms seq.intros(1) unfolding get_target_def ld_trs_def by auto
qed

(* connect TRS with labeled ARS *)
lemma rstep_is_rstep_r_p_s: "rstep R = (\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)"  proof 
 show "rstep R \<subseteq> (\<Union>(rl, p, \<sigma>). rstep_r_p_s R rl p \<sigma>)" proof 
  fix s t assume A:"(s, t) \<in> rstep R" show "(s,t) \<in> (\<Union>(rl, p, \<sigma>). rstep_r_p_s R rl p \<sigma>)" proof -
  obtain rl p \<sigma> where "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" using A rstep_iff_rstep_r_p_s by metis
  thus ?thesis by blast
  qed
 qed
 next 
 show "(\<Union>(rl, p, \<sigma>). rstep_r_p_s R rl p \<sigma>) \<subseteq> rstep R" proof 
  fix s t assume A: "(s,t) \<in> (\<Union>(rl, p, \<sigma>). rstep_r_p_s R rl p \<sigma>)" show "(s,t) \<in> rstep R" proof -
  from A obtain l r p \<sigma> where "(s, t) \<in> rstep_r_p_s R (l, r) p \<sigma>" by auto
  thus ?thesis using rstep_iff_rstep_r_p_s by metis
  qed
 qed
qed

lemma rstep_is_rstep_r_p_s_ast: "(\<Union> (rl,p,\<sigma>). rstep_r_p_s R rl p \<sigma>)^* = (rstep R)^*" using rstep_is_rstep_r_p_s by metis

lemma  local_peak_decompose:
  assumes "local_peak lars P"
  shows "\<exists>s t u a b. (s,a,t) \<in> lars \<and> (s,b,u) \<in> lars \<and> P = ((s,[(a,t)]),(s,[(b,u)]))" 
proof -
  from assms obtain seql seqr where P: "P = (seql,seqr)" and fsteq:"fst seql = fst seqr" 
    and lenl:"length (snd seql) = 1" and lenr:"length (snd seqr) = 1" 
    and seql: "seql \<in> DD_seq lars" and seqr: "seqr \<in> DD_seq lars"
    using assms unfolding local_peak_def peak_def by auto
  from lenl lenr obtain sl sr where "snd seql = [sl]" and "snd seqr = [sr]"
    using length_Suc_conv[of "snd seql" 0] length_Suc_conv[of "snd seqr" 0] length_0_conv by auto
  moreover from seql obtain s where "s = fst seql" using seq.cases by auto
  ultimately obtain a t b u where 
    seql_dec: "seql = (s,[(a,t)])" and
    seqr_dec: "seqr = (s,[(b,u)])" 
    using surjective_pairing fsteq by metis
  have "(s,a,t) \<in> lars" using seql unfolding seql_dec using seq_tail1(2)[of s "(a,t)" "[]"] by auto
  moreover have "(s,b,u) \<in> lars" using seqr unfolding seqr_dec using seq_tail1(2)[of s "(b,u)" "[]"] by auto
  ultimately show ?thesis using P seql_dec seqr_dec by auto
qed

(*labeled steps*)
definition rlstep_r_p_s :: "('f,'v) trs \<Rightarrow> ('f,'v,'a) labeling \<Rightarrow> ('f,'v) rule \<Rightarrow> pos \<Rightarrow> ('f,'v) subst \<Rightarrow> (('f,'v) term\<times>'a\<times>('f,'v) term) set"
 where "rlstep_r_p_s R l rl p \<sigma> = {(s,a,t). (s,t) \<in> rstep_r_p_s R rl p \<sigma> \<and> a = l (s,rl,p,\<sigma>,t)}"

abbreviation label_trs :: "('f,'v) trs \<Rightarrow> ('f,'v,'a) labeling \<Rightarrow> (('f,'v)term \<times> 'a \<times> ('f,'v)term) set" where
 "label_trs R l \<equiv> \<Union> (rl,p,\<sigma>) \<in> UNIV. rlstep_r_p_s R l rl p \<sigma>"

lemma local_peak_introduce:
  assumes "(s,a,t) \<in> rlstep_r_p_s R l rl1 p1 \<sigma>1" and "(s,b,u) \<in> rlstep_r_p_s R l rl2 p2 \<sigma>2"
  shows "((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u)) \<in> local_peaks R" and "a = l (s,rl1,p1,\<sigma>1,t)" and "b = l (s,rl2,p2,\<sigma>2,u)"
using assms unfolding local_peaks_def rlstep_r_p_s_def apply auto using surjective_pairing by metis

lemma R_step_imp_lars_step:
  assumes "(s,t) \<in> rstep_r_p_s R rl p \<sigma>"
  shows "(s,l (s,rl,p,\<sigma>,t),t) \<in> label_trs R l"
using assms unfolding rlstep_r_p_s_def by blast

lemma R_seq_imp_lars_seq:
  assumes "(s,ss) \<in> seq R"
  shows "\<exists> as \<in> DD_seq (label_trs R l). (
    s = fst as \<and> 
    last (s,ss) = lst as \<and>
    Decreasing_Diagrams.labels as = (map l ss))" 
using assms 
proof (induct ss arbitrary:s rule:list.induct)  
  case Nil hence eq: "s = last (s,[])" by auto
  hence ss: "(s, []) \<in> DD_seq (label_trs R l)" (is "?as \<in> _") using Decreasing_Diagrams.seq.intros by auto
  hence "s = fst ?as" and "last (s,[]) = lst ?as" and "Decreasing_Diagrams.labels ?as = (map l [])" 
    using eq apply (metis fst_conv) apply (metis eq fst_conv lst_def snd_conv) unfolding Decreasing_Diagrams.labels_def by auto
  thus ?case using ss by blast
next
  case (Cons step ts)
  from Cons.prems  obtain rl p \<sigma> t where step_dec: "step = (s,rl,p,\<sigma>,t)" by (cases, auto)
  hence ts: "(t,ts) \<in> seq R" using Cons(2) unfolding step_dec using seq_chop_first(1) by metis
  hence  step: "(s,t) \<in> rstep_r_p_s R rl p \<sigma>" using Cons(2) unfolding step_dec using seq_chop_first(2) by fast
  then obtain a where lstep:"(s,a,t) \<in> rlstep_r_p_s R l rl p \<sigma>" unfolding rlstep_r_p_s_def by auto
  hence a:"a = l step" unfolding step_dec unfolding rlstep_r_p_s_def by auto
  from lstep have "(s,a,t) \<in> label_trs R l" unfolding rlstep_r_p_s_def by blast
  hence ss1:"(s,[(a,t)]) \<in> DD_seq (label_trs R l)" (is "?ss1 \<in> _")
    unfolding rlstep_r_p_s_def using Decreasing_Diagrams.seq.intros(1) Decreasing_Diagrams.seq.intros(2) by smt2
  from ts obtain bs where ss2:"bs \<in> DD_seq (label_trs R l)" and "t = fst bs" and "last (t,ts) = lst bs" 
    and lab0:"Decreasing_Diagrams.labels bs = map l ts" using Cons by metis
  hence as:"(s,(a,t) # snd bs) \<in> DD_seq (label_trs R l)" (is "?as \<in> _") 
    using Decreasing_Diagrams.seq_concat[OF ss1 ss2]
    by (metis Decreasing_Diagrams.seq.intros(2) `(s, a, t) \<in> label_trs R l` fst_conv snd_conv surj_pair)
  have fst:"s = fst ?as" and lst:"last (s,step#ts) = lst ?as" 
    apply auto unfolding step_dec get_target_def lst_def
    by (metis List.last.simps `t = fst bs` `last (t, ts) = lst bs` lst_def neq_Nil_conv snd_conv)
  have lab: "Decreasing_Diagrams.labels ?as = map l (step#ts)" using lab0 a unfolding Decreasing_Diagrams.labels_def by auto 
  thus ?case using as fst lst lab by metis 
qed

lemma R_seq_imp_lars_seq2:
  assumes "ss \<in> seq R"
  shows "\<exists> as \<in> DD_seq (label_trs R l). (
   first ss = fst as \<and> 
   last ss = lst as \<and>
   DD_labels as = (map l (snd ss)))" 
proof-
  from assms obtain t ts where ss_dec: "ss = (t,ts)" using surjective_pairing by metis 
  show ?thesis using assms R_seq_imp_lars_seq[of t ts] unfolding ss_dec first.simps last.simps snd_conv by auto
qed

lemma seq_imp_ast: assumes "ss \<in> seq R" and "set (map l (snd ss)) \<subseteq> ds r S" 
defines "rel \<equiv> \<lambda> c. {(s,t). \<exists> r p \<sigma>. (s,t) \<in> rstep_r_p_s R r p \<sigma> \<and> l (s,r,p,\<sigma>,t) = c}"
shows "(first ss,last ss) \<in> (\<Union>c \<in> ds r S. rel c)^*" proof -
 from assms obtain t ts where ss_dec: "ss = (t,ts)" by (metis surjective_pairing)
 from assms  show ?thesis unfolding ss_dec proof (induct ts arbitrary:t rule:list.induct)
 case Nil hence "first (t,[]) = last (t,[])" by auto 
 thus ?case by auto
next
 case (Cons x xs) 
  from Cons(2) obtain rl p \<sigma> u where  x_dec: "x = (t,rl,p,\<sigma>,u)" by (cases, auto)
  have step: "(t,u) \<in> rstep_r_p_s R rl p \<sigma>" using seq_chop_first Cons(2) unfolding x_dec by fast
  have seq: "(u,xs) \<in> seq R"  using seq_chop_first Cons(2) unfolding x_dec by fast
  have lab: "l (t,rl,p,\<sigma>,u) \<in> ds r S" using step Cons(3) x_dec by auto
  hence step: "(t,u) \<in> (UNION (ds r S) rel)" using step unfolding rel_def by fast 
  from Cons(1)[OF seq] Cons(3) have steps:"(first (u, xs), Decreasing_Diagrams2.last (u, xs)) \<in> (UNION (r \<down>s S) rel)\<^sup>*" unfolding x_dec rel_def by auto
  thus ?case using step unfolding x_dec unfolding first.simps last.simps get_target_def apply auto by (metis converse_rtrancl_into_rtrancl local.step)
 qed
qed

lemma seq_imp_refl: assumes "ss \<in> seq R" and "length (snd ss) \<le> 1" and "set (map l (snd ss)) \<subseteq> ds r S" 
defines "rel \<equiv> \<lambda> c. {(s,t). \<exists> r p \<sigma>. (s,t) \<in> rstep_r_p_s R r p \<sigma> \<and> l (s,r,p,\<sigma>,t) = c}"
shows "(first ss,last ss) \<in> (\<Union>c \<in> ds r S. rel c)^=" proof (cases "snd ss")
 case Nil hence "first ss = last ss" using first.simps last.simps by (metis surjective_pairing)
  thus ?thesis by auto
 next
 case (Cons x xs) hence ss_dec: "ss = (fst ss, x#xs)" using assms by (metis (erased, hide_lams) surjective_pairing)
  hence seq:"(fst ss, x#xs) \<in> seq R" using assms by auto
  then obtain t rl p \<sigma> u where  x_dec: "x = (t,rl,p,\<sigma>,u)" by (cases, auto)
  hence xs_dec: "xs = []" using assms Cons by auto
  hence k: "(fst ss,(t,rl,p,\<sigma>,u)#[]) \<in> seq R" using ss_dec seq unfolding x_dec by auto
  hence eq1: "fst ss = t" using seq.cases by (cases, auto)
  hence ss_dec: "ss = (t,[x])" using ss_dec xs_dec by auto
  from k have "(t,(t,rl,p,\<sigma>,u)#[]) \<in> seq R" by (cases, auto)
  hence step: "(t,u) \<in> rstep_r_p_s R rl p \<sigma>" using seq_chop_first by fast
  have lab: "l (t,rl,p,\<sigma>,u) \<in> ds r S" using assms ss_dec unfolding x_dec by (metis (erased, hide_lams) insert_iff list.simps(9) local.Cons set_simps(2) subsetCE x_dec)
  have eq2: "last ss = u" unfolding ss_dec x_dec xs_dec last.simps get_target_def by auto
 show ?thesis using step lab unfolding ss_dec first.simps last.simps x_dec get_target_def snd_conv rel_def by fast
qed

lemma eld_imp_fars_eld_step:
assumes la: "l(x,rl1,p1,s1,y) = a" and lb: "l(x,rl2,p2,s2,z) = b" 
 and eld: "eld R l r ((x,rl1,p1,s1,y),(x,rl2,p2,s2,z))" (is "eld R l r ?p")
defines "rel \<equiv> \<lambda> c. {(s,t). \<exists> r p \<sigma>. (s,t) \<in> rstep_r_p_s R r p \<sigma> \<and> l (s,r,p,\<sigma>,t) = c}"
shows "(\<exists>y1 y2 z1 z2 v.
                             (y, y1) \<in> (\<Union>c\<in>fst r \<down>s {a}. rel c)\<^sup>* \<and>
                             (y1, y2) \<in> (\<Union>c\<in>snd r \<down>s {b}. rel c)\<^sup>= \<and>
                             (y2, v) \<in> (\<Union>c\<in>fst r \<down>s {a, b}. rel c)\<^sup>* \<and>
                             (z, z1) \<in> (\<Union>c\<in>fst r \<down>s {b}. rel c)\<^sup>* \<and>
                             (z1, z2) \<in> (\<Union>c\<in>snd r \<down>s {a}. rel c)\<^sup>= \<and>
                             (z2, v) \<in> (\<Union>c\<in>fst r \<down>s {a, b}. rel c)\<^sup>* )" proof -
 from eld obtain jl jr where d: "ld_trs R ((x, rl1, p1, s1, y), x, rl2, p2, s2, z) jl jr" 
   and eld: "eld_seq l r ((x, rl1, p1, s1, y), x, rl2, p2, s2, z) jl jr" unfolding eld_def by auto
 from d eld have jl: "jl \<in> seq R" and fsty: "first jl = y" and "jl \<in> seq R" unfolding ld_trs_def get_target_def by auto
 then obtain \<sigma>1 \<sigma>2 \<sigma>3 where ml: "map l (snd jl) = (\<sigma>1@\<sigma>2)@\<sigma>3" and eld1: "ELD_1' r (l (fst ?p)) (l (snd ?p)) \<sigma>1 \<sigma>2 \<sigma>3" using eld unfolding eld_seq_def by auto
 then obtain l12 l3 where jl_dec: "jl = seq_concat l12 l3" and l12: "l12 \<in> seq R" and l3: "l3 \<in> seq R" and m12: "map l (snd l12) = \<sigma>1@\<sigma>2" and m3: "map l (snd l3) = \<sigma>3"
   and eq_l3: "last l12 = first l3" and eq_l4: "last jl = last l3" using seq_dec[OF jl ml] by auto
 then obtain l1 l2 where l12_dec: "l12 = seq_concat l1 l2" and l1: "l1 \<in> seq R" and l2: "l2 \<in> seq R" and m1: "map l (snd l1) = \<sigma>1" and m2: "map l (snd l2) = \<sigma>2"
   and eq_l2: "last l1 = first l2" and eq_l12: "last l12 = last l2"  using seq_dec[OF l12 m12] by auto
 have len2: "length (snd l2) \<le> 1" using eld1 unfolding ELD_1'_def m2[symmetric] by auto 
 hence eq_l1: "first l1 = y" by (metis `jl = seq_concat l12 l3` `l12 = seq_concat l1 l2` first.simps fsty seq_concat_def)
 have m1: "set (map l (snd l1)) \<subseteq> (fst r) \<down>s {a}" using m1 using eld1 unfolding ELD_1'_def fst_conv snd_conv la lb by auto
 have l1: "(first l1,last l1) \<in> (\<Union>c \<in> ds (fst r) {a}. rel c)^*" using seq_imp_ast[OF l1 m1] eld unfolding rel_def by auto 
  have m2: "set (map l (snd l2)) \<subseteq> ds (snd r) {b}" using m2 using eld1 unfolding ELD_1'_def fst_conv snd_conv la lb by auto 
 have l2: "(first l2,last l2) \<in> (\<Union>c \<in> ds (snd r) {b}. rel c)^=" using seq_imp_refl[OF l2 len2 m2] eld unfolding rel_def by auto 
    have m3: "set (map l (snd l3)) \<subseteq> ds (fst r) {a,b}"  using m3 using eld1 unfolding ELD_1'_def fst_conv snd_conv la lb ds_def by auto 
 have l3: "(first l3,last l3) \<in> (\<Union>c \<in> ds (fst r) {a,b}. rel c)^*" using seq_imp_ast[OF l3 m3] eld unfolding rel_def by auto 

(* merge with upper twin *)
 from d eld have jr: "jr \<in> seq R" and fstz: "first jr = z" unfolding ld_trs_def get_target_def by auto
 then obtain \<tau>1 \<tau>2 \<tau>3 where mr: "map l (snd jr) = (\<tau>1@\<tau>2)@\<tau>3" and eld1: "ELD_1' r (l (snd ?p)) (l (fst ?p)) \<tau>1 \<tau>2 \<tau>3" using eld unfolding eld_seq_def by auto
 then obtain r12 r3 where jr_dec: "jr = seq_concat r12 r3" and r12: "r12 \<in> seq R" and r3: "r3 \<in> seq R" and m12: "map l (snd r12) = \<tau>1@\<tau>2" and m3: "map l (snd r3) = \<tau>3" 
  and eq_r4: "last jr = last r3" and eq_r3: "last r12 = first r3" using seq_dec[OF jr mr] by auto
 then obtain r1 r2 where r12_dec: "r12 = seq_concat r1 r2" and r1: "r1 \<in> seq R" and r2: "r2 \<in> seq R" and m1: "map l (snd r1) = \<tau>1" and m2: "map l (snd r2) = \<tau>2" 
  and eq_r2: "last r1 = first r2" and eq_r12: "last r12 = last r2"  using seq_dec[OF r12 m12] by auto
 have len2: "length (snd r2) \<le> 1" using eld1 unfolding ELD_1'_def m2[symmetric] by auto 
 hence eq_r1: "first r1 = z" by (metis `jr = seq_concat r12 r3` `r12 = seq_concat r1 r2` first.simps fstz seq_concat_def)
 have m1: "set (map l (snd r1)) \<subseteq> (fst r) \<down>s {b}" using m1 using eld1 unfolding ELD_1'_def fst_conv snd_conv la lb by auto
 have r1: "(first r1,last r1) \<in> (\<Union>c \<in> ds (fst r) {b}. rel c)^*" using seq_imp_ast[OF r1 m1] eld unfolding rel_def by auto 
  have m2: "set (map l (snd r2)) \<subseteq> ds (snd r) {a}" using m2 using eld1 unfolding ELD_1'_def fst_conv snd_conv la lb by auto 
 have r2: "(first r2,last r2) \<in> (\<Union>c \<in> ds (snd r) {a}. rel c)^=" using seq_imp_refl[OF r2 len2 m2] eld unfolding rel_def by auto 
    have m3: "set (map l (snd r3)) \<subseteq> ds (fst r) {a,b}"  using m3 using eld1 unfolding ELD_1'_def fst_conv snd_conv la lb ds_def by auto 
 have r3: "(first r3,last r3) \<in> (\<Union>c \<in> ds (fst r) {a,b}. rel c)^*" using seq_imp_ast[OF r3 m3] eld unfolding rel_def by auto 

 have eq: "last l3 = last r3" using eq_l4 eq_r4 d unfolding ld_trs_def by auto 
 show ?thesis using l1 l2 l3 r1 r2 r3  unfolding eq_l1 eq_l2 eq_l12[symmetric] eq_l3 eq_r1 eq_r2 eq_r12[symmetric] eq_r3 eq  by metis
qed

lemma eld_imp_fars_eld:  
 assumes lp: "\<forall> p \<in> local_peaks R. eld R l r p" 
 shows "fars_eld (fst r) (snd r) (\<lambda> c. {(s,t). \<exists> r p \<sigma>. (s,t) \<in> rstep_r_p_s R r p \<sigma> \<and> l (s,r,p,\<sigma>,t) = c})" (is "fars_eld _ _ ?rel") proof -
    {
    fix a b x y z
    assume "(x, y) \<in> {(s, t). \<exists>r p \<sigma>. (s, t) \<in> rstep_r_p_s R r p \<sigma> \<and> l (s, r, p, \<sigma>, t) = a}"
                          and "(x, z) \<in> {(s, t). \<exists>r p \<sigma>. (s, t) \<in> rstep_r_p_s R r p \<sigma> \<and> l (s, r, p, \<sigma>, t) = b}"
    then obtain rl1 p1 s1 rl2 p2 s2 where "(x,y) \<in> rstep_r_p_s R rl1 p1 s1" and la:"l (x,rl1,p1,s1,y) = a" and
                                          "(x,z) \<in> rstep_r_p_s R rl2 p2 s2" and lb:"l (x,rl2,p2,s2,z) = b" by auto
    hence lp_mem: "((x,rl1,p1,s1,y),(x,rl2,p2,s2,z)) \<in> local_peaks R" (is "?p \<in> _") unfolding local_peaks_def by fast
    hence eld: "eld R l r ?p" using lp by auto
    have "(\<exists>y1 y2 z1 z2 v.  (y, y1) \<in> (\<Union>c\<in>fst r \<down>s {a}. ?rel c)\<^sup>* \<and> (y1, y2) \<in> (\<Union>c\<in>snd r \<down>s {b}. ?rel c)\<^sup>= \<and> (y2, v) \<in> (\<Union>c\<in>fst r \<down>s {a, b}. ?rel c)\<^sup>* \<and>
                            (z, z1) \<in> (\<Union>c\<in>fst r \<down>s {b}. ?rel c)\<^sup>* \<and> (z1, z2) \<in> (\<Union>c\<in>snd r \<down>s {a}. ?rel c)\<^sup>= \<and> (z2, v) \<in> (\<Union>c\<in>fst r \<down>s {a, b}. ?rel c)\<^sup>* )" 
   using eld_imp_fars_eld_step[OF la lb eld] unfolding rel_def by auto 
   }
   thus ?thesis unfolding fars_eld_def rel_def by blast 
 qed
 
lemma main:
  assumes "rel_props (fst r) (snd r)" 
  and lab: "compatible R l r" 
  and dec: "\<forall> (b, p) \<in> critical_peaks R. eld R l r p" 
  shows "CR (rstep R)" 
proof -
  let ?rel = "\<lambda> c. {(s,t). \<exists> r p \<sigma>. (s,t) \<in> rstep_r_p_s R r p \<sigma> \<and> l (s,r,p,\<sigma>,t) = c}"
  from assms have i: "irrefl (fst r)" unfolding rel_props_def by (metis irrefl_def wf_not_sym)
  have lp: "\<forall>p \<in> local_peaks R. eld R l r p" using main_eld[OF dec lab] by fast
  have eq: "(\<Union>c\<in>UNIV. ?rel c) = rstep R" (is "?l = _")  unfolding rstep_is_rstep_r_p_s using assms apply auto by metis 
  hence fars_eld: "fars_eld (fst r) (snd r) ?rel" using eld_imp_fars_eld[OF lp] by auto
 thus ?thesis using fars_eld_imp_cr[OF assms(1) fars_eld] eq by auto
qed

(* Section: Applications *)
(* towards concrete labelings *)

lemma parallel_peak_close: 
assumes "parallel_peak R p"
shows "\<exists> s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u v.(
 p = ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u)) \<and>
 (s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1 \<and>
 (s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<and>
 (t,v) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<and>
 (u,v) \<in> rstep_r_p_s R rl1 p1 \<sigma>1
)" proof -
 from assms(1) obtain r1 p1 s1 t s r2 p2 s2 u where p: "p \<in> local_peaks R" and p_dec: "p = ((s,r1,p1,s1,t),(s,r2,p2,s2,u))" and b:"p1\<bottom>p2" unfolding parallel_peak_def local_peaks_def Let_def by fast
 hence s1: "(s,t) \<in> rstep_r_p_s R r1 p1 s1" and s2: "(s,u) \<in> rstep_r_p_s R r2 p2 s2" unfolding local_peaks_def by auto
 obtain v where j1: "(t,v) \<in> rstep_r_p_s R r2 p2 s2" and j2:"(u,v) \<in> rstep_r_p_s R r1 p1 s1" using parallel_steps[OF _ _ b] s1 s2 surj_pair by metis
 thus ?thesis using s1 s2 p_dec by fast
qed

lemma local_peaks_intros: assumes "parallel_peak R p" shows "p \<in> local_peaks R" using assms unfolding parallel_peak_def by auto

lemma diamond_trs_and_weld_seq_imp_eld: 
assumes "diamond_trs R p jl jr" and weld_seq: "weld_seq l r p jl jr" shows "eld R l r p" proof -
 from assms have ld_trs: "ld_trs R p jl jr" unfolding ld_trs_def diamond_trs_def by auto
 have "eld_seq l r p jl jr" proof -
  have "\<forall>a\<in>set (map l (snd jl)). (a, l (snd p)) \<in> snd r" using assms unfolding weld_seq_def labels_def ds_def by auto 
  hence l: "ELD_1' r (l (fst p)) (l (snd p)) [] (map l (snd jl)) []" using assms unfolding diamond_trs_def weld_seq_def ELD_1'_def ds_def by auto
  have lx: "map l (snd jl) = [] @ map l (snd jl) @ []" by auto
  have "\<forall>a\<in>set (map l (snd jr)). (a, l (fst p)) \<in> snd r" using assms unfolding weld_seq_def labels_def ds_def by auto 
  hence r: "ELD_1' r (l (snd p)) (l (fst p)) [] (map l (snd jr)) []" using assms unfolding diamond_trs_def weld_seq_def ELD_1'_def ds_def by auto
  have rx: "map l (snd jr) = [] @ map l (snd jr) @ []" by auto
  show ?thesis using l lx r rx unfolding eld_seq_def by fast
 qed
 thus ?thesis using ld_trs unfolding eld_def by fast
qed

 (** rule labeling **)
fun rule_labeling :: "(('f,'v) rule \<Rightarrow> nat) \<Rightarrow> ('f,'v) step \<Rightarrow> nat"
 where "rule_labeling i (s,rl,p,\<sigma>,t) = i rl"

lemma rl_is_labeling: shows "is_labeling R (rule_labeling i) r" unfolding is_labeling_def by auto 

lemma rl_parallel:
assumes r: "refl (snd r)" and "parallel_peak R p"
shows  "eld R (rule_labeling i) r p" proof -
 obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u v where
 p_dec: "p = ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u))" (is "p = ?p") and
 steps: "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1 \<and>
 (s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<and>
 (t,v) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<and>
 (u,v) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" using parallel_peak_close[OF assms(2)] by fast
 have lp: "p \<in> local_peaks R" using local_peaks_intros[OF assms(2)] by auto
 from steps have jl: "(t,[(t,rl2,p2,\<sigma>2,v)]) \<in> seq R" (is "?jl \<in> _") using seq.intros by fast
 from steps have jr: "(u,[(u,rl1,p1,\<sigma>1,v)]) \<in> seq R" (is "?jr \<in> _") using seq.intros by fast
 have diamond_trs: "diamond_trs R ?p ?jl ?jr" using lp jl jr unfolding p_dec ld_trs_def last.simps get_target_def diamond_trs_def by auto 
 have weld_seq: "weld_seq (rule_labeling i) r ?p ?jl ?jr" unfolding weld_seq_def labels_def ds_def apply auto using r apply (metis UNIV_I refl_on_def) using r by (metis UNIV_I refl_on_def)
 thus ?thesis using diamond_trs_and_weld_seq_imp_eld[OF diamond_trs weld_seq] unfolding p_dec by metis
qed

lemma rl_variable:
assumes r: "refl (snd r)" and vp: "variable_peak R p" 
and lin: "linear_trs R"
shows "eld R (rule_labeling i) r p" proof -
 from assms have p: "p \<in> local_peaks R" unfolding variable_peak_def by auto
 from assms variable_peak_decompose[OF vp] obtain s t u l1 r1 p1 \<sigma>1 l2 r2 p2 \<sigma>2 q q1 q2 w where
   p_dec: "p = ((s, (l1, r1), p1, \<sigma>1, t), s, (l2, r2), p2, \<sigma>2, u)" and 
   s1: "(s, t) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1" and s2: "(s, u) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2" and
   rest: "p1 <#> q = p2" "q1 <#> q2 = q" "q1 \<in> poss l1" "l1 |_ q1 = Var w" by fast
 let ?\<sigma>3 = "(\<lambda>y. if y = w then (ctxt_of_pos_term q2 (\<sigma>1 y))\<langle>r2 \<cdot> \<sigma>2\<rangle> else \<sigma>1 y)"
 let ?v = "(ctxt_of_pos_term p1 s)\<langle>r1 \<cdot> ?\<sigma>3\<rangle>"
 from lin have linl1: "linear_term l1" and linl2: "linear_term r1" using s1 s2 unfolding rstep_r_p_s_def linear_trs_def Let_def by auto 
 from variable_steps_lin[OF s1 s2 rest linl1] linl2 have
 sr: "(u, ?v) \<in> rstep_r_p_s R (l1, r1) p1 ?\<sigma>3" and
 sl: "t = ?v \<or> (\<exists>q. q \<in> poss r1 \<and> r1 |_ q = Var w \<and> (t, ?v) \<in> rstep_r_p_s R (l2, r2) (p1 <#> q <#> q2) \<sigma>2)" unfolding linear_trs_def by auto
 have jr: "(u,[(u,(l1,r1),p1,?\<sigma>3,?v)]) \<in> seq R" (is "?jr \<in> _") using rstep_is_seq[OF sr] by auto
 show ?thesis proof (cases "t = ?v") 
  case True hence jl: "(t,[]) \<in> seq R" (is "?jl \<in> _") using seq.intros by auto
  have "last ?jl = get_target (fst p)" unfolding p_dec last.simps get_target_def by auto
  hence diamond_trs: "diamond_trs R p ?jl ?jr" using p jl jr True unfolding diamond_trs_def ld_trs_def p_dec by (auto simp add: get_target_def)
  have weld_seq: "weld_seq (rule_labeling i) r p ?jl ?jr" unfolding weld_seq_def labels_def p_dec ds_def apply auto using r by (metis (poly_guards_query) UNIV_I refl_on_def)
  thus ?thesis using diamond_trs_and_weld_seq_imp_eld[OF diamond_trs weld_seq] by fast
 next
  case False then obtain q where sl: "(t, ?v) \<in> rstep_r_p_s R (l2, r2) (p1 <#> q <#> q2) \<sigma>2" using sl by auto
  hence jl: "(t,[(t,(l2,r2),p1<#>q<#>q2,\<sigma>2,?v)]) \<in> seq R" (is "?jl \<in> _") using rstep_is_seq[OF sl] by auto
  hence diamond_trs: "diamond_trs R p ?jl ?jr" using p jl jr unfolding diamond_trs_def ld_trs_def p_dec get_target_def  by (auto simp add: get_target_def)
  have weld_seq: "weld_seq (rule_labeling i) r p ?jl ?jr" unfolding weld_seq_def labels_def p_dec ds_def apply auto using r apply (metis UNIV_I refl_on_def) using r by (metis UNIV_I refl_on_def)  
  thus ?thesis using diamond_trs_and_weld_seq_imp_eld[OF diamond_trs weld_seq] by fast
 qed
qed

lemma rule_labeling_is_compatible:
assumes r: "refl (snd r)" and lin: "linear_trs R"
shows "compatible R (rule_labeling i) r" using assms unfolding compatible_def using rl_is_labeling rl_variable[OF r _ lin] rl_parallel by metis

lemma rule_labeling_has_rel_props:
defines "r \<equiv> ({(n,m). n < (m::nat)},{(n,m). n \<le> (m::nat)})"
shows "rel_props (fst r) (snd r)" proof -
 have r:  "refl (snd r)" using r_def apply auto by (simp add: refl_on_def)
 have t1: "trans (fst r)" using r_def apply auto by (metis less_trans transp_def transp_trans) 
 have t2: "trans (snd r)" using r_def apply auto by (metis order_trans transp_def transp_trans)
 have wf: "wf (fst r)" using r_def apply auto by (metis wf_less) 
 have compat: "snd r O fst r O snd r \<subseteq> fst r" unfolding r_def by auto 
 have "rel_props (fst r) (snd r)" unfolding rel_props_def using t1 t2 wf r compat by auto
 thus ?thesis by auto
qed 
 
corollary rule_labeling_is_sound:
defines "r \<equiv> ({(n,m). n < (m::nat)},{(n,m). n \<le> (m::nat)})"
assumes lin: "linear_trs R"
and dec: "\<forall> (b,p) \<in> critical_peaks R. eld R (rule_labeling i) r p" (is "\<forall> (b,p) \<in> _ . eld _ ?l ?r _")
shows "CR (rstep R)" proof -
 have r:  "refl (snd r)" using r_def apply auto by (simp add: refl_on_def)
 have lab: "compatible R (rule_labeling i) ?r" using rule_labeling_is_compatible[OF r lin]  by auto
 have props: "rel_props (fst r) (snd r)" using rule_labeling_has_rel_props r_def by fast
 show ?thesis using main[OF props lab dec] by auto
qed

(** source labeling **)
fun source_labeling :: "('f,'v) step \<Rightarrow> ('f,'v) term"
 where "source_labeling (s,rl,p,\<sigma>,t) = s"
 
fun sl_rel :: "('f,'v) trs \<Rightarrow> ('f,'v) trs \<Rightarrow> (('f,'v) term) relp"
 where "sl_rel R S = (((relto (rstep R) (rstep S))^+)\<inverse>,((rstep S \<union> rstep R)^* )\<inverse>)"

lemma sl_is_labeling:
shows "is_labeling (R\<union>S) source_labeling (sl_rel R S)" proof -
 {fix s a b t u C \<rho> \<sigma>1 aa p1 ba v \<sigma>2 p2 
 assume        C:"(u, s) \<in> ((rstep S)\<^sup>* O rstep R O (rstep S)\<^sup>* )\<^sup>+" (is "_ \<in> ?rel^+") 
 then obtain n where n1: "n > 0" and n:"(u,s) \<in> ?rel^^n" by (metis trancl_power) 
 from n have "(C\<langle>u \<cdot> \<rho>\<rangle>, C\<langle>s \<cdot> \<rho>\<rangle>) \<in> ?rel^^n"  proof (induct n arbitrary:u)
  case 0 thus ?case by auto
  next
  case (Suc n) then obtain v where step: "(u,v) \<in> ?rel" and steps: "(v,s) \<in> ?rel^^n" by (metis relpow_Suc_E2)
  from step obtain u' v' where "(u,u') \<in> (rstep S)^*" and "(u',v') \<in> rstep R" and "(v',v) \<in> (rstep S)^*" by auto 
  then obtain u' v' where "(C\<langle>u\<cdot>\<rho>\<rangle>,C\<langle>u'\<cdot>\<rho>\<rangle>) \<in> (rstep S)^*" and "(C\<langle>u'\<cdot>\<rho>\<rangle>,C\<langle>v'\<cdot>\<rho>\<rangle>) \<in> rstep R" and "(C\<langle>v'\<cdot>\<rho>\<rangle>,C\<langle>v\<cdot>\<rho>\<rangle>) \<in> (rstep S)^*"
   by (metis rstep.ctxt rstep.subst rsteps_closed_ctxt rsteps_closed_subst)
  hence 1:"(C\<langle>u\<cdot>\<rho>\<rangle>,C\<langle>v\<cdot>\<rho>\<rangle>) \<in> ?rel" by auto
  from steps have 2:"(C\<langle>v\<cdot>\<rho>\<rangle>,C\<langle>s\<cdot>\<rho>\<rangle>) \<in> ?rel^^n" using Suc by auto
  show ?case using 1 2 by (metis relpow_Suc_I2)
  qed
  hence "((C\<langle>u \<cdot> \<rho>\<rangle>, C\<langle>s \<cdot> \<rho>\<rangle>) \<in> ?rel^+)" using n1 by (metis trancl_power)
 } note A = this
 {fix s a b t u C \<rho> \<sigma>1 aa p1 ba v \<sigma>2 p2
  assume "(u, s) \<in> (rstep S \<union> rstep R)\<^sup>*"
   hence "(C\<langle>u \<cdot> \<rho>\<rangle>, C\<langle>s \<cdot> \<rho>\<rangle>) \<in> (rstep S \<union> rstep R)\<^sup>*" by (metis rstep_union rsteps_closed_ctxt rsteps_closed_subst)
 }
 thus ?thesis using A unfolding is_labeling_def step_close.simps by auto
qed

lemma sl_par:
assumes pp: "parallel_peak (R\<union>S) p"
shows "weld (R\<union>S) source_labeling (sl_rel R S) p" (is "weld _ _ ?r _") proof -
 obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u v where
 p_dec: "p = ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u))"  and
 steps: "(s,t) \<in> rstep_r_p_s (R\<union>S) rl1 p1 \<sigma>1 \<and>
 (s,u) \<in> rstep_r_p_s (R\<union>S) rl2 p2 \<sigma>2 \<and>
 (t,v) \<in> rstep_r_p_s (R\<union>S) rl2 p2 \<sigma>2 \<and>
 (u,v) \<in> rstep_r_p_s (R\<union>S) rl1 p1 \<sigma>1" using parallel_peak_close[OF assms] by fast
 have lp: "p \<in> local_peaks (R\<union>S)" using local_peaks_intros[OF assms] by auto
 from steps have jl: "(t,[(t,rl2,p2,\<sigma>2,v)]) \<in> seq (R\<union>S)" (is "?jl \<in> _") using seq.intros by fast
 from steps have jr: "(u,[(u,rl1,p1,\<sigma>1,v)]) \<in> seq (R\<union>S)" (is "?jr \<in> _") using seq.intros by fast
 have diamond_trs: "diamond_trs (R\<union>S) p ?jl ?jr" using lp jl jr unfolding p_dec diamond_trs_def ld_trs_def last.simps get_target_def by auto 
 from steps have s1:"(s,t) \<in> rstep (R \<union> S)" using rstep_r_p_s_imp_rstep by fast
 from steps have s2:"(s,u) \<in> rstep (R \<union> S)" using rstep_r_p_s_imp_rstep by fast
 have weld_seq: "weld_seq source_labeling ?r p ?jl ?jr" unfolding weld_seq_def labels_def p_dec ds_def apply auto
  using s1 s2 unfolding ds_def by auto
 show ?thesis using diamond_trs weld_seq unfolding weld_def ld2_trs_def diamond_trs_def by fast
qed

lemma sl_evar:
assumes vp: "variable_peak (R\<union>S) p" (is "variable_peak ?R _")
and lin: "linear_trs (R\<union>S)"
shows "weld (R\<union>S) source_labeling (sl_rel R S) p" proof -
 from assms have p: "p \<in> local_peaks ?R" unfolding variable_peak_def by auto
 from assms variable_peak_decompose[OF vp] obtain s t u l1 r1 p1 \<sigma>1 l2 r2 p2 \<sigma>2 q q1 q2 w where
   p_dec: "p = ((s, (l1, r1), p1, \<sigma>1, t), s, (l2, r2), p2, \<sigma>2, u)" and 
   s1: "(s, t) \<in> rstep_r_p_s ?R (l1, r1) p1 \<sigma>1" and s2: "(s, u) \<in> rstep_r_p_s ?R (l2, r2) p2 \<sigma>2" and
   rest: "p1 <#> q = p2" "q1 <#> q2 = q" "q1 \<in> poss l1" "l1 |_ q1 = Var w" by fast
 let ?\<sigma>3 = "(\<lambda>y. if y = w then (ctxt_of_pos_term q2 (\<sigma>1 y))\<langle>r2 \<cdot> \<sigma>2\<rangle> else \<sigma>1 y)"
 let ?v = "(ctxt_of_pos_term p1 s)\<langle>r1 \<cdot> ?\<sigma>3\<rangle>"
 from s1 lin have l:"linear_term l1 \<and> linear_term r1" unfolding rstep_r_p_s_def Let_def linear_trs_def by auto
 from variable_steps_lin[OF s1 s2 rest] l have
 sr: "(u, ?v) \<in> rstep_r_p_s ?R (l1, r1) p1 ?\<sigma>3" and
 sl: "t = ?v \<or> (\<exists>q. q \<in> poss r1 \<and> r1 |_ q = Var w \<and> (t, ?v) \<in> rstep_r_p_s ?R (l2, r2) (p1 <#> q <#> q2) \<sigma>2)" by auto
 have jr: "(u,[(u,(l1,r1),p1,?\<sigma>3,?v)]) \<in> seq ?R" (is "?jr \<in> _") using rstep_is_seq[OF sr] by auto
 show ?thesis proof (cases "t = ?v") 
  case True hence jl: "(t,[]) \<in> seq ?R" (is "?jl \<in> _") using seq.intros by auto
  have "last ?jl = get_target (fst p)" unfolding p_dec last.simps get_target_def by auto
  hence diamond_trs: "diamond_trs ?R p ?jl ?jr" using p jl jr True unfolding diamond_trs_def ld_trs_def p_dec by (auto simp add: get_target_def)
  have "weld_seq source_labeling (sl_rel R S) p ?jl ?jr" 
  unfolding weld_seq_def labels_def p_dec ds_def apply auto using s2 
   by (metis (erased, hide_lams) inf_sup_aci(5) r_into_rtrancl rstep_r_p_s_imp_rstep rstep_union)
   thus ?thesis using diamond_trs unfolding weld_def diamond_trs_def ld2_trs_def by fast
 next
  case False then obtain q where sl: "(t, ?v) \<in> rstep_r_p_s ?R (l2, r2) (p1 <#> q <#> q2) \<sigma>2" using sl by auto
  hence jl: "(t,[(t,(l2,r2),p1<#>q<#>q2,\<sigma>2,?v)]) \<in> seq ?R" (is "?jl \<in> _") using rstep_is_seq[OF sl] by auto
  hence diamond_trs: "diamond_trs ?R p ?jl ?jr" using p jl jr unfolding diamond_trs_def ld_trs_def p_dec get_target_def by (auto simp add: get_target_def)
  have r: "refl (snd (sl_rel R S))" by auto
  have weld_seq: "weld_seq source_labeling (sl_rel R S) p ?jl ?jr"
   unfolding weld_seq_def labels_def p_dec ds_def apply auto 
   using s1  apply (metis (erased, hide_lams) inf_sup_aci(5) r_into_rtrancl rstep_r_p_s_imp_rstep rstep_union)
   using s2 by (metis (erased, hide_lams) inf_sup_aci(5) r_into_rtrancl rstep_r_p_s_imp_rstep rstep_union)
    thus ?thesis using diamond_trs unfolding weld_def diamond_trs_def ld2_trs_def by fast
 qed
qed

lemma source_labeling_is_weakly_compatible:
assumes lin: "linear_trs (R \<union> S)"
shows "weakly_compatible (R \<union> S) source_labeling (sl_rel R S)" unfolding weakly_compatible_def 
using sl_is_labeling sl_par sl_evar lin unfolding weld_def eld_def eld_seq_def weld_seq_def by fast

definition rel_rstep :: "('f,'v) trs \<Rightarrow> ('f,'v) trs \<Rightarrow> (('f,'v) term) rel"
 where "rel_rstep R S = relto (rstep R) (rstep S)"
 
lemma S_RS: assumes "(s,t) \<in> (rstep S)^*" and "(t,u) \<in> relto (rstep R) (rstep S)" shows "(s,u) \<in> relto (rstep R) (rstep S)" 
 by (metis (erased, hide_lams) O_assoc assms relcomp.relcompI rtrancl_idemp_self_comp)

lemma S_RS_trancl: assumes "(s,t) \<in> (rstep S)^*" and "(t,u) \<in> (relto (rstep R) (rstep S))^+" 
 shows "(s,u) \<in> (relto (rstep R) (rstep S))^+" (is "_ \<in> ?R^+") proof -
 from assms obtain m where "(t,m) \<in> ?R" and mu:"(m,u) \<in> ?R^*" by (metis tranclD)
 hence "(s,m) \<in> ?R" using assms(1) S_RS by metis
 thus ?thesis using mu by auto
qed

lemma RS_S:
  assumes "(t, u) \<in> relto (rstep R) (rstep S)" and "(u, v) \<in> (rstep S)\<^sup>*"
  shows "(t, v) \<in> relto (rstep R) (rstep S)"
  using assms by force 

lemma RS_S_trancl: 
 assumes "(t,u) \<in> (relto (rstep R) (rstep S))^+" and "(u,v) \<in> (rstep S)^*" 
 shows "(t,v) \<in> (relto (rstep R) (rstep S))^+" (is "_ \<in> ?R") proof -
  from assms obtain m where "(t,m) \<in> ?R^*" and "(m,u) \<in> ?R" by blast
  thus ?thesis  using assms RS_S rtrancl_into_trancl1 tranclD2 by metis
qed

lemma rel_compat:
assumes "(s,t) \<in> (rstep S \<union> rstep R)^*" and "(t,u) \<in> (rel_rstep R S)^+" and "(u,v) \<in> (rstep S \<union> rstep R)^*"
shows "(s,v) \<in> (rel_rstep R S)^+"
proof -
 have k: "(rstep S \<union> rstep R)^* \<subseteq> ((rstep S)^* \<union> (relto (rstep R) (rstep S))^*)" by regexp
 from k have ab: "(s,t) \<in> (relto (rstep R) (rstep S))^* \<or> (s,t) \<in> (rstep S)^*" (is "?a \<or> ?b") using assms by auto
 from k have cd:"(u,v) \<in> (relto (rstep R) (rstep S))^* \<or> (u,v) \<in> (rstep S)^*" (is "?c \<or> ?d") using assms by auto
 have su: "(s,u) \<in> (relto (rstep R) (rstep S))^+" proof (cases "?a")
  case True thus ?thesis using assms unfolding rel_rstep_def by (metis rtrancl_trancl_trancl)
  next
  case False hence "?b" using ab by auto
   thus ?thesis using assms(2) unfolding rel_rstep_def using S_RS_trancl by metis
 qed
 thus ?thesis proof (cases "?c")
  case True thus ?thesis unfolding rel_rstep_def using su by (metis True trancl_rtrancl_trancl)
  next
  case False hence "?d" using cd by auto
   thus ?thesis using su unfolding rel_rstep_def using RS_S_trancl by metis
   qed
qed

corollary source_labeling_and_SN_is_sound:
  fixes R S
  defines "r \<equiv> sl_rel R S"
  assumes lin: "linear_trs (R\<union>S)" (is "linear_trs ?R")
  and SN: "SN (rstep (R\<union>S))"
  and dec: "\<forall>(b,p) \<in> critical_peaks (R\<union>S). eld (R\<union>S) source_labeling (sl_rel R S) p" 
  shows "CR (rstep (R\<union>S))"
proof -
  have lab: "compatible ?R source_labeling (sl_rel R S)" 
    using  source_labeling_is_weakly_compatible[OF lin] using linear_and_weakly_compatible_is_compatible[OF lin] by auto  
  have t1: "trans (fst r)" unfolding r_def sl_rel.simps by auto
  have t2: "trans (snd r)" unfolding r_def sl_rel.simps by auto
  have SN_R:"SN (rstep R)" and SN_S:"SN (rstep S)" using SN apply (metis SN_subset Un_left_absorb rstep_union subset_Un_eq) 
    by (metis (erased, hide_lams) SN SN_subset le_sup_iff rstep_union subset_refl)
  hence "SN_rel (rstep R) (rstep S)"
  proof -
    have "SN_rel_on_alt (rstep (R \<union> S)) {} (Collect top)" by (metis SN SN_rel_empty2 SN_rel_on_conv top_set_def)
    hence "\<exists>x\<^sub>1. SN_rel_on_alt x\<^sub>1 {} (Collect top) \<and> rstep R \<subseteq> x\<^sub>1 \<and> rstep S \<subseteq> x\<^sub>1" using rstep_union by auto
    thus "SN_rel (rstep R) (rstep S)" by (metis (no_types) SN_rel_mono' SN_rel_on_conv sup_bot.right_neutral top_set_def)
  qed
  hence "SN ((fst r)\<inverse>)" using SN_R unfolding r_def sl_rel.simps apply auto by (metis SN_imp_SN_trancl SN_rel_on_def) 
  hence wf: "wf (fst r)" using r_def SN_iff_wf by auto
  have r: "refl (snd r)" using r_def by auto
  have compat: "snd (sl_rel R S) O fst (sl_rel R S) O snd (sl_rel R S) \<subseteq> fst (sl_rel R S)" apply auto 
    using rel_compat unfolding rel_rstep_def by metis
  have rel_props:"rel_props (fst (sl_rel R S)) (snd (sl_rel R S))" unfolding rel_props_def
    using t1 t2 wf r compat r_def by auto
  from lin have ll: "left_linear_trs (R\<union>S)" unfolding left_linear_trs_def linear_trs_def by auto
  show ?thesis using main[OF rel_props lab dec] by fast
qed

(* rule labeling left-linear case *)

lemma rl_variable2:
  assumes r: "refl (snd r)"
  and vp: "variable_peak R p" 
  and ll: "left_linear_trs R"
  shows "weld R (rule_labeling i) r p"
proof -
  from assms have p: "p \<in> local_peaks R" unfolding variable_peak_def by auto
  from assms variable_peak_decompose[OF vp] obtain s t u l1 r1 p1 \<sigma>1 l2 r2 p2 \<sigma>2 q q1 q2 w where
    p_dec: "p = ((s, (l1, r1), p1, \<sigma>1, t), s, (l2, r2), p2, \<sigma>2, u)" and 
    s1: "(s, t) \<in> rstep_r_p_s R (l1, r1) p1 \<sigma>1" and s2: "(s, u) \<in> rstep_r_p_s R (l2, r2) p2 \<sigma>2" and
    rest: "p1 <#> q = p2" "q1 <#> q2 = q" "q1 \<in> poss l1" "l1 |_ q1 = Var w" by fast
  let ?\<sigma>3 = "(\<lambda>y. if y = w then (ctxt_of_pos_term q2 (\<sigma>1 y))\<langle>r2 \<cdot> \<sigma>2\<rangle> else \<sigma>1 y)"
  let ?v = "(ctxt_of_pos_term p1 s)\<langle>r1 \<cdot> ?\<sigma>3\<rangle>"
  from ll have ll1: "linear_term l1" using s1 s2 unfolding rstep_r_p_s_def left_linear_trs_def Let_def by auto 
  show ?thesis
  proof (cases "linear_term r1")
    case False
    from variable_steps_left_lin[OF s1 s2 rest ll1] obtain steps where
      sr: "(u, ?v) \<in> rstep_r_p_s R (l1, r1) p1 ?\<sigma>3" and
      sl: "\<forall>i<length steps. fst (steps ! i) = (l2, r2)" "(t, steps, ?v) \<in> par_rstep2 R" unfolding linear_trs_def by auto
    have jr: "(u,[(u,(l1,r1),p1,?\<sigma>3,?v)]) \<in> seq R" (is "?jr \<in> _") using rstep_is_seq[OF sr] by auto
    obtain jl where "jl \<in> seq R" and "first jl = t" and "last jl = ?v"
      and jl_lab: "\<forall>i < length (snd jl). fst (snd (snd jl ! i)) = (l2,r2)"
      using par_rstep2_imp_seq sl unfolding get_rule_def by force
    hence d: "ld_trs R p jl ?jr" using p jr unfolding ld_trs_def p_dec get_target_def apply auto unfolding get_target_def by auto
    hence d2: "ld2_trs R p jl ?jr" using False unfolding ld2_trs_def p_dec get_rule_def by auto
    have  "weld_seq (rule_labeling i) r p jl ?jr" unfolding weld_seq_def p_dec labels_def using jl_lab r
      unfolding ds_def apply auto apply (metis UNIV_I fst_conv in_set_conv_nth refl_onD snd_conv)
      by (metis UNIV_I refl_onD)
    thus ?thesis unfolding weld_def using d2 by fast
  next
    case True
    hence l: "linear_term r1" by auto
    from variable_steps_lin[OF s1 s2 rest ll1] l have
      sr: "(u, ?v) \<in> rstep_r_p_s R (l1, r1) p1 ?\<sigma>3" and
      sl: "t = ?v \<or> (\<exists>q. q \<in> poss r1 \<and> r1 |_ q = Var w \<and> (t, ?v) \<in> rstep_r_p_s R (l2, r2) (p1 <#> q <#> q2) \<sigma>2)"
      unfolding linear_trs_def by auto
    have jr: "(u,[(u,(l1,r1),p1,?\<sigma>3,?v)]) \<in> seq R" (is "?jr \<in> _") using rstep_is_seq[OF sr] by auto
    show ?thesis
    proof (cases "t = ?v") 
      case True
      hence jl: "(t,[]) \<in> seq R" (is "?jl \<in> _") using seq.intros by auto
      have "?jl \<in> seq R" using jl by auto
      moreover have "length (snd ?jl) \<le> 1" by auto
      moreover have "first ?jl = t" by auto
      moreover have "last ?jl = ?v" using True by auto
      ultimately have jl: "?jl \<in> seq R \<and> length (snd ?jl) \<le> 1 \<and> first ?jl = t \<and> last ?jl = ?v" by auto
      have "p \<in> local_peaks R" using s1 s2 unfolding p_dec local_peaks_def by auto
      hence "ld_trs R p ?jl ?jr" using jl jr unfolding p_dec ld_trs_def local_peaks_def get_target_def apply auto
        unfolding get_target_def by auto
      hence d2: "ld2_trs R p ?jl ?jr" using jl jr unfolding ld2_trs_def by auto
      have  "weld_seq (rule_labeling i) r p ?jl ?jr" unfolding weld_seq_def p_dec labels_def using r
        unfolding ds_def apply auto using assms(1) by (metis UNIV_I fst_conv in_set_conv_nth refl_onD snd_conv)
      thus ?thesis using d2 unfolding weld_def by fast
    next
      case False
      then obtain q where sl: "(t, ?v) \<in> rstep_r_p_s R (l2, r2) (p1 <#> q <#> q2) \<sigma>2" using sl by auto
      hence jl: "(t,[(t,(l2,r2),p1<#>q<#>q2,\<sigma>2,?v)]) \<in> seq R" (is "?jl \<in> _") using rstep_is_seq[OF sl] by auto
      hence diamond_trs: "diamond_trs R p ?jl ?jr" using p jl jr unfolding diamond_trs_def ld_trs_def p_dec get_target_def
        by (auto simp add: get_target_def)
      hence d2: "ld2_trs R p ?jl ?jr" using p jl jr unfolding p_dec get_target_def diamond_trs_def ld_trs_def ld2_trs_def by auto
      have  "weld_seq (rule_labeling i) r p ?jl ?jr" unfolding weld_seq_def p_dec labels_def using r
        unfolding ds_def apply auto using assms(1)
        apply (metis UNIV_I refl_on_def) by (metis UNIV_I refl_on_def)
      thus ?thesis using d2 unfolding weld_def by fast
    qed
  qed
qed

lemma rule_labeling_parallel2:
  assumes "refl (snd r)" and "parallel_peak R p"
  shows "weld R (rule_labeling i) r p"
proof -
  from assms parallel_peak_close obtain s t u v rl1 p1 \<sigma>1 rl2 p2 \<sigma>2
    where p_dec: "p = ((s, rl1, p1, \<sigma>1, t), s, rl2, p2, \<sigma>2, u)" and x: "(s, t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1 \<and>
    (s, u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<and> (t, v) \<in> rstep_r_p_s R rl2 p2 \<sigma>2 \<and> (u, v) \<in> rstep_r_p_s R rl1 p1 \<sigma>1"
    by fast
  from x have 1:"(s,[(s,rl1,p1,\<sigma>1,t)]) \<in> seq R" by (metis rstep_is_seq)
  from x have 2:"(s,[(s,rl2,p2,\<sigma>2,u)]) \<in> seq R" by (metis rstep_is_seq)
  from x have 3:"(t,[(t,rl2,p2,\<sigma>2,v)]) \<in> seq R" (is "?jl \<in> _") by (metis rstep_is_seq)
  from x have 4:"(u,[(u,rl1,p1,\<sigma>1,v)]) \<in> seq R" (is "?jr \<in> _") by (metis rstep_is_seq)
  from p_dec x have "p \<in> local_peaks R" (is "?p \<in> _") unfolding local_peaks_def by fast
  hence "ld_trs R p ?jl ?jr" using 3 4 unfolding p_dec ld_trs_def get_target_def apply auto unfolding get_target_def by auto
  hence x1: "ld2_trs R p ?jl ?jr" unfolding ld2_trs_def by auto
  have x2: "weld_seq (rule_labeling i) r p ?jl ?jr"  
    unfolding weld_seq_def p_dec labels_def ds_def apply auto using assms
    apply (metis UNIV_I refl_on_def) by (metis UNIV_I assms(1) refl_on_def)
  thus ?thesis using x1 unfolding weld_def by auto
qed
 
(* rule labeling (left linear case) *)
lemma rule_labeling_is_weakly_compatible:
  assumes "refl (snd r)" and "left_linear_trs R"
  shows "weakly_compatible R (rule_labeling i) r"
proof -
  have c: "is_labeling R (rule_labeling i) r" using rl_is_labeling by fast
  {
    fix p
    assume "variable_peak R p \<or> parallel_peak R p" and "left_linear_trs R"
    hence "weld R (rule_labeling i) r p" using rl_variable2[OF assms(1) _ assms(2)] rule_labeling_parallel2[OF assms(1)] by metis
  }
  thus ?thesis unfolding weakly_compatible_def using c by fast
qed
 
definition R_d :: "('f,'v) trs \<Rightarrow> ('f,'v) trs"
 where "R_d R = {rl \<in> R. \<not> linear_term (snd rl)}"
 
definition R_nd :: "('f,'v) trs \<Rightarrow> ('f,'v) trs"
 where "R_nd R = {rl \<in> R. linear_term (snd rl)}"
 
fun set2rel :: "('a \<times> 'a) set \<Rightarrow> ('a \<Rightarrow> 'a \<Rightarrow> bool)"
 where "set2rel r = (\<lambda> x y. (x,y) \<in> r)"
 
fun rel2set :: "('a \<Rightarrow> 'a \<Rightarrow> bool) \<Rightarrow> ('a \<times> 'a) set"
 where "rel2set r = {(a,b). r a b}"
 
definition lex_r :: "'a relp \<Rightarrow> 'b relp \<Rightarrow> ('a \<times> 'b) relp"
 where "lex_r r1 r2 = (rel2set (lex_two (set2rel (fst r1)) (set2rel (snd r1)) (set2rel (fst r2))),
                      rel2set (lex_two (set2rel (fst r1)) (set2rel (snd r1)) (set2rel (snd r2))))" 

lemma map_dec: 
  assumes "map f xs = ys1@ys2"
  shows "\<exists> xs1 xs2. xs = xs1@xs2 \<and> map f xs1 = ys1 \<and> map f xs2 = ys2"
using assms
proof (induct ys1 arbitrary: ys2 xs)
  case Nil thus ?case by auto
next
  case (Cons y ys)
  then obtain x' xs' where xs_dec: "xs = x'#xs'" by (metis append_Cons list.distinct(1) list.exhaust list.simps(8)) 
  hence map1: "map f xs' = ys@ys2" using Cons by auto
  then obtain xs1 xs2 where xs'_dec: "xs' = xs1 @ xs2" and map:" map f xs1 = ys \<and> map f xs2 = ys2" using Cons(1)[of xs' ys2] by fast
  hence "xs = (x'#xs1)@xs2" and "map f (x'#xs1) = (f x')#ys" and "map f xs2 = ys2" unfolding xs_dec by auto
  thus ?case by (metis Cons.prems append_Cons list.sel(1) list.simps(9))
qed

lemma R_d_union_R_nd: "R_d R \<union> R_nd R = R" unfolding R_d_def R_nd_def by auto

lemma l_sl:
  fixes l r R
  assumes "(s,t) \<in> (rstep R)^*"
  shows "(t,s) \<in> (snd (sl_rel (R_d R) (R_nd R)))"
using assms unfolding sl_rel.simps apply auto using R_d_union_R_nd by (metis inf_sup_aci(5) rstep_union) 
  
lemma l_imp_ll:
  fixes l r R
  defines "ll \<equiv> (\<lambda> step. (source_labeling step,l step))"
  and "rr \<equiv> (lex_r (sl_rel (R_d R) (R_nd R)) r)"
  assumes "l step  \<in> ds (fst r) {l step', l step''}" 
  and "(get_source step',get_source step) \<in> (rstep R)^*"
  and "(get_source step'',get_source step) \<in> (rstep R)^*"
  shows "ll step \<in> ds (fst rr) {ll step', ll step''}"
proof - 
  from assms obtain s rl p \<sigma> t where step_dec: "step = (s,rl,p,\<sigma>,t)" by (metis (erased, hide_lams) source_labeling.cases)
  show ?thesis
  proof (cases "l step \<in> ds (fst r) {l step'}")
    case True
    from assms obtain s' rl' p' \<sigma>' t' where step'_dec: "step' = (s',rl',p',\<sigma>',t')"
      by (metis (erased, hide_lams) source_labeling.cases)
    have "(source_labeling step',source_labeling step) \<in> (rstep R)^*" using assms unfolding step_dec step'_dec
      unfolding get_source_def by auto
    hence l2: "(source_labeling step,source_labeling step') \<in> (snd (sl_rel (R_d R) (R_nd R)))" using l_sl by auto
    hence l1: "(l step,l step') \<in> (fst r)" using True unfolding ds_def by auto
    hence "(ll step,ll step') \<in> fst rr" using l2 unfolding lex_r_def ll_def rr_def by auto
    thus ?thesis unfolding ds_def by auto
  next
    case False (* symmetric case *)
    from assms obtain s' rl' p' \<sigma>' t' where step''_dec: "step'' = (s',rl',p',\<sigma>',t')"
      by (metis (erased, hide_lams) source_labeling.cases)
    have "(source_labeling step'',source_labeling step) \<in> (rstep R)^*"
      using assms unfolding step_dec step''_dec unfolding get_source_def by auto
    hence l2: "(source_labeling step,source_labeling step'') \<in> (snd (sl_rel (R_d R) (R_nd R)))" using l_sl by auto
    hence l1: "(l step,l step'') \<in> (fst r)" using False assms unfolding ds_def by auto
    hence "(ll step,ll step'') \<in> fst rr" using l2 unfolding lex_r_def ll_def rr_def by auto
    thus ?thesis unfolding ds_def by auto
  qed
qed

lemma l_imp_ll_seq:
  fixes l r R
  defines "ll \<equiv> (\<lambda> step. (source_labeling step,l step))"
  and "rr \<equiv> (lex_r (sl_rel (R_d R) (R_nd R)) r)"
  assumes "set (map l ss) \<subseteq> ds (fst r) {l step',l step''}"
  and "\<forall>i < length ss. (get_source step' ,get_source (ss !i)) \<in> (rstep R)^*"
  and "\<forall>i < length ss. (get_source step'',get_source (ss !i)) \<in> (rstep R)^*"
  shows "set (map ll ss) \<subseteq> ds (fst rr) {ll step',ll step''}"
using assms(3-5)
proof (induct ss)
  case (Cons t ts)  
  hence step:"l t \<in> ds (fst r) {l step',l step''}" and steps:"set (map l ts) \<subseteq> fst r \<down>s {l step', l step''}" by auto
  from Cons have c1: "\<forall>i<length ts. (get_source step', get_source (ts ! i)) \<in> (rstep R)\<^sup>*" by auto
  from Cons have c2: "\<forall>i<length ts. (get_source step'', get_source (ts ! i)) \<in> (rstep R)\<^sup>*" by auto
  hence 2: "set (map ll ts) \<subseteq> fst rr \<down>s {ll step', ll step''}" using Cons(1)[OF steps c1 c2] by auto
  have r1: "(get_source step', get_source t) \<in> (rstep R)\<^sup>*" using Cons assms by auto
  have r2: "(get_source step'', get_source t) \<in> (rstep R)\<^sup>*" using Cons assms by auto 
  have 1: "ll t \<in> ds (fst rr) {ll step', ll step''}" using l_imp_ll[OF step r1 r2] unfolding ll_def rr_def by auto 
  show ?case using 1 2 by auto
qed auto
 
lemma l_imp_ll_eq: (* symmetric to l_imp_ll *)
  fixes l r R
  defines "ll \<equiv> (\<lambda> step. (source_labeling step,l step))"
  and "rr \<equiv> (lex_r (sl_rel (R_d R) (R_nd R)) r)"
  assumes "l step  \<in> ds (snd r) {l step'}" 
  and "(get_source step',get_source step) \<in> (rstep R)^*"
  shows "ll step \<in> ds (snd rr) {ll step'}"
proof - 
  from assms obtain s rl p \<sigma> t where step_dec: "step = (s,rl,p,\<sigma>,t)" by (metis (erased, hide_lams) source_labeling.cases)
  from assms obtain s' rl' p' \<sigma>' t' where step'_dec: "step' = (s',rl',p',\<sigma>',t')" by (metis (erased, hide_lams) source_labeling.cases)
  have "(source_labeling step',source_labeling step) \<in> (rstep R)^*" using assms unfolding step_dec step'_dec unfolding get_source_def by auto
  hence l2: "(source_labeling step,source_labeling step') \<in> (snd (sl_rel (R_d R) (R_nd R)))" using l_sl by auto
  hence l1: "(l step,l step') \<in> (snd r)" using assms unfolding ds_def by auto
  hence "(ll step,ll step') \<in> snd rr" using l2 unfolding lex_r_def ll_def rr_def by auto
  thus ?thesis unfolding ds_def by auto
qed

lemma l_imp_ll_seq_eq: (* symmetric to l_imp_ll_seq *)
  fixes l r R
  defines "ll \<equiv> (\<lambda> step. (source_labeling step,l step))"
  and "rr \<equiv> (lex_r (sl_rel (R_d R) (R_nd R)) r)"
  assumes "set (map l ss) \<subseteq> ds (snd r) {l step'}"
  and "\<forall>i < length ss. (get_source step' ,get_source (ss !i)) \<in> (rstep R)^*"
  shows "set (map ll ss) \<subseteq> ds (snd rr) {ll step'}"
using assms(3-4)
proof (induct ss)
  case (Cons t ts)  
  hence step:"l t \<in> ds (snd r) {l step'}" and steps:"set (map l ts) \<subseteq> snd r \<down>s {l step'}" by auto
  from Cons have c1: "\<forall>i<length ts. (get_source step', get_source (ts ! i)) \<in> (rstep R)\<^sup>*" by auto
  hence 2: "set (map ll ts) \<subseteq> snd rr \<down>s {ll step'}" using Cons(1)[OF steps c1] by auto
  have r1: "(get_source step', get_source t) \<in> (rstep R)\<^sup>*" using Cons assms by auto
  have 1: "ll t \<in> ds (snd rr) {ll step'}" using l_imp_ll_eq[OF step r1] unfolding ll_def rr_def by auto 
  show ?case using 1 2 by auto
qed auto
 
lemma r_ast_seq_imp_r_ast: 
  assumes "(s,t) \<in> (rstep R)^*" and "t = first ss" and "ss \<in> seq R" 
  shows "\<forall>i < length (snd ss). (s,get_source ((snd ss)!i)) \<in> (rstep R)^*"
proof -
  from assms obtain ts where ss_dec: "ss = (t,ts)" by (metis first.elims)
  from assms have ?thesis unfolding ss_dec apply auto
  proof (induct ts arbitrary: t )
    case (Cons step steps) 
    from Cons(3) obtain rl p \<sigma> u where step_dec: "step = (t,rl,p,\<sigma>,u)"
      using seq.cases by blast
    hence "(t,u) \<in> rstep_r_p_s R rl p \<sigma>" using Cons seq.cases by (metis (poly_guards_query) seq_chop_first(2))
    hence ast:"(s,u) \<in> (rstep R)^*" by (metis (poly_guards_query) Cons.prems(1) rstep_r_p_s_imp_rstep rtrancl.rtrancl_into_rtrancl)
    hence rst:"(u,steps) \<in> seq R" by (metis Cons.prems(2) seq_chop_first(1) step_dec)
    from Cons(1)[OF ast rst] have 2: "\<forall>i < length steps.(s, get_source (steps ! i)) \<in> (rstep R)\<^sup>*" by auto
    have 1: "(s,get_source (step)) \<in> (rstep R)^*" unfolding step_dec get_source_def using Cons by auto
    show ?case using 1 2 by (metis (erased, hide_lams) Cons.prems(3) One_nat_def add.commute add_Suc_right diff_Suc_Suc diff_zero gr0_conv_Suc list.size(4) monoid_add_class.add.right_neutral nat_add_left_cancel_less nth_Cons_pos nth_non_equal_first_eq)
  qed auto
  thus ?thesis unfolding ss_dec by auto
qed

lemma len_append: "\<forall> i < length xs. (xs ! i = (xs @ ys) ! i)" by (metis nth_append)

lemma eld_seq_imp_eld_seq:
  assumes ld_trs: "ld_trs R p jl jr" 
  and "eld_seq l r p jl jr"
  defines "ll \<equiv> (\<lambda> step. (source_labeling step,l step))"
  and "rr \<equiv> (lex_r (sl_rel (R_d R) (R_nd R)) r)"
  shows "eld_seq ll rr p jl jr"
proof -
  {
    fix jl \<sigma>1 \<sigma>2 \<sigma>3 lp rp ss
    assume A1: "map l (snd ss) = \<sigma>1@\<sigma>2@\<sigma>3" and A2:"ELD_1' r (l lp) (l rp) \<sigma>1 \<sigma>2 \<sigma>3"
      and A3: "ss \<in> seq R" and A4: "(get_source lp, first ss) \<in> (rstep R)^*" and A5: "get_source lp = get_source rp"
    from A2 have l:"length \<sigma>2 \<le> 1" unfolding ELD_1'_def by auto
    
    from A1 obtain ss1 ss2 ss3 where ss_dec:"snd ss = ss1@ss2@ss3" and dec1: "map l ss1 = \<sigma>1" and
      dec2: "map l ss2 = \<sigma>2"  and dec3:"map l ss3 = \<sigma>3" 
      using map_dec by metis
    hence 1: "set (map l ss1) \<subseteq> ds (fst r) {l lp,l lp}" and 2: "set (map l ss2) \<subseteq> ds (snd r) {l rp}" and
      3: "set (map l ss3) \<subseteq> ds (fst r) {l rp,l lp}" 
      using A2 unfolding ELD_1'_def by auto
    hence ll_dec: "map ll (snd ss) = (map ll ss1)@(map ll ss2)@(map ll ss3)" using ss_dec by (metis map_append)
    have l: "length (map ll ss2) \<le> 1" using l dec2 by auto
    
    have seq:"\<forall>i < length (snd ss). (get_source lp,get_source ((snd ss)!i)) \<in> (rstep R)^*"
      using r_ast_seq_imp_r_ast[OF A4 _ A3] by fast
    have s1: "\<forall>i<length ss1. (get_source lp, get_source (ss1 ! i)) \<in> (rstep R)\<^sup>*"
    proof (intro allI impI)
      fix i
      assume A: "i < length ss1" 
      hence eq: "((ss1@ss2@ss3)!i) = (ss1!i)" using nth_append by metis
      thus "(get_source lp, get_source (ss1 !i)) \<in> (rstep R)^*" using seq unfolding ss_dec A unfolding get_source_def eq 
        apply auto by (metis (erased, hide_lams) A add.commute len_append trans_less_add2)
    qed
    have s2: "\<forall>i<length ss2. (get_source rp, get_source (ss2 ! i)) \<in> (rstep R)\<^sup>*"
    proof (intro allI impI)
      fix i
      assume A: "i < length ss2"
      hence eq: "((ss1@ss2@ss3)!(length ss1+i)) =  (ss2!i)" by (metis len_append nth_append_length_plus)
      thus "(get_source rp, get_source (ss2 !i)) \<in> (rstep R)^*" using seq unfolding ss_dec A A5 unfolding get_source_def eq 
        apply auto by (metis (erased, hide_lams) A eq le_add1 less_le_trans nat_add_left_cancel_less)
    qed
    have s3: "\<forall>i<length ss3. (get_source lp, get_source (ss3 ! i)) \<in> (rstep R)\<^sup>*"
    proof (intro allI impI)
      fix i
      assume A: "i < length ss3"
      hence eq: "((ss1@ss2@ss3)!(length (ss1@ss2)+i)) =  (ss3!i)" using List.append_assoc[symmetric] by (metis nth_append_length_plus)
      thus "(get_source lp, get_source (ss3 !i)) \<in> (rstep R)^*" using seq unfolding ss_dec A
        unfolding get_source_def eq List.append_assoc[symmetric]
        by (metis (erased, hide_lams) A length_append nat_add_left_cancel_less)
    qed
    have s3': "\<forall>i<length ss3. (get_source rp, get_source (ss3 ! i)) \<in> (rstep R)\<^sup>*" using s3 A5 by auto 
    have "set (map ll ss1) \<subseteq> ds (fst rr) {ll lp}" 
      and "set (map ll ss2) \<subseteq> ds (snd rr) {ll rp}" 
      and "set (map ll ss3) \<subseteq> ds (fst rr) {ll rp,ll lp}"
      using l_imp_ll_seq[OF 1 s1 s1] using l_imp_ll_seq_eq[OF 2 s2] using l_imp_ll_seq[OF 3 s3' s3] unfolding ll_def rr_def by auto
    hence "\<exists> s1 s2 s3. (map ll (snd ss) = s1@s2@s3 \<and> ELD_1' rr (ll lp) (ll rp) s1 s2 s3)"
      using ll_dec unfolding ELD_1'_def using l by fast
  } note F = this
  from assms obtain s rl q \<sigma> t rl2 q2 \<sigma>2 t2 where p_dec: "p = ((s,rl,q,\<sigma>,t),(s,rl2,q2,\<sigma>2,t2))" 
    and step1: "(s,t) \<in> rstep_r_p_s R rl q \<sigma>" and step2: "(s,t2) \<in> rstep_r_p_s R rl2 q2 \<sigma>2"
    using assms unfolding ld_trs_def local_peaks_def get_source_def get_target_def by auto
  from assms obtain \<sigma>1 \<sigma>2 \<sigma>3 where decl: "map l (snd jl) = \<sigma>1@\<sigma>2@\<sigma>3" 
    and ELDl: "ELD_1' r (l (fst p)) (l (snd p)) \<sigma>1 \<sigma>2 \<sigma>3" and seql: "jl \<in> seq R" unfolding ld_trs_def eld_seq_def by fast
  have "(get_source (fst p),get_target (fst p)) \<in> rstep R" using step1
    unfolding p_dec get_source_def get_target_def apply auto by (metis rstep_r_p_s_imp_rstep)
  hence stepl: "(get_source (fst p), first jl) \<in> (rstep R)\<^sup>*" using assms unfolding ld_trs_def by auto
  have "(get_source (snd p),get_target (snd p)) \<in> rstep R" using step2
    unfolding p_dec get_source_def get_target_def apply auto by (metis rstep_r_p_s_imp_rstep)
  hence stepr: "(get_source (snd p), first jr) \<in> (rstep R)\<^sup>*" using assms unfolding ld_trs_def by auto
  from assms obtain \<tau>1 \<tau>2 \<tau>3 where decr: "map l (snd jr) = \<tau>1@\<tau>2@\<tau>3" 
    and ELDr: "ELD_1' r (l (snd p)) (l (fst p)) \<tau>1 \<tau>2 \<tau>3" and seqr: "jr \<in> seq R" unfolding ld_trs_def eld_seq_def by fast
  have A5: "get_source (fst p) = get_source (snd p)" using ld_trs unfolding ld_trs_def local_peaks_def get_source_def by auto
  show ?thesis unfolding ld_trs_def local_peaks_def eld_seq_def 
    using F[OF decl ELDl seql stepl A5] F[OF decr ELDr seqr stepr A5[symmetric]] by blast
qed
 
(* eld for l implies eld for (l_sn \<times> l) *)
lemma eld_imp_eld:
  assumes "eld R l r p"
  defines "ll \<equiv> (\<lambda> step. (source_labeling step,l step))"
  and "rr \<equiv> (lex_r (sl_rel (R_d R) (R_nd R)) r)"
  shows "eld R ll rr p"
proof -
  from assms obtain jl jr where d:"ld_trs R p jl jr" and ELD:"eld_seq l r p jl jr" unfolding eld_def by auto
  show ?thesis using d eld_seq_imp_eld_seq[OF d ELD] unfolding ll_def rr_def eld_def by fast
qed

lemma rel_props_lex:
  assumes "rel_props (fst r1) (snd r1)"
  and   "rel_props (fst r2) (snd r2)"
  defines "r \<equiv> lex_r r1 r2"
  shows  "rel_props (fst r) (snd r)"
proof -
  have t1: "trans (fst r1)"
    and t1': "trans (snd r1)"
    and t2: "trans (fst r2)" 
    and t2': "trans (snd r2)"
    using assms unfolding rel_props_def by auto
  (* transitivity1 *)
  {
    fix x1 x2 y1 y2 z1 z2 
    assume A: "((x1,x2),(y1,y2)) \<in> fst r" and B: "((y1,y2),(z1,z2)) \<in> fst r"
    have "((x1,x2),(z1,z2)) \<in> fst r"
    proof (cases "(x1,y1) \<in> fst r1")
      case True note T1 = this
      hence "(x1,z1) \<in> fst r1"
      proof (cases "(y1,z1) \<in> fst r1")
        case True
        thus ?thesis using T1 t1 trans_def by metis
      next
        case False
        hence T2: "(y1,z1) \<in> snd r1" using B unfolding r_def lex_r_def by auto
        have "(x1,x1) \<in> snd r1" using assms(1) unfolding rel_props_def by (metis UNIV_I refl_on_def)
        thus ?thesis using T1 assms(1) T2 unfolding rel_props_def r_def lex_r_def t1 by auto
      qed
      thus ?thesis unfolding r_def lex_r_def by auto
    next
      case False
      hence T1: "(x1,y1) \<in> snd r1" and T2: "(x2,y2) \<in> fst r2" using A unfolding r_def lex_r_def by auto
      thus ?thesis
      proof (cases "(y1,z1) \<in> fst r1")
      next
        case True
        have T3: "(z1,z1) \<in> snd r1" using assms(1) unfolding rel_props_def by (metis UNIV_I refl_on_def)
        thus ?thesis using T1 True T3 assms(1-2) unfolding rel_props_def r_def lex_r_def by auto
      next
        case False
        hence T3: "(y1,z1) \<in> snd r1" and T4: "(y2,z2) \<in> fst r2" using B unfolding r_def lex_r_def by auto
        have "(x1,z1) \<in> snd r1" using T1 T3 t1' unfolding trans_def by metis
        moreover have "(x2,z2) \<in> fst r2" using T2 T4 t2 unfolding trans_def by metis
        ultimately show ?thesis unfolding  r_def lex_r_def by auto
      qed
    qed
  }
  hence tA: "trans (fst r)" unfolding trans_def by (metis prod.collapse)
  (* transitivity2 *)
  {
    fix x1 x2 y1 y2 z1 z2 (* symmetric of above *) 
    assume A: "((x1,x2),(y1,y2)) \<in> snd r" and B: "((y1,y2),(z1,z2)) \<in> snd r"
    have "((x1,x2),(z1,z2)) \<in> snd r"
    proof (cases "(x1,y1) \<in> fst r1")
      case True note T1 = this
      hence "(x1,z1) \<in> fst r1"
      proof (cases "(y1,z1) \<in> fst r1")
        case True 
        thus ?thesis using T1 t1 unfolding trans_def  by metis 
      next
        case False
        hence T2: "(y1,z1) \<in> snd r1" using B unfolding r_def lex_r_def by auto
        have "(x1,x1) \<in> snd r1" using assms(1) unfolding rel_props_def by (metis UNIV_I refl_on_def)
        thus ?thesis using T1 assms(1) T2 unfolding rel_props_def r_def lex_r_def t1 by auto
      qed
      thus ?thesis unfolding r_def lex_r_def by auto
    next
      case False
      hence T1: "(x1,y1) \<in> snd r1" and T2: "(x2,y2) \<in> snd r2" using A unfolding r_def lex_r_def by auto
      thus ?thesis
      proof (cases "(y1,z1) \<in> fst r1")
        case True
        have T3: "(z1,z1) \<in> snd r1" using assms(1) unfolding rel_props_def by (metis UNIV_I refl_on_def)
        thus ?thesis using T1 True T3 assms(1-2) unfolding rel_props_def r_def lex_r_def by auto
      next
        case False
        hence T3: "(y1,z1) \<in> snd r1" and T4: "(y2,z2) \<in> snd r2" using B unfolding r_def lex_r_def by auto
        have "(x1,z1) \<in> snd r1" using T1 T3 t1' unfolding trans_def by metis
        moreover have "(x2,z2) \<in> snd r2" using T2 T4 t2' unfolding trans_def by metis
        ultimately show ?thesis unfolding  r_def lex_r_def by auto
      qed
    qed
  }
  hence tB: "trans (snd r)"  unfolding trans_def by (metis prod.collapse)
  (* termination *)
  let ?s1 = "\<lambda> a b. (a,b) \<in> (fst r1)\<inverse>"
  let ?ns1 = "\<lambda> a b. (a,b) \<in> (snd r1)\<inverse>"
  let ?s2 = "\<lambda> a b. (a,b) \<in> (fst r2)\<inverse>"
  have SN1: "SN ((fst r1)\<inverse>)" using assms unfolding rel_props_def by (metis SN_iff_wf converse_converse)
  hence SNx: "SN {(a,b). ?s1 a b}" using SN1 apply auto by (metis converse_unfold)
  have SN2: "SN ((fst r2)\<inverse>)" using assms unfolding rel_props_def by (metis SN_iff_wf converse_converse)
  hence SNy: "SN {(a,b). ?s2 a b}" using SN2 apply auto by (metis converse_unfold)
  have compat: "\<And> a b c. (?ns1 a b \<Longrightarrow> ?s1 b c \<Longrightarrow> ?s1 a c)"
  proof -
    fix a b c
    assume A: "?ns1 a b" and B: "?s1 b c"
    have C: "(c,c) \<in> snd r1" using assms unfolding rel_props_def by (metis UNIV_I refl_on_def)
    show "?s1 a c" using A B C using assms unfolding rel_props_def by auto
  qed
  have SN: "SN {(a,b). lex_two ?s1 ?ns1 ?s2 a b}" (is "SN ?R") using lex_two[OF compat SNx SNy] by auto
  hence "?R = ((fst r)\<inverse>)" unfolding r_def lex_r_def by auto
  hence "SN ((fst r)\<inverse>)" using SN unfolding r_def lex_r_def by auto 
  hence wf: "wf ((fst r))" by (metis SN_iff_wf converse_converse)
  (* reflexivity *)
  {
    fix x1 x2
    assume "(x1,x1) \<in> snd r1" and "(x2,x2) \<in> snd r2"
    hence "((x1,x2),(x1,x2)) \<in> snd r" unfolding r_def lex_r_def by auto 
  } note r_key = this
  have r1: "\<forall> x1 \<in> UNIV. (x1,x1) \<in> snd r1"
    using assms unfolding rel_props_def by (metis (erased, lifting) UNIV_I case_prodI2 refl_on_def)
  have r2: "\<forall> x2 \<in> UNIV. (x2,x2) \<in> snd r2"
    using assms unfolding rel_props_def by (metis (erased, lifting) UNIV_I case_prodI2 refl_on_def)
  have r: "\<forall> (x1,x2) \<in> UNIV. ((x1,x2),x1,x2) \<in> snd r" using r_key r1 r2 by auto 
  hence r: "refl (snd r)" by (metis (erased, lifting) UNIV_I old.prod.case prod.collapse refl_on_def' snd_conv) 
  (* compatibility *)
  {
    fix x1 x2 y1 y2 z1 z2 v1 v2 
    assume A: "((x1,x2),(y1,y2)) \<in> snd r" and B: "((y1,y2),(z1,z2)) \<in> fst r" and C: "((z1,z2),(v1,v2)) \<in> snd r"
    have D: "((x1,x2),(z1,z2)) \<in> fst r"
    proof (cases "(x1,y1) \<in> fst r1") 
      case True note T1 = this
      show ?thesis
      proof (cases "(y1,z1) \<in> fst r1")
        case True
        hence "(x1,z1) \<in> fst r1" using T1 t1 unfolding trans_def by metis
        thus ?thesis unfolding r_def lex_r_def by auto
      next
        case False
        hence "(y1,z1) \<in> snd r1" using B unfolding r_def lex_r_def by auto
        moreover have "(x1,x1) \<in> snd r1" using assms unfolding rel_props_def by (metis UNIV_I r1)
        ultimately have "(x1,z1) \<in> fst r1" using T1 assms unfolding rel_props_def by auto
        thus ?thesis unfolding r_def lex_r_def by auto
      qed
    next
      case False
      hence T2: "(x1,y1) \<in> snd r1" and W1:"(x2,y2) \<in> snd r2" using A unfolding r_def lex_r_def by auto
      thus ?thesis
      proof (cases "(y1,z1) \<in> fst r1")
        case True
        have "(z1,z1) \<in> snd r1" using assms unfolding rel_props_def by (metis UNIV_I r1)
        hence "(x1,z1) \<in> fst r1" using T2 True assms unfolding rel_props_def by auto
        thus ?thesis unfolding r_def lex_r_def by auto
      next
        case False
        hence "(y1,z1) \<in> snd r1" and W2:"(y2,z2) \<in> fst r2" using B unfolding r_def lex_r_def by auto
        hence k1: "(x1,z1) \<in> snd r1" using T2 t1' unfolding trans_def by metis
        hence "(z2,z2) \<in> snd r2" using assms unfolding rel_props_def by (metis UNIV_I r2)
        hence "(x2,z2) \<in> fst r2" using W1 W2 assms unfolding rel_props_def by auto
        thus ?thesis using k1 unfolding r_def lex_r_def by auto
      qed
    qed
    have "((x1,x2),(v1,v2)) \<in> fst r"
    proof (cases "(x1,z1) \<in> fst r1") 
      case True note T1 = this
      thus ?thesis
      proof (cases "(z1,v1) \<in> fst r1")
        case True
        hence "(x1,v1) \<in> fst r1" using T1 t1 unfolding trans_def by metis
        thus ?thesis unfolding r_def lex_r_def by auto
      next
        case False
        hence T2: "(z1,v1) \<in> snd r1" using C unfolding r_def lex_r_def by auto
        have "(x1,x1) \<in> snd r1" using assms unfolding rel_props_def by (metis UNIV_I r1)
        hence "(x1,v1) \<in> fst r1" using T1 T2 assms unfolding rel_props_def by auto
        thus ?thesis unfolding r_def lex_r_def by auto
      qed
    next
      case False
      hence T1: "(x1,z1) \<in> snd r1" and W0: "(x2,z2) \<in> fst r2" using D unfolding r_def lex_r_def by auto
      thus ?thesis
      proof (cases "(z1,v1) \<in> fst r1")
        case True 
        have "(v1,v1) \<in> snd r1" using assms unfolding rel_props_def by (metis UNIV_I r1)
        hence "(x1,v1) \<in> fst r1" using assms T1 True unfolding rel_props_def by auto
        thus ?thesis unfolding r_def lex_r_def by auto
      next
        case False
        have W2: "(x2,x2) \<in> snd r2" using assms unfolding rel_props_def using UNIV_I r2 by metis
        from False have "(z1,v1) \<in> snd r1" and W1: "(z2,v2) \<in> snd r2" using C unfolding r_def lex_r_def by auto
        hence "(x1,v1) \<in> snd r1" using T1 t1' unfolding trans_def by metis
        moreover have "(x2,v2) \<in> fst r2" using W0 W1 W2 using assms unfolding rel_props_def by auto
        ultimately show ?thesis unfolding r_def lex_r_def by auto
      qed
    qed
  }
  hence compat: "snd r O fst r O snd r \<subseteq> fst r" by auto
  show ?thesis using tA tB wf r compat unfolding rel_props_def by simp
qed
  
lemma SN_rel_has_rel_props:
  assumes SN: "SN_rel (rstep (R_d R)) (rstep (R_nd R))"
  defines "r \<equiv> (sl_rel (R_d R) (R_nd R))" 
  shows "rel_props (fst r) (snd r)"
proof -
  have t1: "trans (fst r)" unfolding r_def sl_rel.simps fst_conv by (metis trans_converse trans_trancl)
  have t2: "trans (snd r)" unfolding r_def sl_rel.simps snd_conv by (metis trans_converse trans_rtrancl)
  have "SN (((rstep (R_nd R))\<^sup>* O rstep (R_d R) O (rstep (R_nd R))\<^sup>*))" using SN unfolding r_def sl_rel.simps fst_conv by (metis SN_rel_on_def)
  hence wf: "wf (fst r)" unfolding r_def sl_rel.simps fst_conv by (metis SN_iff_wf wf_converse_trancl)
  have r: "refl (snd r)" unfolding r_def sl_rel.simps by auto
  have compat: "snd r O fst r O snd r \<subseteq> fst r" unfolding r_def apply auto by (metis RS_S R_d_union_R_nd S_RS S_RS_trancl inf_sup_aci(5) rel_rstep_def relto_trancl_conv rstep_union rtrancl.rtrancl_refl)  
  show ?thesis unfolding rel_props_def using t1 t2 wf r compat by auto
qed

lemma is_labeling_lex:
  assumes "is_labeling R l1 r1"
  and "is_labeling R l2 r2"
  shows "is_labeling R (\<lambda> step. (l1 step,l2 step)) (lex_r r1 r2)"
using assms unfolding is_labeling_def Let_def lex_r_def by auto

lemma SN_rel_is_labeling:
  shows "is_labeling R source_labeling (sl_rel (R_d R) (R_nd R))" (is "is_labeling ?R ?l ?r")
proof -
  have eq: "R = (R_d R) \<union> (R_nd R)" unfolding R_d_def R_nd_def by auto 
  show ?thesis using sl_is_labeling eq by metis
qed

 (* similar to l_imp_ll_eq*)
lemma l_imp_ll_eq3:
  fixes l1 l2 r1 r2 R
  defines "ll \<equiv> (\<lambda> step. (l1 step,l2 step))"
  and "rr \<equiv> (lex_r r1 r2)"
  assumes "l1 step \<in> ds (snd r1) {l1 step'}"
  and "l2 step \<in> ds (snd r2) {l2 step'}" 
  and "(get_source step',get_source step) \<in> (rstep R)^*"
  shows "ll step \<in> ds (snd rr) {ll step'}"
unfolding ll_def rr_def using assms unfolding lex_r_def ds_def by auto 
 
 (* similar to ll_imp_ll_seq_eq *)
lemma l_imp_ll_seq_eq3: (* symmetric to l_imp_ll_seq *)
  fixes l1 l r1 r R
  defines "ll \<equiv> (\<lambda> step. (l step,l1 step))"
  and "rr \<equiv> (lex_r r r1)"
  assumes "set (map l ss) \<subseteq> ds (snd r) {l step'}"
  and "set (map l1 ss) \<subseteq> ds (snd r1) {l1 step'}"
  and "\<forall>i < length ss. (get_source step' ,get_source (ss !i)) \<in> (rstep R)^*"
  shows "set (map ll ss) \<subseteq> ds (snd rr) {ll step'}"
using assms(3-5)
proof (induct ss)
  case Nil thus ?case by auto
next
  case (Cons t ts)  
  hence step:"l t \<in> ds (snd r) {l step'}" and steps:"set (map l ts) \<subseteq> snd r \<down>s {l step'}" by auto
  from Cons have step':"l1 t \<in> ds (snd r1) {l1 step'}" and steps':"set (map l1 ts) \<subseteq> snd r1 \<down>s {l1 step'}" by auto
  from Cons have c1: "\<forall>i<length ts. (get_source step', get_source (ts ! i)) \<in> (rstep R)\<^sup>*" by auto
  hence 2: "set (map ll ts) \<subseteq> snd rr \<down>s {ll step'}" using Cons(1)[OF steps steps' c1] by auto
  have r1: "(get_source step', get_source t) \<in> (rstep R)\<^sup>*" using Cons assms by auto
  have 1: "ll t \<in> ds (snd rr) {ll step'}" using l_imp_ll_eq3[OF step step' r1] unfolding ll_def rr_def by blast 
  show ?case using 1 2 by auto
qed
 
lemma l_imp_ll3:
  fixes l l1 r r1 R
  defines "ll \<equiv> (\<lambda> step. (l step,l1 step))"
  and "rr \<equiv> (lex_r r r1)"
  assumes "l step  \<in> ds (fst r) {l step', l step''}"
  and "(get_source step',get_source step) \<in> (rstep R)^*"
  and "(get_source step'',get_source step) \<in> (rstep R)^*"
  shows "ll step \<in> ds (fst rr) {ll step', ll step''}"
using assms unfolding ds_def ll_def rr_def lex_r_def by auto
 
 (* similar to l_imp_ll_seq *)
lemma l_imp_ll_seq3:
  fixes l1 l r1 r R
  defines "ll \<equiv> (\<lambda> step. (l step,l1 step))"
  and "rr \<equiv> (lex_r r r1)"
  assumes "set (map l ss) \<subseteq> ds (fst r) {l step',l step''}"
  and "\<forall>i < length ss. (get_source step' ,get_source (ss !i)) \<in> (rstep R)^*"
  and "\<forall>i < length ss. (get_source step'',get_source (ss !i)) \<in> (rstep R)^*"
  shows "set (map ll ss) \<subseteq> ds (fst rr) {ll step',ll step''}"
using assms(3-)
proof (induct ss)
  case Nil thus ?case by auto
next
  case (Cons t ts)  
  hence step:"l t \<in> ds (fst r) {l step',l step''}" and steps:"set (map l ts) \<subseteq> fst r \<down>s {l step', l step''}" by auto
  from Cons have c1: "\<forall>i<length ts. (get_source step', get_source (ts ! i)) \<in> (rstep R)\<^sup>*" by auto
  from Cons have c2: "\<forall>i<length ts. (get_source step'', get_source (ts ! i)) \<in> (rstep R)\<^sup>*" by auto
  hence 2: "set (map ll ts) \<subseteq> fst rr \<down>s {ll step', ll step''}" using Cons(1)[OF steps (* steps'*) c1 c2] by auto
  have r1: "(get_source step', get_source t) \<in> (rstep R)\<^sup>*" using Cons assms by auto
  have r2: "(get_source step'', get_source t) \<in> (rstep R)\<^sup>*" using Cons assms by auto 
  have 1: "ll t \<in> ds (fst rr) {ll step', ll step''}" using l_imp_ll3[OF step  r1 r2] unfolding ll_def rr_def by fast
  show ?case using 1 2 by auto
qed

lemma eld_seq_and_weld_seq_imp_eld_seq:
  assumes "ld2_trs R p jl jr" "eld_seq l1 r1 p jl jr" and "weld_seq l2 r2 p jl jr"
  shows "eld_seq (\<lambda> step. (l1 step,l2 step)) (lex_r r1 r2) p jl jr"
proof -
  (* similar to proof from eld_seq_imp_eld_seq *)
  {
    fix l1 l2 jl \<sigma>1 \<sigma>2 \<sigma>3 lp rp ss
    assume A1: "map l1 (snd ss) = \<sigma>1@\<sigma>2@\<sigma>3" and A2:"ELD_1' r1 (l1 lp) (l1 rp) \<sigma>1 \<sigma>2 \<sigma>3"
      and A3: "ss \<in> seq R" and A4: "(get_source lp, first ss) \<in> (rstep R)^*"
      and A5: "get_source lp = get_source rp"
      and A6: "set (map l2 (snd ss)) \<subseteq> snd r2 \<down>s {l2 rp}"
    from A2 have l:"length \<sigma>2 \<le> 1" unfolding ELD_1'_def by auto
    let ?l = "\<lambda>step. (l1 step, l2 step)"
    from A1 obtain ss1 ss2 ss3 where ss_dec:"snd ss = ss1@ss2@ss3" and dec1: "map l1 ss1 = \<sigma>1"
      and dec2: "map l1 ss2 = \<sigma>2"  and dec3:"map l1 ss3 = \<sigma>3" 
      using map_dec by metis
    hence 1: "set (map l1 ss1) \<subseteq> ds (fst r1) {l1 lp,l1 lp}"
      and 2: "set (map l1 ss2) \<subseteq> ds (snd r1) {l1 rp}"
      and 3: "set (map l1 ss3) \<subseteq> ds (fst r1) {l1 rp,l1 lp}" 
      using A2 unfolding ELD_1'_def by auto
    hence ll_dec: "map ?l (snd ss) = (map ?l ss1)@(map ?l ss2)@(map ?l ss3)"
      using ss_dec by (metis map_append)
    have l: "length (map ?l ss2) \<le> 1" using l dec2 by auto
    have seq:"\<forall>i < length (snd ss). (get_source lp,get_source ((snd ss)!i)) \<in> (rstep R)^*"
      using r_ast_seq_imp_r_ast[OF A4 _ A3] by fast
    have s1: "\<forall>i<length ss1. (get_source lp, get_source (ss1 ! i)) \<in> (rstep R)\<^sup>*"
    proof (intro allI impI)
      fix i
      assume A: "i < length ss1" 
      hence eq: "((ss1@ss2@ss3)!i) = (ss1!i)" using nth_append by metis
      thus "(get_source lp, get_source (ss1 !i)) \<in> (rstep R)^*"
        using seq unfolding ss_dec A unfolding get_source_def eq 
        apply auto by (metis (erased, hide_lams) A add.commute len_append trans_less_add2)
    qed
    have s2: "\<forall>i<length ss2. (get_source rp, get_source (ss2 ! i)) \<in> (rstep R)\<^sup>*"
    proof (intro allI impI)
      fix i
      assume A: "i < length ss2"
      hence eq: "((ss1@ss2@ss3)!(length ss1+i)) =  (ss2!i)" by (metis len_append nth_append_length_plus)
      thus "(get_source rp, get_source (ss2 !i)) \<in> (rstep R)^*"
        using seq unfolding ss_dec A A5 unfolding get_source_def eq 
        apply auto by (metis (erased, hide_lams) A eq le_add1 less_le_trans nat_add_left_cancel_less)
    qed
    have s3: "\<forall>i<length ss3. (get_source lp, get_source (ss3 ! i)) \<in> (rstep R)\<^sup>*"
    proof (intro allI impI)
      fix i assume A: "i < length ss3"
      hence eq: "((ss1@ss2@ss3)!(length (ss1@ss2)+i)) =  (ss3!i)"
        using List.append_assoc[symmetric] by (metis len_append nth_append_length_plus append_assoc)
      thus "(get_source lp, get_source (ss3 !i)) \<in> (rstep R)^*"
        using seq unfolding ss_dec A unfolding get_source_def eq List.append_assoc[symmetric]
        by (metis (erased, hide_lams) A length_append nat_add_left_cancel_less)
    qed
    have s3': "\<forall>i<length ss3. (get_source rp, get_source (ss3 ! i)) \<in> (rstep R)\<^sup>*" using s3 A5 by auto 
    have x2: "set (map l2 ss2) \<subseteq> snd r2 \<down>s {l2 rp}" using A6 unfolding ss_dec by auto
    have "set (map (\<lambda>step. (l1 step, l2 step)) ss1) \<subseteq> ds (fst (lex_r r1 r2)) {(\<lambda>step. (l1 step, l2 step)) lp,(\<lambda>step. (l1 step, l2 step)) lp}"
      using l_imp_ll_seq3[OF 1 s1 s1] by fast
    hence "set (map ?l ss1) \<subseteq> ds (fst (lex_r r1 r2)) {?l lp}" by auto
    moreover have "set (map ?l ss2) \<subseteq> ds (snd (lex_r r1 r2)) {?l rp}"
      using l_imp_ll_seq_eq3[OF 2 x2 s2] by fast
    moreover have "set (map ?l ss3) \<subseteq> ds (fst (lex_r r1 r2)) {?l rp,?l lp}"
      using l_imp_ll_seq3[OF 3 s3' s3] by fast
    ultimately have "\<exists> s1 s2 s3. (map ?l (snd ss) = s1@s2@s3 \<and> 
      ELD_1' (lex_r r1 r2) (?l lp) (?l rp) s1 s2 s3)"
      unfolding ll_dec ELD_1'_def using l by fast
  } note F = this
  from assms obtain s rl q \<sigma> t rl2 q2 \<sigma>2 t2 where p_dec: "p = ((s,rl,q,\<sigma>,t),(s,rl2,q2,\<sigma>2,t2))" 
    and step1: "(s,t) \<in> rstep_r_p_s R rl q \<sigma>" and step2: "(s,t2) \<in> rstep_r_p_s R rl2 q2 \<sigma>2"
    using assms unfolding ld2_trs_def ld_trs_def local_peaks_def get_source_def get_target_def by auto
  from assms obtain \<sigma>1 \<sigma>2 \<sigma>3 where decl: "map l1 (snd jl) = \<sigma>1@\<sigma>2@\<sigma>3" 
    and ELDl: "ELD_1' r1 (l1 (fst p)) (l1 (snd p)) \<sigma>1 \<sigma>2 \<sigma>3" and seql: "jl \<in> seq R"
    unfolding ld2_trs_def ld_trs_def eld_seq_def by fast
  have "(get_source (fst p),get_target (fst p)) \<in> rstep R"
    using step1 unfolding p_dec get_source_def get_target_def apply auto by (metis rstep_r_p_s_imp_rstep)
  hence stepl: "(get_source (fst p), first jl) \<in> (rstep R)\<^sup>*"
    using assms unfolding ld2_trs_def ld_trs_def by auto
  have "(get_source (snd p),get_target (snd p)) \<in> rstep R"
    using step2 unfolding p_dec get_source_def get_target_def apply auto by (metis rstep_r_p_s_imp_rstep)
  hence stepr: "(get_source (snd p), first jr) \<in> (rstep R)\<^sup>*"
    using assms unfolding ld2_trs_def ld_trs_def by auto
  from assms obtain \<tau>1 \<tau>2 \<tau>3 where decr: "map l1 (snd jr) = \<tau>1@\<tau>2@\<tau>3" 
    and ELDr: "ELD_1' r1 (l1 (snd p)) (l1 (fst p)) \<tau>1 \<tau>2 \<tau>3" and seqr: "jr \<in> seq R"
    unfolding ld2_trs_def ld_trs_def eld_seq_def by fast
  have A5: "get_source (fst p) = get_source (snd p)"
    using assms unfolding ld2_trs_def ld_trs_def local_peaks_def get_source_def by auto
  have A6: "set (map l2 (snd jl)) \<subseteq> snd r2 \<down>s {l2 (snd p)}"
    using assms unfolding weld_seq_def labels_def by auto
  have A6': "set (map l2 (snd jr)) \<subseteq> snd r2 \<down>s {l2 (fst p)}"
    using assms unfolding weld_seq_def labels_def by auto
  show ?thesis unfolding ld_trs_def local_peaks_def eld_seq_def 
    using F[OF decl ELDl seql stepl A5 A6] F[OF decr ELDr seqr stepr A5[symmetric] A6'] by fast
qed

 
(* (* this direction is currently not used *) 
 (* similar to l_imp_ll_seq *)
lemma l_imp_ll_seq2:
 fixes l1 l r1 r R
 defines "ll \<equiv> (\<lambda> step. (l1 step,l step))"
 and "rr \<equiv> (lex_r r1 r)"
 assumes "set (map l ss) \<subseteq> ds (fst r) {l step',l step''}"
  and "\<forall>i < length ss. (get_source step' ,get_source (ss !i)) \<in> (rstep R)^*"
  and "\<forall>i < length ss. (get_source step'',get_source (ss !i)) \<in> (rstep R)^*"
 shows "set (map ll ss) \<subseteq> ds (fst rr) {ll step',ll step''}" sorry
 
  (* similar to ll_imp_ll_seq_eq *)
lemma l_imp_ll_seq_eq2: (* symmetric to l_imp_ll_seq *)
 fixes l1 l r1 r R
 defines "ll \<equiv> (\<lambda> step. (l1 step,l step))"
 and "rr \<equiv> (lex_r r1 r)"
 assumes "set (map l ss) \<subseteq> ds (snd r) {l step'}"
  and "\<forall>i < length ss. (get_source step' ,get_source (ss !i)) \<in> (rstep R)^*"
   shows "set (map ll ss) \<subseteq> ds (snd rr) {ll step'}" sorry


 
lemma weld_seq_and_eld_seq_imp_eld_seq:
 assumes "ld2_trs R p jl jr" "eld_seq l2 r2 p jl jr" and "weld_seq l1 r1 p jl jr"
 shows "eld_seq (\<lambda> step. (l1 step,l2 step)) (lex_r r1 r2) p jl jr" (is "eld_seq ?l ?r _ _ _") proof -
 (* from eld_seq_imp_eld_seq *)
  {fix l1 l2 r1 r2 jl \<sigma>1 \<sigma>2 \<sigma>3 lp rp ss
    assume A1: "map l2 (snd ss) = \<sigma>1@\<sigma>2@\<sigma>3" and A2:"ELD_1' r2 (l2 lp) (l2 rp) \<sigma>1 \<sigma>2 \<sigma>3"
     and A3: "ss \<in> seq R" and A4: "(get_source lp, first ss) \<in> (rstep R)^*" and A5: "get_source lp = get_source rp"

   from A2 have l:"length \<sigma>2 \<le> 1" unfolding ELD_1'_def by auto
    let ?l = "\<lambda>step. (l1 step, l2 step)"
  from A1 obtain ss1 ss2 ss3 where ss_dec:"snd ss = ss1@ss2@ss3" and dec1: "map l2 ss1 = \<sigma>1" and dec2: "map l2 ss2 = \<sigma>2"  and dec3:"map l2 ss3 = \<sigma>3" 
   using map_dec by metis
  hence 1: "set (map l2 ss1) \<subseteq> ds (fst r2) {l2 lp,l2 lp}" and 2: "set (map l2 ss2) \<subseteq> ds (snd r2) {l2 rp}" and 3: "set (map l2 ss3) \<subseteq> ds (fst r2) {l2 rp,l2 lp}" 
   using A2 unfolding ELD_1'_def by auto
   
  hence ll_dec: "map ?l (snd ss) = (map ?l ss1)@(map ?l ss2)@(map ?l ss3)" using ss_dec by (metis map_append)
  have l: "length (map ?l ss2) \<le> 1" using l dec2 by auto
  
  have seq:"\<forall>i < length (snd ss). (get_source lp,get_source ((snd ss)!i)) \<in> (rstep R)^*" using r_ast_seq_imp_r_ast[OF A4 _ A3] by fast
  have s1: "\<forall>i<length ss1. (get_source lp, get_source (ss1 ! i)) \<in> (rstep R)\<^sup>*" apply auto proof - 
   fix i assume A: "i < length ss1" 
    hence eq: "((ss1@ss2@ss3)!i) = (ss1!i)" using nth_append by metis
    thus "(get_source lp, get_source (ss1 !i)) \<in> (rstep R)^*" using seq unfolding ss_dec A unfolding get_source_def eq 
     apply auto by (metis (erased, hide_lams) A add.commute len_append trans_less_add2)
  qed
  have s2: "\<forall>i<length ss2. (get_source rp, get_source (ss2 ! i)) \<in> (rstep R)\<^sup>*" apply auto proof -
   fix i assume A: "i < length ss2"
   hence eq: "((ss1@ss2@ss3)!(length ss1+i)) =  (ss2!i)" by (metis len_append nth_append_length_plus)
       thus "(get_source rp, get_source (ss2 !i)) \<in> (rstep R)^*" using seq unfolding ss_dec A A5 unfolding get_source_def eq 
     apply auto by (metis (erased, hide_lams) A eq le_add1 less_le_trans nat_add_left_cancel_less)
  qed
  have s3: "\<forall>i<length ss3. (get_source lp, get_source (ss3 ! i)) \<in> (rstep R)\<^sup>*" apply auto proof -
   fix i assume A: "i < length ss3"
   hence eq: "((ss1@ss2@ss3)!(length (ss1@ss2)+i)) =  (ss3!i)" using List.append_assoc[symmetric] by (metis len_append nth_append_length_plus append_assoc)
       thus "(get_source lp, get_source (ss3 !i)) \<in> (rstep R)^*" using seq unfolding ss_dec A unfolding get_source_def eq List.append_assoc[symmetric]
       by (metis (erased, hide_lams) A length_append nat_add_left_cancel_less)
  qed
  have s3': "\<forall>i<length ss3. (get_source rp, get_source (ss3 ! i)) \<in> (rstep R)\<^sup>*" using s3 A5 by auto 


  have "set (map (\<lambda>step. (l1 step, l2 step)) ss1) \<subseteq> ds (fst (lex_r r1 r2)) {(\<lambda>step. (l1 step, l2 step)) lp,(\<lambda>step. (l1 step, l2 step)) lp}" using l_imp_ll_seq2[OF 1 s1 s1] by fast
  hence "set (map ?l ss1) \<subseteq> ds (fst (lex_r r1 r2)) {?l lp}" by auto
  moreover have "set (map ?l ss2) \<subseteq> ds (snd (lex_r r1 r2)) {?l rp}" using l_imp_ll_seq_eq2[OF 2 s2] by fast
  moreover have "set (map ?l ss3) \<subseteq> ds (fst (lex_r r1 r2)) {?l rp,?l lp}" using l_imp_ll_seq2[OF 3 s3' s3] by metis
  ultimately have "\<exists> s1 s2 s3. (map ?l (snd ss) = s1@s2@s3 \<and> ELD_1' (lex_r r1 r2) (?l lp) (?l rp) s1 s2 s3)"  unfolding ll_dec ELD_1'_def using l by fast
 } note F = this
  from assms obtain s rl q \<sigma> t rl2 q2 \<sigma>2 t2 where p_dec: "p = ((s,rl,q,\<sigma>,t),(s,rl2,q2,\<sigma>2,t2))" 
   and step1: "(s,t) \<in> rstep_r_p_s R rl q \<sigma>" and step2: "(s,t2) \<in> rstep_r_p_s R rl2 q2 \<sigma>2" using assms unfolding ld2_trs_def ld_trs_def local_peaks_def get_source_def get_target_def by auto
  from assms obtain \<sigma>1 \<sigma>2 \<sigma>3 where decl: "map l2 (snd jl) = \<sigma>1@\<sigma>2@\<sigma>3" 
   and ELDl: "ELD_1' r2 (l2 (fst p)) (l2 (snd p)) \<sigma>1 \<sigma>2 \<sigma>3" and seql: "jl \<in> seq R" unfolding ld2_trs_def ld_trs_def eld_seq_def by fast
  have "(get_source (fst p),get_target (fst p)) \<in> rstep R" using step1 unfolding p_dec get_source_def get_target_def apply auto by (metis rstep_r_p_s_imp_rstep)
  hence stepl: "(get_source (fst p), first jl) \<in> (rstep R)\<^sup>*" using assms unfolding ld2_trs_def ld_trs_def by auto
  have "(get_source (snd p),get_target (snd p)) \<in> rstep R" using step2 unfolding p_dec get_source_def get_target_def apply auto by (metis rstep_r_p_s_imp_rstep)
  hence stepr: "(get_source (snd p), first jr) \<in> (rstep R)\<^sup>*" using assms unfolding ld2_trs_def ld_trs_def by auto
  from assms obtain \<tau>1 \<tau>2 \<tau>3 where decr: "map l2 (snd jr) = \<tau>1@\<tau>2@\<tau>3" 
   and ELDr: "ELD_1' r2 (l2 (snd p)) (l2 (fst p)) \<tau>1 \<tau>2 \<tau>3" and seqr: "jr \<in> seq R" unfolding ld2_trs_def ld_trs_def eld_seq_def by fast
 have A5: "get_source (fst p) = get_source (snd p)" using assms unfolding ld2_trs_def ld_trs_def local_peaks_def get_source_def by auto
   show ?thesis unfolding ld_trs_def local_peaks_def eld_seq_def 
   using F[OF decl ELDl seql stepl A5] F[OF decr ELDr seqr stepr A5[symmetric]] by fast
 qed
  *)
 
lemma SN_rel_diamond_trs_imp_eld_seq:
  assumes "diamond_trs R p jl jr"
  shows "eld_seq source_labeling (sl_rel (R_d R) (R_nd R)) p jl jr" (is "eld_seq ?l ?r _ _ _")
proof -
  have "ld_trs R p jl jr" and l1: "length (snd jl) \<le> 1" and l2: "length (snd jr) \<le> 1"
    using assms unfolding diamond_trs_def by auto
  obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u where p_dec: "p = ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u))"
    using assms unfolding local_peaks_def diamond_trs_def ld_trs_def by blast
  have st: "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" and su: "(s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2"
    using assms unfolding local_peaks_def diamond_trs_def ld_trs_def p_dec by auto
  have jl: "jl \<in> seq R" and jr: "jr \<in> seq R" using assms unfolding ld_trs_def diamond_trs_def by auto
  have eq: "fst jl = first jl" by (metis first.simps prod.collapse)
  have step: "(s,t) \<in> (rstep R)^*" using st by (metis r_into_rtrancl rstep_r_p_s_imp_rstep)
  hence ast:"(s,fst jl) \<in> (rstep R)^*" using assms
    unfolding diamond_trs_def ld_trs_def p_dec get_target_def eq[symmetric] by auto
  have "\<forall>i < length (snd jl). (s,get_source ((snd jl)!i)) \<in> (rstep R)^*"
    using r_ast_seq_imp_r_ast[OF ast eq jl] by auto
  hence la: "set (map ?l (snd jl)) \<subseteq> ds (snd ?r) {?l (snd p)}"
    unfolding p_dec ds_def get_source_def apply auto 
    by (metis (no_types, hide_lams) R_d_union_R_nd fst_conv in_set_conv_nth inf_sup_aci(5) rstep_union)
  have 1: "ELD_1' ?r (?l (fst p)) (?l (snd p)) [] (map ?l (snd jl)) []"
    using la l1 unfolding ELD_1'_def by auto
  (* symmetric *)
  have eq: "fst jr = first jr" by (metis first.simps prod.collapse)
  have step: "(s,u) \<in> (rstep R)^*" using su by (metis r_into_rtrancl rstep_r_p_s_imp_rstep)
  hence ast:"(s,fst jr) \<in> (rstep R)^*"
    using assms unfolding diamond_trs_def ld_trs_def p_dec get_target_def eq[symmetric] by auto
  have "\<forall>i < length (snd jr). (s,get_source ((snd jr)!i)) \<in> (rstep R)^*"
    using r_ast_seq_imp_r_ast[OF ast eq jr] by auto
  hence lb: "set (map ?l (snd jr)) \<subseteq> ds (snd ?r) {?l (fst p)}"
    unfolding p_dec ds_def get_source_def apply auto 
    by (metis (no_types, hide_lams) R_d_union_R_nd fst_conv in_set_conv_nth inf_sup_aci(5) rstep_union)
  have 2: "ELD_1' ?r (?l (snd p)) (?l (fst p)) [] (map ?l (snd jr)) []"
    using lb l2 unfolding ELD_1'_def by auto
  show ?thesis using 1 2 assms unfolding eld_seq_def diamond_def
    by (metis (mono_tags, hide_lams) append_Nil append_Nil2)
qed

lemma SN_rel_ld2_trs_imp_eld_seq: 
  assumes "ld2_trs R p jl jr"
  and "SN_rel (rstep (R_d R)) (rstep (R_nd R))"
  shows "eld_seq source_labeling (sl_rel (R_d R) (R_nd R)) p jl jr" (is "eld_seq ?l ?r _ _ _")
proof - 
  show ?thesis proof (cases "diamond_trs R p jl jr")
    case True thus ?thesis using SN_rel_diamond_trs_imp_eld_seq by fast
  next
    case False 
    {
      fix p jl jr 
      assume d2: "ld2_trs R p jl jr" 
      (* copy from SN_rel_diamond_trs_imp_eld_seq *)
      obtain s rl1 p1 \<sigma>1 t rl2 p2 \<sigma>2 u where p_dec: "p = ((s,rl1,p1,\<sigma>1,t),(s,rl2,p2,\<sigma>2,u))"
        using d2 unfolding ld2_trs_def local_peaks_def diamond_trs_def ld_trs_def by blast
      have st: "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" and su: "(s,u) \<in> rstep_r_p_s R rl2 p2 \<sigma>2"
        using d2 unfolding ld2_trs_def local_peaks_def diamond_trs_def ld_trs_def p_dec by auto
      have jl: "jl \<in> seq R" and jr: "jr \<in> seq R" using d2 unfolding ld2_trs_def ld_trs_def diamond_trs_def
        by auto
      have l1: "length (snd jr) \<le> 1" using d2 unfolding ld2_trs_def by auto
      have eq: "fst jr = first jr" by (metis first.simps prod.collapse)
      have step: "(s,u) \<in> (rstep R)^*" using su by (metis r_into_rtrancl rstep_r_p_s_imp_rstep)
      hence ast:"(s,fst jr) \<in> (rstep R)^*" using d2
        unfolding ld2_trs_def diamond_trs_def ld_trs_def p_dec get_target_def eq[symmetric] by auto
      have "\<forall>i < length (snd jr). (s,get_source ((snd jr)!i)) \<in> (rstep R)^*"
        using r_ast_seq_imp_r_ast[OF ast eq jr] by auto
      hence la: "set (map ?l (snd jr)) \<subseteq> ds (snd ?r) {?l (fst p)}"
        unfolding p_dec ds_def get_source_def apply auto 
        by (metis (no_types, hide_lams) R_d_union_R_nd fst_conv in_set_conv_nth inf_sup_aci(5) rstep_union)
      have 1: "ELD_1' ?r (?l (snd p)) (?l (fst p)) [] (map ?l (snd jr)) []"
        using la l1 unfolding ELD_1'_def by auto
      have 2: "\<exists> \<tau>1 \<tau>2 \<tau>3. map ?l (snd jl) = \<tau>1@\<tau>2@\<tau>3 \<and> ELD_1' ?r (?l (fst p)) (?l (snd p)) \<tau>1 \<tau>2 \<tau>3"
      proof (cases "length (snd jl) \<le> 1")
        case True
        have eq: "fst jl = first jl" by (metis first.simps prod.collapse)
        have step: "(s,t) \<in> (rstep R)^*" using st by (metis r_into_rtrancl rstep_r_p_s_imp_rstep)
        hence ast:"(s,fst jl) \<in> (rstep R)^*" using d2 unfolding ld2_trs_def diamond_trs_def ld_trs_def p_dec get_target_def eq[symmetric] by auto
        have "\<forall>i < length (snd jl). (s,get_source ((snd jl)!i)) \<in> (rstep R)^*" using r_ast_seq_imp_r_ast[OF ast eq jl] by auto
        hence la: "set (map ?l (snd jl)) \<subseteq> ds (snd ?r) {?l (snd p)}" unfolding p_dec ds_def get_source_def apply auto 
          by (metis (no_types, hide_lams) R_d_union_R_nd fst_conv in_set_conv_nth inf_sup_aci(5) rstep_union)
        have "ELD_1' ?r (?l (fst p)) (?l (snd p)) [] (map ?l (snd jl)) []" using la True unfolding ELD_1'_def by (metis empty_subsetI length_map list.set(1))
        thus ?thesis by (metis append_Nil append_Nil2)
      next
        case False hence nl: "\<not> linear_term (snd (get_rule (fst p)))" using d2 unfolding ld2_trs_def by auto
        have "(s,t) \<in> rstep_r_p_s R rl1 p1 \<sigma>1" using st by auto
        hence x: "(s,t) \<in> rstep_r_p_s (R_d R \<union> R_nd R) rl1 p1 \<sigma>1" unfolding R_d_def R_nd_def by (metis R_d_def R_d_union_R_nd R_nd_def)
        hence "(s,t) \<in> rstep_r_p_s (R_d R) rl1 p1 \<sigma>1"
        proof (cases "rl1 \<in> R_d R")
          case True thus ?thesis using x unfolding R_d_def rstep_r_p_s_def by auto
        next
          case False 
          have "rl1 \<in> R" using st unfolding rstep_r_p_s_def apply auto by metis  
          hence "rl1 \<in> R_nd R" using False unfolding R_d_def R_nd_def by auto
          thus ?thesis using nl unfolding p_dec get_rule_def R_nd_def by auto
        qed
        hence st2: "(s,t) \<in> rstep (R_d R)" by (metis rstep_r_p_s_imp_rstep)
        hence "(t,s) \<in> fst ?r" unfolding sl_rel.simps by auto
        from st2 have st3: "(s,t) \<in> (relto (rstep (R_d R)) (rstep (R_nd R)))^+" by auto
        have uu: "(t,t) \<in> (rstep R)^*" by auto
        have eq: "t = first jl" using d2 unfolding ld2_trs_def ld_trs_def unfolding p_dec get_target_def by auto 
        from r_ast_seq_imp_r_ast[OF uu eq jl] have y: "\<forall>i<length (snd jl). (t, get_source (snd jl ! i)) \<in> (rstep R)\<^sup>*" by auto
        {
          fix i assume A: "i < length (snd jl)"
          have "(t,get_source (snd jl !i)) \<in> (rstep R)^*" (is "(t,?v) \<in> _") using A y by auto
          hence l: "(t,?v) \<in> (rstep (R_d R) \<union> rstep (R_nd R))^*" by (metis R_d_union_R_nd rstep_union)
          have k: "(rstep (R_d R) \<union> rstep (R_nd R))^* \<subseteq> ((rstep (R_nd R))^* \<union> (relto (rstep (R_d R)) (rstep (R_nd R)))^*)" by regexp
          hence "(t,?v) \<in> (relto (rstep (R_d R)) (rstep (R_nd R)))^* \<or> (t,?v) \<in> (rstep (R_nd R))^*" using l by auto 
          hence "(?v, s) \<in> fst (sl_rel (R_d R) (R_nd R))" using st3 apply auto by (metis RS_S_trancl)
        }
        hence "\<forall>i<length (snd jl). (get_source (snd jl ! i),s) \<in> (fst ?r)"  using y by auto
        hence "set (map ?l (snd jl)) \<subseteq> ds (fst ?r) {?l (fst p)}" unfolding p_dec get_source_def ds_def apply auto by (metis fst_conv in_set_idx)
        thus ?thesis unfolding ELD_1'_def apply auto by (metis (erased, hide_lams) add.commute append_Nil2 empty_subsetI le_add1 list.set(1) list.size(3) monoid_add_class.add.right_neutral set_map)
      qed
      have "eld_seq source_labeling ?r p jl jr" using 1 2 unfolding eld_seq_def by (metis append_Nil append_Nil2) 
    } note T1 = this
    from assms have "ld2_trs R p jl jr" unfolding ld2_trs_def unfolding ld_trs_def by auto
    thus ?thesis using T1 by auto 
  qed
qed

lemma SN_rel_compatible:
  assumes wll: "weakly_compatible R l r" and rp: "rel_props (fst r) (snd r)"
  and SN: "SN_rel (rstep (R_d R)) (rstep (R_nd R))"
  and ll: "left_linear_trs R"
  defines "ll \<equiv> (\<lambda> step. (source_labeling step,l step))"
  and "rr \<equiv> (lex_r (sl_rel (R_d R) (R_nd R)) r)" (is "_ \<equiv> lex_r ?r1 ?r2")
  shows "compatible R ll rr"
unfolding compatible_def
proof
  have "is_labeling R source_labeling ?r1" using SN_rel_is_labeling by auto
  thus "is_labeling R ll rr" using wll is_labeling_lex unfolding ll_def rr_def weakly_compatible_def by fast
next
  {
    fix p assume A: "variable_peak R p \<or> parallel_peak R p"
    hence "weld R l r p" using ll using wll ll unfolding weakly_compatible_def by fast
    then obtain jl jr where d2: "ld2_trs R p jl jr" and weld_seq:"weld_seq l r p jl jr" unfolding weld_def by auto
    from d2 have d: "ld_trs R p jl jr" using d2 unfolding ld2_trs_def by auto
    have eld_seq: "eld_seq source_labeling ?r1 p jl jr" using SN_rel_ld2_trs_imp_eld_seq[OF d2 SN]  by auto
    hence "eld_seq ll rr p jl jr" using weld_seq d2 eld_seq_and_weld_seq_imp_eld_seq unfolding ll_def rr_def by fast
    hence "eld R ll rr p" using d unfolding weld_def eld_def by fast
  }
  thus "\<forall>p. variable_peak R p \<or> parallel_peak R p \<longrightarrow> eld R ll rr p" (is "?xp") by auto
qed
  
lemma SN_rel_lex:
  assumes wll: "weakly_compatible R l r" and rp: "rel_props (fst r) (snd r)"
  and SN: "SN_rel (rstep (R_d R)) (rstep (R_nd R))" 
  and dec: "\<forall> (b,p) \<in> critical_peaks R. eld R l r p"
  and ll: "left_linear_trs R"
  defines "ll \<equiv> (\<lambda> step. (source_labeling step,l step))"
  and "rr \<equiv> (lex_r (sl_rel (R_d R) (R_nd R)) r)" (is "_ \<equiv> lex_r ?r1 ?r2")
  shows "rel_props (fst rr) (snd rr)"
  and "compatible R ll rr"
  and  "\<forall> (b,p) \<in> critical_peaks R. eld R ll rr p"
proof -
  have rp1:"rel_props (fst ?r1) (snd ?r1)" using SN_rel_has_rel_props[OF SN] by auto 
  show "rel_props (fst rr) (snd rr)" using rel_props_lex[OF rp1 rp] unfolding rr_def by fast
  show "compatible R ll rr" using assms SN_rel_compatible by fast
  show "\<forall>(b, p)\<in>critical_peaks R. eld R ll rr p" using eld_imp_eld dec unfolding ll_def rr_def by fast
qed

lemma rule_labeling_is_sound_ll:
  defines "r \<equiv> ({(n,m). n < (m::nat)},{(n,m). n \<le> (m::nat)})"
  assumes ll: "left_linear_trs R"
  and SN: "SN_rel (rstep (R_d R)) (rstep (R_nd R))"
  and dec: "\<forall> (b,p) \<in> critical_peaks R. eld R (rule_labeling i) r p" (is "\<forall> (b,p) \<in> _ . eld _ ?l ?r _")
  shows "CR (rstep R)"
proof -
  have r: "refl (snd r)" unfolding r_def by (simp add: refl_on_def)
  have rp: "rel_props (fst r) (snd r)" using rule_labeling_has_rel_props unfolding r_def by fast
  have el: "weakly_compatible R (rule_labeling i) r" using rule_labeling_is_weakly_compatible[OF r ll] by auto
  let ?l = "(\<lambda> step. (source_labeling step,rule_labeling i step))"
  let ?r = "(lex_r (sl_rel (R_d R) (R_nd R)) r)"
  from SN_rel_lex[OF el rp SN dec ll] show ?thesis using main ll by blast
qed

hide_const (open) last

end
